# Python modules
import os

# 3rd party modules
import numpy as np

from scipy.signal import fftconvolve, convolve
from math import ceil
from math import sqrt



def dwt(wavelet, data, levels):
    """
    Assumes that we deal with data wrap as a periodic function
    
    """
    alerts = np.array([])
    coeffs = np.array([])
    widths = np.array([])
    
    lmax   = int(np.ceil(np.log2(len(data))))-2 
    levels = levels if levels < lmax else lmax 
    
    # Ensure data is even length
    if len(data)%2 != 0:
        data   = np.append(data, data[-1])
        alerts = np.append(alerts, 1)
    else:
        alerts = np.append(alerts, 0)

    widths = np.append(widths, len(data))
    alerts = np.append(alerts, levels)
    alerts = np.append(alerts, 0)
        
    orig = data

    mwave = get_mwave(wavelet)

    # Iterate through levels applying wavelet filter
    
    for level in range(levels):
        ca,cd   = dwt1(orig, mwave)
        coeffs  = np.append(cd, coeffs)
        widths  = np.append(len(cd), widths)
        if level==levels-1:
            coeffs = np.append(ca, coeffs)
            widths = np.append(len(ca), widths)
        orig = ca

    return coeffs, widths, alerts


def dwt1(data, mwave):
    lo, hi, _, _ = mwave

    npad = int(len(lo) / 2 + len(hi) / 2)
    npad2 = int(npad / 2.0)

    # Apply periodic extension to data as needed
    # - ensure even number of points
    # - pad with periodic data
    data = np.append(data, data[-1]) if len(data) % 2 else data
    data = np.pad(data, (npad2, npad2), 'wrap')

    # Filter data to get approx and detail coefficients
    ca = np.real(convolve(data, lo))  # was fftconvolve(), this is faster
    cd = np.real(convolve(data, hi))

    # Extract from extended array and down sample by 2
    ca = ca[npad:-npad:2].copy()
    cd = cd[npad:-npad:2].copy()

    return ca, cd


def idwt1(ca, cd, mwave):

    _, _, lo, hi = mwave

    navg  = int(int(len(lo))/2 + int(len(hi))/2)
    navg2 = int(navg / 2.0)
    npts  = len(cd)*2
    ufact = 2           # up sample factor of 2 

    # Processing steps
    # - upsample by factor of 2
    # - ensure even number of points
    # - pad with periodic data
    # - convolve with wavelet mother kernel
    # - repeat for both approx and detail coeffs
    
    ca = np.append(np.insert(ca, slice(1,None,ufact-1), 0.0), 0.0)  
    ca = np.append(ca, ca[-1]) if len(ca)%2 else ca 
    ca = np.pad(ca, (navg2,navg2), 'wrap')
    recona = np.real(convolve(ca, lo))      # was fftconvolve(), this is faster

    cd = np.append(np.insert(cd, slice(1,None,ufact-1), 0.0), 0.0)  
    cd = np.append(cd, cd[-1]) if len(cd)%2 else cd 
    cd = np.pad(cd, (navg2,navg2), 'wrap')
    recond = np.real(convolve(cd, hi))
    
    recon = recona[navg-1:npts+navg-1] + recond[navg-1:npts+navg-1]
    
    return recon



def idwt(wavelet, coeffs, widths, alerts):

    levels = int(alerts[1])
    nca    = int(widths[0])
    ncd    = int(widths[1])
    ca     = coeffs[0:nca]
    cd     = coeffs[nca:2*nca]

    mwave = get_mwave(wavelet)

    for i in range(levels):
        
        recon = idwt1(ca,cd,mwave)
            
        nca += ncd
        if i < levels-1:
            ca  = recon
            ncd = int(widths[i+2])
            cd  = coeffs[nca:nca+ncd]

            if len(ca) > len(cd):
                t = len(ca)-len(cd)
                lent = int(np.floor(t * 0.5))
                ca = ca[:lent+len(cd)]
                ca = ca[lent:]
    
    pad   = int(alerts[0])
    recon = recon[0:len(recon)-pad]
    
    return recon


def get_mwave_names():
    '''
    Mother wavelet family names, just FYI.
    Scrounged from PyWavelets 'import pywt' via brute force

    '''
    val = ['bior1.1', 'bior1.3', 'bior1.5', 'bior2.2', 'bior2.4', 'bior2.6', 'bior2.8', 'bior3.1',
           'bior3.3', 'bior3.5', 'bior3.7', 'bior3.9', 'bior4.4', 'bior5.5', 'bior6.8',
           'coif1', 'coif2', 'coif3', 'coif4', 'coif5', 'coif6', 'coif7', 'coif8', 'coif9',
           'coif10', 'coif11', 'coif12', 'coif13', 'coif14', 'coif15', 'coif16', 'coif17',
           'db1', 'db2', 'db3', 'db4', 'db5', 'db6', 'db7', 'db8','db9',
           'db10', 'db11', 'db12', 'db13', 'db14', 'db15', 'db16', 'db17',
           'db18', 'db19', 'db20', 'db21', 'db22', 'db23', 'db24', 'db25',
           'db26', 'db27', 'db28', 'db29', 'db30', 'db31', 'db32', 'db33',
           'db34', 'db35', 'db36', 'db37', 'db38',
           'dmey', 'haar',
           'rbio1.1', 'rbio1.3', 'rbio1.5', 'rbio2.2', 'rbio2.4', 'rbio2.6', 'rbio2.8', 'rbio3.1',
           'rbio3.3', 'rbio3.5', 'rbio3.7', 'rbio3.9', 'rbio4.4', 'rbio5.5', 'rbio6.8',
           'sym2', 'sym3', 'sym4', 'sym5', 'sym6', 'sym7', 'sym8', 'sym9', 'sym10', 'sym11',
           'sym12', 'sym13', 'sym14', 'sym15', 'sym16', 'sym17', 'sym18', 'sym19', 'sym20']
    return(val)


def get_mwave(family):
    '''
    Mother wavelet, separated into deconstruct/reconstruct and lo/hi pass.
    Scrounged from PyWavelets 'import pywt' via brute force

    '''
    if family=="bior1.1":
        dec_lo = np.array([0.70710678118655, 0.70710678118655])
        dec_hi = np.array([-0.70710678118655, 0.70710678118655])
        rec_lo = np.array([0.70710678118655, 0.70710678118655])
        rec_hi = np.array([0.70710678118655, -0.70710678118655])
    elif family=="bior1.3":
        dec_lo = np.array([-0.08838834764832, 0.08838834764832, 0.70710678118655, 0.70710678118655, 0.08838834764832, -0.08838834764832])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.70710678118655, 0.70710678118655, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.70710678118655, 0.70710678118655, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([-0.08838834764832, -0.08838834764832, 0.70710678118655, -0.70710678118655, 0.08838834764832, 0.08838834764832])
    elif family=="bior1.5":
        dec_lo = np.array([0.01657281518406, -0.01657281518406, -0.12153397801644, 0.12153397801644, 0.70710678118655, 0.70710678118655, 0.12153397801644, -0.12153397801644, -0.01657281518406, 0.01657281518406])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.70710678118655, 0.70710678118655, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.70710678118655, 0.70710678118655, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.01657281518406, 0.01657281518406, -0.12153397801644, -0.12153397801644, 0.70710678118655, -0.70710678118655, 0.12153397801644, 0.12153397801644, -0.01657281518406, -0.01657281518406])
    elif family=="bior2.2":
        dec_lo = np.array([0.00000000000000, -0.17677669529664, 0.35355339059327, 1.06066017177982, 0.35355339059327, -0.17677669529664])
        dec_hi = np.array([-0.00000000000000, 0.35355339059327, -0.70710678118655, 0.35355339059327, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.35355339059327, 0.70710678118655, 0.35355339059327, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, 0.17677669529664, 0.35355339059327, -1.06066017177982, 0.35355339059327, 0.17677669529664])
    elif family=="bior2.4":
        dec_lo = np.array([0.00000000000000, 0.03314563036812, -0.06629126073624, -0.17677669529664, 0.41984465132951, 0.99436891104358, 0.41984465132951, -0.17677669529664, -0.06629126073624, 0.03314563036812])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.35355339059327, -0.70710678118655, 0.35355339059327, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.35355339059327, 0.70710678118655, 0.35355339059327, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.03314563036812, -0.06629126073624, 0.17677669529664, 0.41984465132951, -0.99436891104358, 0.41984465132951, 0.17677669529664, -0.06629126073624, -0.03314563036812])
    elif family=="bior2.6":
        dec_lo = np.array([0.00000000000000, -0.00690533966002, 0.01381067932005, 0.04695630968817, -0.10772329869639, -0.16987135563661, 0.44746600996961, 0.96674755240348, 0.44746600996961, -0.16987135563661, -0.10772329869639, 0.04695630968817, 0.01381067932005, -0.00690533966002])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.35355339059327, -0.70710678118655, 0.35355339059327, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.35355339059327, 0.70710678118655, 0.35355339059327, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, 0.00690533966002, 0.01381067932005, -0.04695630968817, -0.10772329869639, 0.16987135563661, 0.44746600996961, -0.96674755240348, 0.44746600996961, 0.16987135563661, -0.10772329869639, -0.04695630968817, 0.01381067932005, 0.00690533966002])
    elif family=="bior2.8":
        dec_lo = np.array([0.00000000000000, 0.00151054305063, -0.00302108610126, -0.01294751186255, 0.02891610982635, 0.05299848189069, -0.13491307360774, -0.16382918343409, 0.46257144047592, 0.95164212189718, 0.46257144047592, -0.16382918343409, -0.13491307360774, 0.05299848189069, 0.02891610982635, -0.01294751186255, -0.00302108610126, 0.00151054305063])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.35355339059327, -0.70710678118655, 0.35355339059327, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.35355339059327, 0.70710678118655, 0.35355339059327, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.00151054305063, -0.00302108610126, 0.01294751186255, 0.02891610982635, -0.05299848189069, -0.13491307360774, 0.16382918343409, 0.46257144047592, -0.95164212189718, 0.46257144047592, 0.16382918343409, -0.13491307360774, -0.05299848189069, 0.02891610982635, 0.01294751186255, -0.00302108610126, -0.00151054305063])
    elif family=="bior3.1":
        dec_lo = np.array([-0.35355339059327, 1.06066017177982, 1.06066017177982, -0.35355339059327])
        dec_hi = np.array([-0.17677669529664, 0.53033008588991, -0.53033008588991, 0.17677669529664])
        rec_lo = np.array([0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664])
        rec_hi = np.array([-0.35355339059327, -1.06066017177982, 1.06066017177982, 0.35355339059327])
    elif family=="bior3.3":
        dec_lo = np.array([0.06629126073624, -0.19887378220872, -0.15467960838456, 0.99436891104358, 0.99436891104358, -0.15467960838456, -0.19887378220872, 0.06629126073624])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.17677669529664, 0.53033008588991, -0.53033008588991, 0.17677669529664, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.06629126073624, 0.19887378220872, -0.15467960838456, -0.99436891104358, 0.99436891104358, 0.15467960838456, -0.19887378220872, -0.06629126073624])
    elif family=="bior3.5":
        dec_lo = np.array([-0.01381067932005, 0.04143203796015, 0.05248058141619, -0.26792717880897, -0.07181553246426, 0.96674755240348, 0.96674755240348, -0.07181553246426, -0.26792717880897, 0.05248058141619, 0.04143203796015, -0.01381067932005])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.17677669529664, 0.53033008588991, -0.53033008588991, 0.17677669529664, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([-0.01381067932005, -0.04143203796015, 0.05248058141619, 0.26792717880897, -0.07181553246426, -0.96674755240348, 0.96674755240348, 0.07181553246426, -0.26792717880897, -0.05248058141619, 0.04143203796015, 0.01381067932005])
    elif family=="bior3.7":
        dec_lo = np.array([0.00302108610126, -0.00906325830378, -0.01683176542131, 0.07466398507402, 0.03133297870736, -0.30115912592284, -0.02649924094535, 0.95164212189718, 0.95164212189718, -0.02649924094535, -0.30115912592284, 0.03133297870736, 0.07466398507402, -0.01683176542131, -0.00906325830378, 0.00302108610126])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.17677669529664, 0.53033008588991, -0.53033008588991, 0.17677669529664, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00302108610126, 0.00906325830378, -0.01683176542131, -0.07466398507402, 0.03133297870736, 0.30115912592284, -0.02649924094535, -0.95164212189718, 0.95164212189718, 0.02649924094535, -0.30115912592284, -0.03133297870736, 0.07466398507402, 0.01683176542131, -0.00906325830378, -0.00302108610126])
    elif family=="bior3.9":
        dec_lo = np.array([-0.00067974437278, 0.00203923311835, 0.00506031921961, -0.02061891264111, -0.01411278793018, 0.09913478249423, 0.01230013626942, -0.32019196836078, 0.00205002271157, 0.94212570067821, 0.94212570067821, 0.00205002271157, -0.32019196836078, 0.01230013626942, 0.09913478249423, -0.01411278793018, -0.02061891264111, 0.00506031921961, 0.00203923311835, -0.00067974437278])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.17677669529664, 0.53033008588991, -0.53033008588991, 0.17677669529664, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([-0.00067974437278, -0.00203923311835, 0.00506031921961, 0.02061891264111, -0.01411278793018, -0.09913478249423, 0.01230013626942, 0.32019196836078, 0.00205002271157, -0.94212570067821, 0.94212570067821, -0.00205002271157, -0.32019196836078, -0.01230013626942, 0.09913478249423, 0.01411278793018, -0.02061891264111, -0.00506031921961, 0.00203923311835, 0.00067974437278])
    elif family=="bior4.4":
        dec_lo = np.array([0.00000000000000, 0.03782845550726, -0.02384946501956, -0.11062440441844, 0.37740285561283, 0.85269867900889, 0.37740285561283, -0.11062440441844, -0.02384946501956, 0.03782845550726])
        dec_hi = np.array([-0.00000000000000, -0.06453888262870, 0.04068941760916, 0.41809227322162, -0.78848561640558, 0.41809227322162, 0.04068941760916, -0.06453888262870, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, -0.06453888262870, -0.04068941760916, 0.41809227322162, 0.78848561640558, 0.41809227322162, -0.04068941760916, -0.06453888262870, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.03782845550726, -0.02384946501956, 0.11062440441844, 0.37740285561283, -0.85269867900889, 0.37740285561283, 0.11062440441844, -0.02384946501956, -0.03782845550726])
    elif family=="bior5.5":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.03968708834741, 0.00794810863724, -0.05446378846824, 0.34560528195603, 0.73666018142821, 0.34560528195603, -0.05446378846824, 0.00794810863724, 0.03968708834741, 0.00000000000000])
        dec_hi = np.array([-0.01345670945912, -0.00269496688011, 0.13670658466433, -0.09350469740094, -0.47680326579848, 0.89950610974865, -0.47680326579848, -0.09350469740094, 0.13670658466433, -0.00269496688011, -0.01345670945912, 0.00000000000000])
        rec_lo = np.array([0.01345670945912, -0.00269496688011, -0.13670658466433, -0.09350469740094, 0.47680326579848, 0.89950610974865, 0.47680326579848, -0.09350469740094, -0.13670658466433, -0.00269496688011, 0.01345670945912, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.03968708834741, -0.00794810863724, -0.05446378846824, -0.34560528195603, 0.73666018142821, -0.34560528195603, -0.05446378846824, -0.00794810863724, 0.03968708834741, -0.00000000000000])
    elif family=="bior6.8":
        dec_lo = np.array([0.00000000000000, 0.00190883173648, -0.00191428612909, -0.01699063986760, 0.01193456527973, 0.04973290349094, -0.07726317316720, -0.09405920349574, 0.42079628460983, 0.82592299745840, 0.42079628460983, -0.09405920349574, -0.07726317316720, 0.04973290349094, 0.01193456527973, -0.01699063986760, -0.00191428612909, 0.00190883173648])
        dec_hi = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.01442628250562, -0.01446750489679, -0.07872200106263, 0.04036797903034, 0.41784910915027, -0.75890772945365, 0.41784910915027, 0.04036797903034, -0.07872200106263, -0.01446750489679, 0.01442628250562, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.01442628250562, 0.01446750489679, -0.07872200106263, -0.04036797903034, 0.41784910915027, 0.75890772945365, 0.41784910915027, -0.04036797903034, -0.07872200106263, 0.01446750489679, 0.01442628250562, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.00190883173648, -0.00191428612909, 0.01699063986760, 0.01193456527973, -0.04973290349094, -0.07726317316720, 0.09405920349574, 0.42079628460983, -0.82592299745840, 0.42079628460983, 0.09405920349574, -0.07726317316720, -0.04973290349094, 0.01193456527973, 0.01699063986760, -0.00191428612909, -0.00190883173648])
    elif family=="coif1":
        dec_lo = np.array([-0.01565572813579, -0.07273261951253, 0.38486484686486, 0.85257202021160, 0.33789766245748, -0.07273261951253])
        dec_hi = np.array([0.07273261951253, 0.33789766245748, -0.85257202021160, 0.38486484686486, 0.07273261951253, -0.01565572813579])
        rec_lo = np.array([-0.07273261951253, 0.33789766245748, 0.85257202021160, 0.38486484686486, -0.07273261951253, -0.01565572813579])
        rec_hi = np.array([-0.01565572813579, 0.07273261951253, 0.38486484686486, -0.85257202021160, 0.33789766245748, 0.07273261951253])
    elif family=="coif2":
        dec_lo = np.array([-0.00072054944552, -0.00182320887091, 0.00561143481937, 0.02368017194685, -0.05943441864643, -0.07648859907828, 0.41700518442324, 0.81272363544941, 0.38611006682276, -0.06737255472373, -0.04146493678687, 0.01638733646320])
        dec_hi = np.array([-0.01638733646320, -0.04146493678687, 0.06737255472373, 0.38611006682276, -0.81272363544941, 0.41700518442324, 0.07648859907828, -0.05943441864643, -0.02368017194685, 0.00561143481937, 0.00182320887091, -0.00072054944552])
        rec_lo = np.array([0.01638733646320, -0.04146493678687, -0.06737255472373, 0.38611006682276, 0.81272363544941, 0.41700518442324, -0.07648859907828, -0.05943441864643, 0.02368017194685, 0.00561143481937, -0.00182320887091, -0.00072054944552])
        rec_hi = np.array([-0.00072054944552, 0.00182320887091, 0.00561143481937, -0.02368017194685, -0.05943441864643, 0.07648859907828, 0.41700518442324, -0.81272363544941, 0.38611006682276, 0.06737255472373, -0.04146493678687, -0.01638733646320])
    elif family=="coif3":
        dec_lo = np.array([-0.00003459977320, -0.00007098330251, 0.00046621695982, 0.00111751877083, -0.00257451768814, -0.00900797613673, 0.01588054486367, 0.03455502757330, -0.08230192710630, -0.07179982161915, 0.42848347637737, 0.79377722262609, 0.40517690240912, -0.06112339000297, -0.06577191128147, 0.02345269614208, 0.00778259642567, -0.00379351286438])
        dec_hi = np.array([0.00379351286438, 0.00778259642567, -0.02345269614208, -0.06577191128147, 0.06112339000297, 0.40517690240912, -0.79377722262609, 0.42848347637737, 0.07179982161915, -0.08230192710630, -0.03455502757330, 0.01588054486367, 0.00900797613673, -0.00257451768814, -0.00111751877083, 0.00046621695982, 0.00007098330251, -0.00003459977320])
        rec_lo = np.array([-0.00379351286438, 0.00778259642567, 0.02345269614208, -0.06577191128147, -0.06112339000297, 0.40517690240912, 0.79377722262609, 0.42848347637737, -0.07179982161915, -0.08230192710630, 0.03455502757330, 0.01588054486367, -0.00900797613673, -0.00257451768814, 0.00111751877083, 0.00046621695982, -0.00007098330251, -0.00003459977320])
        rec_hi = np.array([-0.00003459977320, 0.00007098330251, 0.00046621695982, -0.00111751877083, -0.00257451768814, 0.00900797613673, 0.01588054486367, -0.03455502757330, -0.08230192710630, 0.07179982161915, 0.42848347637737, -0.79377722262609, 0.40517690240912, 0.06112339000297, -0.06577191128147, -0.02345269614208, 0.00778259642567, 0.00379351286438])
    elif family=="coif4":
        dec_lo = np.array([-0.00000178499091, -0.00000325964794, 0.00003122986160, 0.00006233885431, -0.00025997433712, -0.00058902022463, 0.00126656107893, 0.00375143469715, -0.00565828380013, -0.01521172818770, 0.02508225333795, 0.03933442260559, -0.09622042453595, -0.06662747236682, 0.43438603311436, 0.78223893442428, 0.41530842700068, -0.05607731960357, -0.08126671024919, 0.02668230466960, 0.01606894713158, -0.00734616793627, -0.00162949242523, 0.00089231390254])
        dec_hi = np.array([-0.00089231390254, -0.00162949242523, 0.00734616793627, 0.01606894713158, -0.02668230466960, -0.08126671024919, 0.05607731960357, 0.41530842700068, -0.78223893442428, 0.43438603311436, 0.06662747236682, -0.09622042453595, -0.03933442260559, 0.02508225333795, 0.01521172818770, -0.00565828380013, -0.00375143469715, 0.00126656107893, 0.00058902022463, -0.00025997433712, -0.00006233885431, 0.00003122986160, 0.00000325964794, -0.00000178499091])
        rec_lo = np.array([0.00089231390254, -0.00162949242523, -0.00734616793627, 0.01606894713158, 0.02668230466960, -0.08126671024919, -0.05607731960357, 0.41530842700068, 0.78223893442428, 0.43438603311436, -0.06662747236682, -0.09622042453595, 0.03933442260559, 0.02508225333795, -0.01521172818770, -0.00565828380013, 0.00375143469715, 0.00126656107893, -0.00058902022463, -0.00025997433712, 0.00006233885431, 0.00003122986160, -0.00000325964794, -0.00000178499091])
        rec_hi = np.array([-0.00000178499091, 0.00000325964794, 0.00003122986160, -0.00006233885431, -0.00025997433712, 0.00058902022463, 0.00126656107893, -0.00375143469715, -0.00565828380013, 0.01521172818770, 0.02508225333795, -0.03933442260559, -0.09622042453595, 0.06662747236682, 0.43438603311436, -0.78223893442428, 0.41530842700068, 0.05607731960357, -0.08126671024919, -0.02668230466960, 0.01606894713158, 0.00734616793627, -0.00162949242523, -0.00089231390254])
    elif family=="coif5":
        dec_lo = np.array([-0.00000009604010, -0.00000016237995, 0.00000206122040, 0.00000370072771, -0.00002127022167, -0.00004121986192, 0.00014035632812, 0.00030185794167, -0.00063755892613, -0.00166162730393, 0.00243157544254, 0.00676152022062, -0.00915950733868, -0.01975839160097, 0.03267479946706, 0.04128753047212, -0.10556315130734, -0.06203775157498, 0.43798230665916, 0.77429362286033, 0.42157126673075, -0.05204667025355, -0.09192158806009, 0.02816974427053, 0.02340832211893, -0.01013158484690, -0.00415931262758, 0.00217829437785, 0.00035857774116, -0.00021208186207])
        dec_hi = np.array([0.00021208186207, 0.00035857774116, -0.00217829437785, -0.00415931262758, 0.01013158484690, 0.02340832211893, -0.02816974427053, -0.09192158806009, 0.05204667025355, 0.42157126673075, -0.77429362286033, 0.43798230665916, 0.06203775157498, -0.10556315130734, -0.04128753047212, 0.03267479946706, 0.01975839160097, -0.00915950733868, -0.00676152022062, 0.00243157544254, 0.00166162730393, -0.00063755892613, -0.00030185794167, 0.00014035632812, 0.00004121986192, -0.00002127022167, -0.00000370072771, 0.00000206122040, 0.00000016237995, -0.00000009604010])
        rec_lo = np.array([-0.00021208186207, 0.00035857774116, 0.00217829437785, -0.00415931262758, -0.01013158484690, 0.02340832211893, 0.02816974427053, -0.09192158806009, -0.05204667025355, 0.42157126673075, 0.77429362286033, 0.43798230665916, -0.06203775157498, -0.10556315130734, 0.04128753047212, 0.03267479946706, -0.01975839160097, -0.00915950733868, 0.00676152022062, 0.00243157544254, -0.00166162730393, -0.00063755892613, 0.00030185794167, 0.00014035632812, -0.00004121986192, -0.00002127022167, 0.00000370072771, 0.00000206122040, -0.00000016237995, -0.00000009604010])
        rec_hi = np.array([-0.00000009604010, 0.00000016237995, 0.00000206122040, -0.00000370072771, -0.00002127022167, 0.00004121986192, 0.00014035632812, -0.00030185794167, -0.00063755892613, 0.00166162730393, 0.00243157544254, -0.00676152022062, -0.00915950733868, 0.01975839160097, 0.03267479946706, -0.04128753047212, -0.10556315130734, 0.06203775157498, 0.43798230665916, -0.77429362286033, 0.42157126673075, 0.05204667025355, -0.09192158806009, -0.02816974427053, 0.02340832211893, 0.01013158484690, -0.00415931262758, -0.00217829437785, 0.00035857774116, 0.00021208186207])
    elif family=="coif6":
        dec_lo = np.array([-0.00000000530909, -0.00000000848714, 0.00000013503245, 0.00000022559979, -0.00000165961930, -0.00000292438556, 0.00001313985135, 0.00002473655933, -0.00007528004307, -0.00015457719928, 0.00032522235901, 0.00076985473075, -0.00115743501343, -0.00307393950721, 0.00385765827059, 0.00959109017590, -0.01265006790873, -0.02295015327985, 0.03888132625151, 0.04185249067614, -0.11226080796482, -0.05810891797261, 0.44040119112685, 0.76840325757989, 0.42581954501284, -0.04876407217567, -0.09967300204601, 0.02878611434667, 0.02964577289132, -0.01223157779004, -0.00702940639100, 0.00353901987154, 0.00109162471233, -0.00062461304393, -0.00008117002627, 0.00005077548784])
        dec_hi = np.array([-0.00005077548784, -0.00008117002627, 0.00062461304393, 0.00109162471233, -0.00353901987154, -0.00702940639100, 0.01223157779004, 0.02964577289132, -0.02878611434667, -0.09967300204601, 0.04876407217567, 0.42581954501284, -0.76840325757989, 0.44040119112685, 0.05810891797261, -0.11226080796482, -0.04185249067614, 0.03888132625151, 0.02295015327985, -0.01265006790873, -0.00959109017590, 0.00385765827059, 0.00307393950721, -0.00115743501343, -0.00076985473075, 0.00032522235901, 0.00015457719928, -0.00007528004307, -0.00002473655933, 0.00001313985135, 0.00000292438556, -0.00000165961930, -0.00000022559979, 0.00000013503245, 0.00000000848714, -0.00000000530909])
        rec_lo = np.array([0.00005077548784, -0.00008117002627, -0.00062461304393, 0.00109162471233, 0.00353901987154, -0.00702940639100, -0.01223157779004, 0.02964577289132, 0.02878611434667, -0.09967300204601, -0.04876407217567, 0.42581954501284, 0.76840325757989, 0.44040119112685, -0.05810891797261, -0.11226080796482, 0.04185249067614, 0.03888132625151, -0.02295015327985, -0.01265006790873, 0.00959109017590, 0.00385765827059, -0.00307393950721, -0.00115743501343, 0.00076985473075, 0.00032522235901, -0.00015457719928, -0.00007528004307, 0.00002473655933, 0.00001313985135, -0.00000292438556, -0.00000165961930, 0.00000022559979, 0.00000013503245, -0.00000000848714, -0.00000000530909])
        rec_hi = np.array([-0.00000000530909, 0.00000000848714, 0.00000013503245, -0.00000022559979, -0.00000165961930, 0.00000292438556, 0.00001313985135, -0.00002473655933, -0.00007528004307, 0.00015457719928, 0.00032522235901, -0.00076985473075, -0.00115743501343, 0.00307393950721, 0.00385765827059, -0.00959109017590, -0.01265006790873, 0.02295015327985, 0.03888132625151, -0.04185249067614, -0.11226080796482, 0.05810891797261, 0.44040119112685, -0.76840325757989, 0.42581954501284, 0.04876407217567, -0.09967300204601, -0.02878611434667, 0.02964577289132, 0.01223157779004, -0.00702940639100, -0.00353901987154, 0.00109162471233, 0.00062461304393, -0.00008117002627, -0.00005077548784])
    elif family=="coif7":
        dec_lo = np.array([-0.00000000029906, -0.00000000045783, 0.00000000879659, 0.00000001393510, -0.00000012550913, -0.00000020693205, 0.00000115797691, 0.00000200207805, -0.00000777124355, -0.00001423563698, 0.00004043048242, 0.00007971050026, -0.00016781721215, -0.00036906682873, 0.00057949944823, 0.00143474185665, -0.00180153728333, -0.00461784213043, 0.00543131644288, 0.01205233824184, -0.01594684681957, -0.02515425756854, 0.04399304616308, 0.04170535760258, -0.11729357104319, -0.05475124164815, 0.44213746140184, 0.76381536541673, 0.42888880724942, -0.04603339703847, -0.10555616822156, 0.02893704198352, 0.03491050510474, -0.01380255423629, -0.00993889526908, 0.00482944656070, 0.00210577204141, -0.00116931442858, -0.00028720237536, 0.00017510216778, 0.00001871135500, -0.00001222225062])
        dec_hi = np.array([0.00001222225062, 0.00001871135500, -0.00017510216778, -0.00028720237536, 0.00116931442858, 0.00210577204141, -0.00482944656070, -0.00993889526908, 0.01380255423629, 0.03491050510474, -0.02893704198352, -0.10555616822156, 0.04603339703847, 0.42888880724942, -0.76381536541673, 0.44213746140184, 0.05475124164815, -0.11729357104319, -0.04170535760258, 0.04399304616308, 0.02515425756854, -0.01594684681957, -0.01205233824184, 0.00543131644288, 0.00461784213043, -0.00180153728333, -0.00143474185665, 0.00057949944823, 0.00036906682873, -0.00016781721215, -0.00007971050026, 0.00004043048242, 0.00001423563698, -0.00000777124355, -0.00000200207805, 0.00000115797691, 0.00000020693205, -0.00000012550913, -0.00000001393510, 0.00000000879659, 0.00000000045783, -0.00000000029906])
        rec_lo = np.array([-0.00001222225062, 0.00001871135500, 0.00017510216778, -0.00028720237536, -0.00116931442858, 0.00210577204141, 0.00482944656070, -0.00993889526908, -0.01380255423629, 0.03491050510474, 0.02893704198352, -0.10555616822156, -0.04603339703847, 0.42888880724942, 0.76381536541673, 0.44213746140184, -0.05475124164815, -0.11729357104319, 0.04170535760258, 0.04399304616308, -0.02515425756854, -0.01594684681957, 0.01205233824184, 0.00543131644288, -0.00461784213043, -0.00180153728333, 0.00143474185665, 0.00057949944823, -0.00036906682873, -0.00016781721215, 0.00007971050026, 0.00004043048242, -0.00001423563698, -0.00000777124355, 0.00000200207805, 0.00000115797691, -0.00000020693205, -0.00000012550913, 0.00000001393510, 0.00000000879659, -0.00000000045783, -0.00000000029906])
        rec_hi = np.array([-0.00000000029906, 0.00000000045783, 0.00000000879659, -0.00000001393510, -0.00000012550913, 0.00000020693205, 0.00000115797691, -0.00000200207805, -0.00000777124355, 0.00001423563698, 0.00004043048242, -0.00007971050026, -0.00016781721215, 0.00036906682873, 0.00057949944823, -0.00143474185665, -0.00180153728333, 0.00461784213043, 0.00543131644288, -0.01205233824184, -0.01594684681957, 0.02515425756854, 0.04399304616308, -0.04170535760258, -0.11729357104319, 0.05475124164815, 0.44213746140184, -0.76381536541673, 0.42888880724942, 0.04603339703847, -0.10555616822156, -0.02893704198352, 0.03491050510474, 0.01380255423629, -0.00993889526908, -0.00482944656070, 0.00210577204141, 0.00116931442858, -0.00028720237536, -0.00017510216778, 0.00001871135500, 0.00001222225062])
    elif family=="coif8":
        dec_lo = np.array([-0.00000000001708, -0.00000000002525, 0.00000000057048, 0.00000000086700, -0.00000000927121, -0.00000001454001, 0.00000009772419, 0.00000015893517, -0.00000075150216, -0.00000127545430, 0.00000449693644, 0.00000803150300, -0.00002180200077, -0.00004147478607, 0.00008754452092, 0.00018169287648, -0.00029777893220, -0.00068717164335, 0.00089677606308, 0.00223564942205, -0.00254400371025, -0.00615665954826, 0.00706582701104, 0.01411747007762, -0.01898524469525, -0.02665671054265, 0.04825237108568, 0.04118580667626, -0.12121116823150, -0.05186074316119, 0.44344254984153, 0.76011330201794, 0.43120981555509, -0.04371898336595, -0.11016997698347, 0.02882862175929, 0.03937203787798, -0.01497846208171, -0.01274237063272, 0.00599484919216, 0.00330082501062, -0.00178326000860, -0.00062356044746, 0.00037129499561, 0.00007547367838, -0.00004829631521, -0.00000436826482, 0.00000295433652])
        dec_hi = np.array([-0.00000295433652, -0.00000436826482, 0.00004829631521, 0.00007547367838, -0.00037129499561, -0.00062356044746, 0.00178326000860, 0.00330082501062, -0.00599484919216, -0.01274237063272, 0.01497846208171, 0.03937203787798, -0.02882862175929, -0.11016997698347, 0.04371898336595, 0.43120981555509, -0.76011330201794, 0.44344254984153, 0.05186074316119, -0.12121116823150, -0.04118580667626, 0.04825237108568, 0.02665671054265, -0.01898524469525, -0.01411747007762, 0.00706582701104, 0.00615665954826, -0.00254400371025, -0.00223564942205, 0.00089677606308, 0.00068717164335, -0.00029777893220, -0.00018169287648, 0.00008754452092, 0.00004147478607, -0.00002180200077, -0.00000803150300, 0.00000449693644, 0.00000127545430, -0.00000075150216, -0.00000015893517, 0.00000009772419, 0.00000001454001, -0.00000000927121, -0.00000000086700, 0.00000000057048, 0.00000000002525, -0.00000000001708])
        rec_lo = np.array([0.00000295433652, -0.00000436826482, -0.00004829631521, 0.00007547367838, 0.00037129499561, -0.00062356044746, -0.00178326000860, 0.00330082501062, 0.00599484919216, -0.01274237063272, -0.01497846208171, 0.03937203787798, 0.02882862175929, -0.11016997698347, -0.04371898336595, 0.43120981555509, 0.76011330201794, 0.44344254984153, -0.05186074316119, -0.12121116823150, 0.04118580667626, 0.04825237108568, -0.02665671054265, -0.01898524469525, 0.01411747007762, 0.00706582701104, -0.00615665954826, -0.00254400371025, 0.00223564942205, 0.00089677606308, -0.00068717164335, -0.00029777893220, 0.00018169287648, 0.00008754452092, -0.00004147478607, -0.00002180200077, 0.00000803150300, 0.00000449693644, -0.00000127545430, -0.00000075150216, 0.00000015893517, 0.00000009772419, -0.00000001454001, -0.00000000927121, 0.00000000086700, 0.00000000057048, -0.00000000002525, -0.00000000001708])
        rec_hi = np.array([-0.00000000001708, 0.00000000002525, 0.00000000057048, -0.00000000086700, -0.00000000927121, 0.00000001454001, 0.00000009772419, -0.00000015893517, -0.00000075150216, 0.00000127545430, 0.00000449693644, -0.00000803150300, -0.00002180200077, 0.00004147478607, 0.00008754452092, -0.00018169287648, -0.00029777893220, 0.00068717164335, 0.00089677606308, -0.00223564942205, -0.00254400371025, 0.00615665954826, 0.00706582701104, -0.01411747007762, -0.01898524469525, 0.02665671054265, 0.04825237108568, -0.04118580667626, -0.12121116823150, 0.05186074316119, 0.44344254984153, -0.76011330201794, 0.43120981555509, 0.04371898336595, -0.11016997698347, -0.02882862175929, 0.03937203787798, 0.01497846208171, -0.01274237063272, -0.00599484919216, 0.00330082501062, 0.00178326000860, -0.00062356044746, -0.00037129499561, 0.00007547367838, 0.00004829631521, -0.00000436826482, -0.00000295433652])
    elif family=="coif9":
        dec_lo = np.array([-0.00000000000099, -0.00000000000142, 0.00000000003686, 0.00000000005417, -0.00000000067235, -0.00000000101363, 0.00000000797401, 0.00000001237526, -0.00000006916547, -0.00000011096670, 0.00000046795848, 0.00000078024803, -0.00000257238357, -0.00000448811148, 0.00001181440945, 0.00002177639641, -0.00004613708198, -0.00009135595509, 0.00015541352127, 0.00033691386928, -0.00046272908551, -0.00109574562795, 0.00126969092514, 0.00311322778838, -0.00335767452659, -0.00761404244826, 0.00870275744623, 0.01581871581593, -0.02175455351095, -0.02766123949868, 0.05184461568625, 0.04047376745573, -0.12434558953929, -0.04934886629363, 0.44445789317645, 0.75704552338438, 0.43302675110315, -0.04172611020585, -0.11388350819005, 0.02857266755695, 0.04318172760825, -0.01586022389479, -0.01537664962972, 0.00702234046020, 0.00459705642492, -0.00242124167365, -0.00107545827274, 0.00062647303214, 0.00018228485966, -0.00011443395279, -0.00001978720443, 0.00001315888565, 0.00000102932007, -0.00000071649204])
        dec_hi = np.array([0.00000071649204, 0.00000102932007, -0.00001315888565, -0.00001978720443, 0.00011443395279, 0.00018228485966, -0.00062647303214, -0.00107545827274, 0.00242124167365, 0.00459705642492, -0.00702234046020, -0.01537664962972, 0.01586022389479, 0.04318172760825, -0.02857266755695, -0.11388350819005, 0.04172611020585, 0.43302675110315, -0.75704552338438, 0.44445789317645, 0.04934886629363, -0.12434558953929, -0.04047376745573, 0.05184461568625, 0.02766123949868, -0.02175455351095, -0.01581871581593, 0.00870275744623, 0.00761404244826, -0.00335767452659, -0.00311322778838, 0.00126969092514, 0.00109574562795, -0.00046272908551, -0.00033691386928, 0.00015541352127, 0.00009135595509, -0.00004613708198, -0.00002177639641, 0.00001181440945, 0.00000448811148, -0.00000257238357, -0.00000078024803, 0.00000046795848, 0.00000011096670, -0.00000006916547, -0.00000001237526, 0.00000000797401, 0.00000000101363, -0.00000000067235, -0.00000000005417, 0.00000000003686, 0.00000000000142, -0.00000000000099])
        rec_lo = np.array([-0.00000071649204, 0.00000102932007, 0.00001315888565, -0.00001978720443, -0.00011443395279, 0.00018228485966, 0.00062647303214, -0.00107545827274, -0.00242124167365, 0.00459705642492, 0.00702234046020, -0.01537664962972, -0.01586022389479, 0.04318172760825, 0.02857266755695, -0.11388350819005, -0.04172611020585, 0.43302675110315, 0.75704552338438, 0.44445789317645, -0.04934886629363, -0.12434558953929, 0.04047376745573, 0.05184461568625, -0.02766123949868, -0.02175455351095, 0.01581871581593, 0.00870275744623, -0.00761404244826, -0.00335767452659, 0.00311322778838, 0.00126969092514, -0.00109574562795, -0.00046272908551, 0.00033691386928, 0.00015541352127, -0.00009135595509, -0.00004613708198, 0.00002177639641, 0.00001181440945, -0.00000448811148, -0.00000257238357, 0.00000078024803, 0.00000046795848, -0.00000011096670, -0.00000006916547, 0.00000001237526, 0.00000000797401, -0.00000000101363, -0.00000000067235, 0.00000000005417, 0.00000000003686, -0.00000000000142, -0.00000000000099])
        rec_hi = np.array([-0.00000000000099, 0.00000000000142, 0.00000000003686, -0.00000000005417, -0.00000000067235, 0.00000000101363, 0.00000000797401, -0.00000001237526, -0.00000006916547, 0.00000011096670, 0.00000046795848, -0.00000078024803, -0.00000257238357, 0.00000448811148, 0.00001181440945, -0.00002177639641, -0.00004613708198, 0.00009135595509, 0.00015541352127, -0.00033691386928, -0.00046272908551, 0.00109574562795, 0.00126969092514, -0.00311322778838, -0.00335767452659, 0.00761404244826, 0.00870275744623, -0.01581871581593, -0.02175455351095, 0.02766123949868, 0.05184461568625, -0.04047376745573, -0.12434558953929, 0.04934886629363, 0.44445789317645, -0.75704552338438, 0.43302675110315, 0.04172611020585, -0.11388350819005, -0.02857266755695, 0.04318172760825, 0.01586022389479, -0.01537664962972, -0.00702234046020, 0.00459705642492, 0.00242124167365, -0.00107545827274, -0.00062647303214, 0.00018228485966, 0.00011443395279, -0.00001978720443, -0.00001315888565, 0.00000102932007, 0.00000071649204])
    elif family=="coif10":
        dec_lo = np.array([-0.00000000000006, -0.00000000000008, 0.00000000000237, 0.00000000000339, -0.00000000004804, -0.00000000007013, 0.00000000063331, 0.00000000094678, -0.00000000611891, -0.00000000939633, 0.00000004620903, 0.00000007315543, -0.00000028409076, -0.00000046576244, 0.00000146244503, 0.00000249720279, -0.00000643432949, -0.00001153105839, 0.00002454191012, 0.00004672498135, -0.00008202162256, -0.00016857983433, 0.00024363173072, 0.00054516737086, -0.00065986625324, -0.00157485822319, 0.00168996979264, 0.00402022279070, -0.00421811388481, -0.00895320728654, 0.01030537800245, 0.01720591249832, -0.02426732868280, -0.02831006394443, 0.05490896399592, 0.03966834927954, -0.12690910430555, -0.04714526253802, 0.44526919771961, 0.75445010949478, 0.43448818216271, -0.03998711301587, -0.11693607050207, 0.02823291273880, 0.04646274705470, -0.01652151126871, -0.01782044578129, 0.00791715706771, 0.00593737326590, -0.00305399249381, -0.00162077810885, 0.00092493996042, 0.00034345502618, -0.00021177413649, -0.00005264472186, 0.00003445969323, 0.00000517396261, -0.00000355120554, -0.00000024427649, 0.00000017423675])
        dec_hi = np.array([-0.00000017423675, -0.00000024427649, 0.00000355120554, 0.00000517396261, -0.00003445969323, -0.00005264472186, 0.00021177413649, 0.00034345502618, -0.00092493996042, -0.00162077810885, 0.00305399249381, 0.00593737326590, -0.00791715706771, -0.01782044578129, 0.01652151126871, 0.04646274705470, -0.02823291273880, -0.11693607050207, 0.03998711301587, 0.43448818216271, -0.75445010949478, 0.44526919771961, 0.04714526253802, -0.12690910430555, -0.03966834927954, 0.05490896399592, 0.02831006394443, -0.02426732868280, -0.01720591249832, 0.01030537800245, 0.00895320728654, -0.00421811388481, -0.00402022279070, 0.00168996979264, 0.00157485822319, -0.00065986625324, -0.00054516737086, 0.00024363173072, 0.00016857983433, -0.00008202162256, -0.00004672498135, 0.00002454191012, 0.00001153105839, -0.00000643432949, -0.00000249720279, 0.00000146244503, 0.00000046576244, -0.00000028409076, -0.00000007315543, 0.00000004620903, 0.00000000939633, -0.00000000611891, -0.00000000094678, 0.00000000063331, 0.00000000007013, -0.00000000004804, -0.00000000000339, 0.00000000000237, 0.00000000000008, -0.00000000000006])
        rec_lo = np.array([0.00000017423675, -0.00000024427649, -0.00000355120554, 0.00000517396261, 0.00003445969323, -0.00005264472186, -0.00021177413649, 0.00034345502618, 0.00092493996042, -0.00162077810885, -0.00305399249381, 0.00593737326590, 0.00791715706771, -0.01782044578129, -0.01652151126871, 0.04646274705470, 0.02823291273880, -0.11693607050207, -0.03998711301587, 0.43448818216271, 0.75445010949478, 0.44526919771961, -0.04714526253802, -0.12690910430555, 0.03966834927954, 0.05490896399592, -0.02831006394443, -0.02426732868280, 0.01720591249832, 0.01030537800245, -0.00895320728654, -0.00421811388481, 0.00402022279070, 0.00168996979264, -0.00157485822319, -0.00065986625324, 0.00054516737086, 0.00024363173072, -0.00016857983433, -0.00008202162256, 0.00004672498135, 0.00002454191012, -0.00001153105839, -0.00000643432949, 0.00000249720279, 0.00000146244503, -0.00000046576244, -0.00000028409076, 0.00000007315543, 0.00000004620903, -0.00000000939633, -0.00000000611891, 0.00000000094678, 0.00000000063331, -0.00000000007013, -0.00000000004804, 0.00000000000339, 0.00000000000237, -0.00000000000008, -0.00000000000006])
        rec_hi = np.array([-0.00000000000006, 0.00000000000008, 0.00000000000237, -0.00000000000339, -0.00000000004804, 0.00000000007013, 0.00000000063331, -0.00000000094678, -0.00000000611891, 0.00000000939633, 0.00000004620903, -0.00000007315543, -0.00000028409076, 0.00000046576244, 0.00000146244503, -0.00000249720279, -0.00000643432949, 0.00001153105839, 0.00002454191012, -0.00004672498135, -0.00008202162256, 0.00016857983433, 0.00024363173072, -0.00054516737086, -0.00065986625324, 0.00157485822319, 0.00168996979264, -0.00402022279070, -0.00421811388481, 0.00895320728654, 0.01030537800245, -0.01720591249832, -0.02426732868280, 0.02831006394443, 0.05490896399592, -0.03966834927954, -0.12690910430555, 0.04714526253802, 0.44526919771961, -0.75445010949478, 0.43448818216271, 0.03998711301587, -0.11693607050207, -0.02823291273880, 0.04646274705470, 0.01652151126871, -0.01782044578129, -0.00791715706771, 0.00593737326590, 0.00305399249381, -0.00162077810885, -0.00092493996042, 0.00034345502618, 0.00021177413649, -0.00005264472186, -0.00003445969323, 0.00000517396261, 0.00000355120554, -0.00000024427649, -0.00000017423675])
    elif family=="coif11":
        dec_lo = np.array([-0.00000000000000, -0.00000000000000, 0.00000000000015, 0.00000000000021, -0.00000000000339, -0.00000000000482, 0.00000000004920, 0.00000000007132, -0.00000000052408, -0.00000000077682, 0.00000000437151, 0.00000000664300, -0.00000002973714, -0.00000004647552, 0.00000016966517, 0.00000027376934, -0.00000082879674, -0.00000138737548, 0.00000352114598, 0.00000615316487, -0.00001316375093, -0.00002423559202, 0.00004368364304, 0.00008582415436, -0.00012986682089, -0.00027542687785, 0.00035122675042, 0.00080174521034, -0.00088617751806, -0.00210361007647, 0.00214902071763, 0.00492267326110, -0.00510511284128, -0.01016088486651, 0.01185197442751, 0.01832964333662, -0.02654507086218, -0.02870259601973, 0.05755024470407, 0.03882560372412, -0.12904373222526, -0.04519495629358, 0.44593144282059, 0.75221754404024, 0.43568960835151, -0.03845246068584, -0.11948939622817, 0.02784705124132, 0.04931253893772, -0.01701592242448, -0.02007325197616, 0.00869153830286, 0.00728301715594, -0.00366386352916, -0.00223684251959, 0.00125207303989, 0.00055772048695, -0.00033810804201, -0.00010766668138, 0.00006942759575, 0.00001503949885, -0.00001018591923, -0.00000134935218, 0.00000095105732, 0.00000005829773, -0.00000004246659])
        dec_hi = np.array([0.00000004246659, 0.00000005829773, -0.00000095105732, -0.00000134935218, 0.00001018591923, 0.00001503949885, -0.00006942759575, -0.00010766668138, 0.00033810804201, 0.00055772048695, -0.00125207303989, -0.00223684251959, 0.00366386352916, 0.00728301715594, -0.00869153830286, -0.02007325197616, 0.01701592242448, 0.04931253893772, -0.02784705124132, -0.11948939622817, 0.03845246068584, 0.43568960835151, -0.75221754404024, 0.44593144282059, 0.04519495629358, -0.12904373222526, -0.03882560372412, 0.05755024470407, 0.02870259601973, -0.02654507086218, -0.01832964333662, 0.01185197442751, 0.01016088486651, -0.00510511284128, -0.00492267326110, 0.00214902071763, 0.00210361007647, -0.00088617751806, -0.00080174521034, 0.00035122675042, 0.00027542687785, -0.00012986682089, -0.00008582415436, 0.00004368364304, 0.00002423559202, -0.00001316375093, -0.00000615316487, 0.00000352114598, 0.00000138737548, -0.00000082879674, -0.00000027376934, 0.00000016966517, 0.00000004647552, -0.00000002973714, -0.00000000664300, 0.00000000437151, 0.00000000077682, -0.00000000052408, -0.00000000007132, 0.00000000004920, 0.00000000000482, -0.00000000000339, -0.00000000000021, 0.00000000000015, 0.00000000000000, -0.00000000000000])
        rec_lo = np.array([-0.00000004246659, 0.00000005829773, 0.00000095105732, -0.00000134935218, -0.00001018591923, 0.00001503949885, 0.00006942759575, -0.00010766668138, -0.00033810804201, 0.00055772048695, 0.00125207303989, -0.00223684251959, -0.00366386352916, 0.00728301715594, 0.00869153830286, -0.02007325197616, -0.01701592242448, 0.04931253893772, 0.02784705124132, -0.11948939622817, -0.03845246068584, 0.43568960835151, 0.75221754404024, 0.44593144282059, -0.04519495629358, -0.12904373222526, 0.03882560372412, 0.05755024470407, -0.02870259601973, -0.02654507086218, 0.01832964333662, 0.01185197442751, -0.01016088486651, -0.00510511284128, 0.00492267326110, 0.00214902071763, -0.00210361007647, -0.00088617751806, 0.00080174521034, 0.00035122675042, -0.00027542687785, -0.00012986682089, 0.00008582415436, 0.00004368364304, -0.00002423559202, -0.00001316375093, 0.00000615316487, 0.00000352114598, -0.00000138737548, -0.00000082879674, 0.00000027376934, 0.00000016966517, -0.00000004647552, -0.00000002973714, 0.00000000664300, 0.00000000437151, -0.00000000077682, -0.00000000052408, 0.00000000007132, 0.00000000004920, -0.00000000000482, -0.00000000000339, 0.00000000000021, 0.00000000000015, -0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, 0.00000000000000, 0.00000000000015, -0.00000000000021, -0.00000000000339, 0.00000000000482, 0.00000000004920, -0.00000000007132, -0.00000000052408, 0.00000000077682, 0.00000000437151, -0.00000000664300, -0.00000002973714, 0.00000004647552, 0.00000016966517, -0.00000027376934, -0.00000082879674, 0.00000138737548, 0.00000352114598, -0.00000615316487, -0.00001316375093, 0.00002423559202, 0.00004368364304, -0.00008582415436, -0.00012986682089, 0.00027542687785, 0.00035122675042, -0.00080174521034, -0.00088617751806, 0.00210361007647, 0.00214902071763, -0.00492267326110, -0.00510511284128, 0.01016088486651, 0.01185197442751, -0.01832964333662, -0.02654507086218, 0.02870259601973, 0.05755024470407, -0.03882560372412, -0.12904373222526, 0.04519495629358, 0.44593144282059, -0.75221754404024, 0.43568960835151, 0.03845246068584, -0.11948939622817, -0.02784705124132, 0.04931253893772, 0.01701592242448, -0.02007325197616, -0.00869153830286, 0.00728301715594, 0.00366386352916, -0.00223684251959, -0.00125207303989, 0.00055772048695, 0.00033810804201, -0.00010766668138, -0.00006942759575, 0.00001503949885, 0.00001018591923, -0.00000134935218, -0.00000095105732, 0.00000005829773, 0.00000004246659])
    elif family=="coif12":
        dec_lo = np.array([-0.00000000000000, -0.00000000000000, 0.00000000000001, 0.00000000000001, -0.00000000000024, -0.00000000000033, 0.00000000000375, 0.00000000000530, -0.00000000004369, -0.00000000006289, 0.00000000039906, 0.00000000058647, -0.00000000297691, -0.00000000447752, 0.00000001865414, 0.00000002879532, -0.00000010022581, -0.00000015931231, 0.00000046902731, 0.00000077084373, -0.00000193560105, -0.00000330623921, 0.00000711297735, 0.00001271854495, -0.00002344975070, -0.00004433611443, 0.00006982465534, 0.00014116860337, -0.00018957894276, -0.00041198987136, 0.00047699273329, 0.00109944423886, -0.00113848525514, -0.00266309869589, 0.00263853207607, 0.00579785317180, -0.00600279370881, -0.01123692841484, 0.01333075543551, 0.01923522882934, -0.02861160212352, -0.02890894744141, 0.05984819686064, 0.03797767256571, -0.13084806656616, -0.04345493993504, 0.44648152288957, 0.75027077765067, 0.43669517449051, -0.03708505815412, -0.12165660970269, 0.02743798196811, 0.05180772024805, -0.01738271453899, -0.02214438902479, 0.00935963057432, 0.00860864981494, -0.00424102101492, -0.00290334962500, 0.00159569952227, 0.00082048776033, -0.00048946907653, -0.00018759057842, 0.00011925526521, 0.00003319558671, -0.00002219929936, -0.00000425536511, 0.00000296541195, 0.00000035104114, -0.00000025311180, -0.00000001397637, 0.00000001037010])
        dec_hi = np.array([-0.00000001037010, -0.00000001397637, 0.00000025311180, 0.00000035104114, -0.00000296541195, -0.00000425536511, 0.00002219929936, 0.00003319558671, -0.00011925526521, -0.00018759057842, 0.00048946907653, 0.00082048776033, -0.00159569952227, -0.00290334962500, 0.00424102101492, 0.00860864981494, -0.00935963057432, -0.02214438902479, 0.01738271453899, 0.05180772024805, -0.02743798196811, -0.12165660970269, 0.03708505815412, 0.43669517449051, -0.75027077765067, 0.44648152288957, 0.04345493993504, -0.13084806656616, -0.03797767256571, 0.05984819686064, 0.02890894744141, -0.02861160212352, -0.01923522882934, 0.01333075543551, 0.01123692841484, -0.00600279370881, -0.00579785317180, 0.00263853207607, 0.00266309869589, -0.00113848525514, -0.00109944423886, 0.00047699273329, 0.00041198987136, -0.00018957894276, -0.00014116860337, 0.00006982465534, 0.00004433611443, -0.00002344975070, -0.00001271854495, 0.00000711297735, 0.00000330623921, -0.00000193560105, -0.00000077084373, 0.00000046902731, 0.00000015931231, -0.00000010022581, -0.00000002879532, 0.00000001865414, 0.00000000447752, -0.00000000297691, -0.00000000058647, 0.00000000039906, 0.00000000006289, -0.00000000004369, -0.00000000000530, 0.00000000000375, 0.00000000000033, -0.00000000000024, -0.00000000000001, 0.00000000000001, 0.00000000000000, -0.00000000000000])
        rec_lo = np.array([0.00000001037010, -0.00000001397637, -0.00000025311180, 0.00000035104114, 0.00000296541195, -0.00000425536511, -0.00002219929936, 0.00003319558671, 0.00011925526521, -0.00018759057842, -0.00048946907653, 0.00082048776033, 0.00159569952227, -0.00290334962500, -0.00424102101492, 0.00860864981494, 0.00935963057432, -0.02214438902479, -0.01738271453899, 0.05180772024805, 0.02743798196811, -0.12165660970269, -0.03708505815412, 0.43669517449051, 0.75027077765067, 0.44648152288957, -0.04345493993504, -0.13084806656616, 0.03797767256571, 0.05984819686064, -0.02890894744141, -0.02861160212352, 0.01923522882934, 0.01333075543551, -0.01123692841484, -0.00600279370881, 0.00579785317180, 0.00263853207607, -0.00266309869589, -0.00113848525514, 0.00109944423886, 0.00047699273329, -0.00041198987136, -0.00018957894276, 0.00014116860337, 0.00006982465534, -0.00004433611443, -0.00002344975070, 0.00001271854495, 0.00000711297735, -0.00000330623921, -0.00000193560105, 0.00000077084373, 0.00000046902731, -0.00000015931231, -0.00000010022581, 0.00000002879532, 0.00000001865414, -0.00000000447752, -0.00000000297691, 0.00000000058647, 0.00000000039906, -0.00000000006289, -0.00000000004369, 0.00000000000530, 0.00000000000375, -0.00000000000033, -0.00000000000024, 0.00000000000001, 0.00000000000001, -0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, 0.00000000000000, 0.00000000000001, -0.00000000000001, -0.00000000000024, 0.00000000000033, 0.00000000000375, -0.00000000000530, -0.00000000004369, 0.00000000006289, 0.00000000039906, -0.00000000058647, -0.00000000297691, 0.00000000447752, 0.00000001865414, -0.00000002879532, -0.00000010022581, 0.00000015931231, 0.00000046902731, -0.00000077084373, -0.00000193560105, 0.00000330623921, 0.00000711297735, -0.00001271854495, -0.00002344975070, 0.00004433611443, 0.00006982465534, -0.00014116860337, -0.00018957894276, 0.00041198987136, 0.00047699273329, -0.00109944423886, -0.00113848525514, 0.00266309869589, 0.00263853207607, -0.00579785317180, -0.00600279370881, 0.01123692841484, 0.01333075543551, -0.01923522882934, -0.02861160212352, 0.02890894744141, 0.05984819686064, -0.03797767256571, -0.13084806656616, 0.04345493993504, 0.44648152288957, -0.75027077765067, 0.43669517449051, 0.03708505815412, -0.12165660970269, -0.02743798196811, 0.05180772024805, 0.01738271453899, -0.02214438902479, -0.00935963057432, 0.00860864981494, 0.00424102101492, -0.00290334962500, -0.00159569952227, 0.00082048776033, 0.00048946907653, -0.00018759057842, -0.00011925526521, 0.00003319558671, 0.00002219929936, -0.00000425536511, -0.00000296541195, 0.00000035104114, 0.00000025311180, -0.00000001397637, -0.00000001037010])
    elif family=="coif13":
        dec_lo = np.array([-0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000002, -0.00000000000002, 0.00000000000028, 0.00000000000039, -0.00000000000356, -0.00000000000500, 0.00000000003535, 0.00000000005051, -0.00000000028699, -0.00000000041821, 0.00000000195996, 0.00000000291864, -0.00000001149213, -0.00000001753055, 0.00000005876532, 0.00000009209611, -0.00000026532815, -0.00000042870074, 0.00000106842762, 0.00000178756063, -0.00000386853334, -0.00000674079922, 0.00001267653882, 0.00002318774251, -0.00003779966210, -0.00007329758910, 0.00010322189107, 0.00021391688900, -0.00026078674318, -0.00057668812596, 0.00061961533962, 0.00143010518686, -0.00141352653690, -0.00323755941382, 0.00315087763819, 0.00663157452845, -0.00689911532291, -0.01218803467103, 0.01473631763164, 0.01996119614595, -0.03049009579423, -0.02897913316299, 0.06186422181011, 0.03714297920871, -0.13239268372863, -0.04189131400120, 0.44694511891208, 0.74855383245819, 0.43754954331414, -0.03585654727086, -0.12351915633983, 0.02701986705945, 0.05400870551306, -0.01765094147449, -0.02404734005036, 0.00993524324843, 0.00989833068184, -0.00478067594933, -0.00360331127877, 0.00194629541239, 0.00112553391308, -0.00066123808672, -0.00029338106940, 0.00018399068205, 0.00006181045384, -0.00004083713797, -0.00001008560197, 0.00000695230175, 0.00000119388808, -0.00000085244059, -0.00000009111838, 0.00000006701079, 0.00000000336330, -0.00000000253646])
        dec_hi = np.array([0.00000000253646, 0.00000000336330, -0.00000006701079, -0.00000009111838, 0.00000085244059, 0.00000119388808, -0.00000695230175, -0.00001008560197, 0.00004083713797, 0.00006181045384, -0.00018399068205, -0.00029338106940, 0.00066123808672, 0.00112553391308, -0.00194629541239, -0.00360331127877, 0.00478067594933, 0.00989833068184, -0.00993524324843, -0.02404734005036, 0.01765094147449, 0.05400870551306, -0.02701986705945, -0.12351915633983, 0.03585654727086, 0.43754954331414, -0.74855383245819, 0.44694511891208, 0.04189131400120, -0.13239268372863, -0.03714297920871, 0.06186422181011, 0.02897913316299, -0.03049009579423, -0.01996119614595, 0.01473631763164, 0.01218803467103, -0.00689911532291, -0.00663157452845, 0.00315087763819, 0.00323755941382, -0.00141352653690, -0.00143010518686, 0.00061961533962, 0.00057668812596, -0.00026078674318, -0.00021391688900, 0.00010322189107, 0.00007329758910, -0.00003779966210, -0.00002318774251, 0.00001267653882, 0.00000674079922, -0.00000386853334, -0.00000178756063, 0.00000106842762, 0.00000042870074, -0.00000026532815, -0.00000009209611, 0.00000005876532, 0.00000001753055, -0.00000001149213, -0.00000000291864, 0.00000000195996, 0.00000000041821, -0.00000000028699, -0.00000000005051, 0.00000000003535, 0.00000000000500, -0.00000000000356, -0.00000000000039, 0.00000000000028, 0.00000000000002, -0.00000000000002, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000])
        rec_lo = np.array([-0.00000000253646, 0.00000000336330, 0.00000006701079, -0.00000009111838, -0.00000085244059, 0.00000119388808, 0.00000695230175, -0.00001008560197, -0.00004083713797, 0.00006181045384, 0.00018399068205, -0.00029338106940, -0.00066123808672, 0.00112553391308, 0.00194629541239, -0.00360331127877, -0.00478067594933, 0.00989833068184, 0.00993524324843, -0.02404734005036, -0.01765094147449, 0.05400870551306, 0.02701986705945, -0.12351915633983, -0.03585654727086, 0.43754954331414, 0.74855383245819, 0.44694511891208, -0.04189131400120, -0.13239268372863, 0.03714297920871, 0.06186422181011, -0.02897913316299, -0.03049009579423, 0.01996119614595, 0.01473631763164, -0.01218803467103, -0.00689911532291, 0.00663157452845, 0.00315087763819, -0.00323755941382, -0.00141352653690, 0.00143010518686, 0.00061961533962, -0.00057668812596, -0.00026078674318, 0.00021391688900, 0.00010322189107, -0.00007329758910, -0.00003779966210, 0.00002318774251, 0.00001267653882, -0.00000674079922, -0.00000386853334, 0.00000178756063, 0.00000106842762, -0.00000042870074, -0.00000026532815, 0.00000009209611, 0.00000005876532, -0.00000001753055, -0.00000001149213, 0.00000000291864, 0.00000000195996, -0.00000000041821, -0.00000000028699, 0.00000000005051, 0.00000000003535, -0.00000000000500, -0.00000000000356, 0.00000000000039, 0.00000000000028, -0.00000000000002, -0.00000000000002, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000002, 0.00000000000002, 0.00000000000028, -0.00000000000039, -0.00000000000356, 0.00000000000500, 0.00000000003535, -0.00000000005051, -0.00000000028699, 0.00000000041821, 0.00000000195996, -0.00000000291864, -0.00000001149213, 0.00000001753055, 0.00000005876532, -0.00000009209611, -0.00000026532815, 0.00000042870074, 0.00000106842762, -0.00000178756063, -0.00000386853334, 0.00000674079922, 0.00001267653882, -0.00002318774251, -0.00003779966210, 0.00007329758910, 0.00010322189107, -0.00021391688900, -0.00026078674318, 0.00057668812596, 0.00061961533962, -0.00143010518686, -0.00141352653690, 0.00323755941382, 0.00315087763819, -0.00663157452845, -0.00689911532291, 0.01218803467103, 0.01473631763164, -0.01996119614595, -0.03049009579423, 0.02897913316299, 0.06186422181011, -0.03714297920871, -0.13239268372863, 0.04189131400120, 0.44694511891208, -0.74855383245819, 0.43754954331414, 0.03585654727086, -0.12351915633983, -0.02701986705945, 0.05400870551306, 0.01765094147449, -0.02404734005036, -0.00993524324843, 0.00989833068184, 0.00478067594933, -0.00360331127877, -0.00194629541239, 0.00112553391308, 0.00066123808672, -0.00029338106940, -0.00018399068205, 0.00006181045384, 0.00004083713797, -0.00001008560197, -0.00000695230175, 0.00000119388808, 0.00000085244059, -0.00000009111838, -0.00000006701079, 0.00000000336330, 0.00000000253646])
    elif family=="coif14":
        dec_lo = np.array([-0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000002, 0.00000000000003, -0.00000000000028, -0.00000000000039, 0.00000000000305, 0.00000000000426, -0.00000000002679, -0.00000000003801, 0.00000000019809, 0.00000000028628, -0.00000000125908, -0.00000000185687, 0.00000000698739, 0.00000001053864, -0.00000003427757, -0.00000005300651, 0.00000015013414, 0.00000023876494, -0.00000059199746, -0.00000097184235, 0.00000211607576, 0.00000360289360, -0.00000689570940, -0.00001225451573, 0.00002058327894, 0.00003849201590, -0.00005654418404, -0.00011221961944, 0.00014390413588, 0.00030435990061, -0.00034297286151, -0.00076682514885, 0.00077772015661, 0.00178561819890, -0.00170805386611, -0.00381452407658, 0.00367932350198, 0.00741577823077, -0.00778520241707, -0.01302408800539, 0.01606728392607, 0.02053949987453, -0.03220185256406, -0.02894925613195, 0.06364620314135, 0.03633187623818, -0.13372942834322, -0.04047709315229, 0.44734065675374, 0.74702493796151, 0.43828475995016, -0.03474484970545, -0.12513713600885, 0.02660154627199, 0.05596345341138, -0.01784235415244, -0.02579684377083, 0.01043094917065, 0.01114255210951, -0.00528120283631, -0.00432306228077, 0.00229671072646, 0.00146612145699, -0.00084880838479, -0.00042463732870, 0.00026286235206, 0.00010248333220, -0.00006692604994, -0.00001999557640, 0.00001364035768, 0.00000302478082, -0.00000213946393, -0.00000033246584, 0.00000024242798, 0.00000002360236, -0.00000001766250, -0.00000000081190, 0.00000000062128])
        dec_hi = np.array([-0.00000000062128, -0.00000000081190, 0.00000001766250, 0.00000002360236, -0.00000024242798, -0.00000033246584, 0.00000213946393, 0.00000302478082, -0.00001364035768, -0.00001999557640, 0.00006692604994, 0.00010248333220, -0.00026286235206, -0.00042463732870, 0.00084880838479, 0.00146612145699, -0.00229671072646, -0.00432306228077, 0.00528120283631, 0.01114255210951, -0.01043094917065, -0.02579684377083, 0.01784235415244, 0.05596345341138, -0.02660154627199, -0.12513713600885, 0.03474484970545, 0.43828475995016, -0.74702493796151, 0.44734065675374, 0.04047709315229, -0.13372942834322, -0.03633187623818, 0.06364620314135, 0.02894925613195, -0.03220185256406, -0.02053949987453, 0.01606728392607, 0.01302408800539, -0.00778520241707, -0.00741577823077, 0.00367932350198, 0.00381452407658, -0.00170805386611, -0.00178561819890, 0.00077772015661, 0.00076682514885, -0.00034297286151, -0.00030435990061, 0.00014390413588, 0.00011221961944, -0.00005654418404, -0.00003849201590, 0.00002058327894, 0.00001225451573, -0.00000689570940, -0.00000360289360, 0.00000211607576, 0.00000097184235, -0.00000059199746, -0.00000023876494, 0.00000015013414, 0.00000005300651, -0.00000003427757, -0.00000001053864, 0.00000000698739, 0.00000000185687, -0.00000000125908, -0.00000000028628, 0.00000000019809, 0.00000000003801, -0.00000000002679, -0.00000000000426, 0.00000000000305, 0.00000000000039, -0.00000000000028, -0.00000000000003, 0.00000000000002, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000])
        rec_lo = np.array([0.00000000062128, -0.00000000081190, -0.00000001766250, 0.00000002360236, 0.00000024242798, -0.00000033246584, -0.00000213946393, 0.00000302478082, 0.00001364035768, -0.00001999557640, -0.00006692604994, 0.00010248333220, 0.00026286235206, -0.00042463732870, -0.00084880838479, 0.00146612145699, 0.00229671072646, -0.00432306228077, -0.00528120283631, 0.01114255210951, 0.01043094917065, -0.02579684377083, -0.01784235415244, 0.05596345341138, 0.02660154627199, -0.12513713600885, -0.03474484970545, 0.43828475995016, 0.74702493796151, 0.44734065675374, -0.04047709315229, -0.13372942834322, 0.03633187623818, 0.06364620314135, -0.02894925613195, -0.03220185256406, 0.02053949987453, 0.01606728392607, -0.01302408800539, -0.00778520241707, 0.00741577823077, 0.00367932350198, -0.00381452407658, -0.00170805386611, 0.00178561819890, 0.00077772015661, -0.00076682514885, -0.00034297286151, 0.00030435990061, 0.00014390413588, -0.00011221961944, -0.00005654418404, 0.00003849201590, 0.00002058327894, -0.00001225451573, -0.00000689570940, 0.00000360289360, 0.00000211607576, -0.00000097184235, -0.00000059199746, 0.00000023876494, 0.00000015013414, -0.00000005300651, -0.00000003427757, 0.00000001053864, 0.00000000698739, -0.00000000185687, -0.00000000125908, 0.00000000028628, 0.00000000019809, -0.00000000003801, -0.00000000002679, 0.00000000000426, 0.00000000000305, -0.00000000000039, -0.00000000000028, 0.00000000000003, 0.00000000000002, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000002, -0.00000000000003, -0.00000000000028, 0.00000000000039, 0.00000000000305, -0.00000000000426, -0.00000000002679, 0.00000000003801, 0.00000000019809, -0.00000000028628, -0.00000000125908, 0.00000000185687, 0.00000000698739, -0.00000001053864, -0.00000003427757, 0.00000005300651, 0.00000015013414, -0.00000023876494, -0.00000059199746, 0.00000097184235, 0.00000211607576, -0.00000360289360, -0.00000689570940, 0.00001225451573, 0.00002058327894, -0.00003849201590, -0.00005654418404, 0.00011221961944, 0.00014390413588, -0.00030435990061, -0.00034297286151, 0.00076682514885, 0.00077772015661, -0.00178561819890, -0.00170805386611, 0.00381452407658, 0.00367932350198, -0.00741577823077, -0.00778520241707, 0.01302408800539, 0.01606728392607, -0.02053949987453, -0.03220185256406, 0.02894925613195, 0.06364620314135, -0.03633187623818, -0.13372942834322, 0.04047709315229, 0.44734065675374, -0.74702493796151, 0.43828475995016, 0.03474484970545, -0.12513713600885, -0.02660154627199, 0.05596345341138, 0.01784235415244, -0.02579684377083, -0.01043094917065, 0.01114255210951, 0.00528120283631, -0.00432306228077, -0.00229671072646, 0.00146612145699, 0.00084880838479, -0.00042463732870, -0.00026286235206, 0.00010248333220, 0.00006692604994, -0.00001999557640, -0.00001364035768, 0.00000302478082, 0.00000213946393, -0.00000033246584, -0.00000024242798, 0.00000002360236, 0.00000001766250, -0.00000000081190, -0.00000000062128])
    elif family=="coif15":
        dec_lo = np.array([-0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000002, -0.00000000000003, 0.00000000000026, 0.00000000000035, -0.00000000000243, -0.00000000000337, 0.00000000001936, 0.00000000002728, -0.00000000013262, -0.00000000019012, 0.00000000079407, 0.00000000116018, -0.00000000420710, -0.00000000627695, 0.00000001992151, 0.00000003042031, -0.00000008500698, -0.00000013320727, 0.00000032915263, 0.00000053099583, -0.00000116336853, -0.00000193980039, 0.00000377209106, 0.00000653441559, -0.00001126710177, -0.00002041399072, 0.00003112248367, 0.00005943677222, -0.00007986582334, -0.00016180532955, 0.00019175221056, 0.00041206868139, -0.00043554294379, -0.00097907305461, 0.00094989997549, 0.00215849846653, -0.00201892979680, -0.00438452413454, 0.00421808950096, 0.00814664268259, -0.00865469059482, -0.01375609348897, 0.01732475364848, 0.02099633854680, -0.03376590813620, -0.02884567344425, 0.06523194483317, 0.03554988160863, -0.13489723657820, -0.03919056961250, 0.44768169876169, 0.74565221337108, 0.43892440853136, -0.03373249605098, -0.12655584892864, 0.02618853484172, 0.05771035117515, -0.01797342382926, -0.02740743945378, 0.01085781218464, 0.01233614757165, -0.00574290284726, -0.00505189364251, 0.00264175381472, 0.00183566044669, -0.00104793891684, -0.00058001104415, 0.00035458138561, 0.00015621764400, -0.00010088156289, -0.00003505374322, 0.00002367321573, 0.00000636368402, -0.00000446042318, -0.00000089682120, 0.00000064858694, 0.00000009197257, -0.00000006831332, -0.00000000610221, 0.00000000463776, 0.00000000019652, -0.00000000015237])
        dec_hi = np.array([0.00000000015237, 0.00000000019652, -0.00000000463776, -0.00000000610221, 0.00000006831332, 0.00000009197257, -0.00000064858694, -0.00000089682120, 0.00000446042318, 0.00000636368402, -0.00002367321573, -0.00003505374322, 0.00010088156289, 0.00015621764400, -0.00035458138561, -0.00058001104415, 0.00104793891684, 0.00183566044669, -0.00264175381472, -0.00505189364251, 0.00574290284726, 0.01233614757165, -0.01085781218464, -0.02740743945378, 0.01797342382926, 0.05771035117515, -0.02618853484172, -0.12655584892864, 0.03373249605098, 0.43892440853136, -0.74565221337108, 0.44768169876169, 0.03919056961250, -0.13489723657820, -0.03554988160863, 0.06523194483317, 0.02884567344425, -0.03376590813620, -0.02099633854680, 0.01732475364848, 0.01375609348897, -0.00865469059482, -0.00814664268259, 0.00421808950096, 0.00438452413454, -0.00201892979680, -0.00215849846653, 0.00094989997549, 0.00097907305461, -0.00043554294379, -0.00041206868139, 0.00019175221056, 0.00016180532955, -0.00007986582334, -0.00005943677222, 0.00003112248367, 0.00002041399072, -0.00001126710177, -0.00000653441559, 0.00000377209106, 0.00000193980039, -0.00000116336853, -0.00000053099583, 0.00000032915263, 0.00000013320727, -0.00000008500698, -0.00000003042031, 0.00000001992151, 0.00000000627695, -0.00000000420710, -0.00000000116018, 0.00000000079407, 0.00000000019012, -0.00000000013262, -0.00000000002728, 0.00000000001936, 0.00000000000337, -0.00000000000243, -0.00000000000035, 0.00000000000026, 0.00000000000003, -0.00000000000002, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000])
        rec_lo = np.array([-0.00000000015237, 0.00000000019652, 0.00000000463776, -0.00000000610221, -0.00000006831332, 0.00000009197257, 0.00000064858694, -0.00000089682120, -0.00000446042318, 0.00000636368402, 0.00002367321573, -0.00003505374322, -0.00010088156289, 0.00015621764400, 0.00035458138561, -0.00058001104415, -0.00104793891684, 0.00183566044669, 0.00264175381472, -0.00505189364251, -0.00574290284726, 0.01233614757165, 0.01085781218464, -0.02740743945378, -0.01797342382926, 0.05771035117515, 0.02618853484172, -0.12655584892864, -0.03373249605098, 0.43892440853136, 0.74565221337108, 0.44768169876169, -0.03919056961250, -0.13489723657820, 0.03554988160863, 0.06523194483317, -0.02884567344425, -0.03376590813620, 0.02099633854680, 0.01732475364848, -0.01375609348897, -0.00865469059482, 0.00814664268259, 0.00421808950096, -0.00438452413454, -0.00201892979680, 0.00215849846653, 0.00094989997549, -0.00097907305461, -0.00043554294379, 0.00041206868139, 0.00019175221056, -0.00016180532955, -0.00007986582334, 0.00005943677222, 0.00003112248367, -0.00002041399072, -0.00001126710177, 0.00000653441559, 0.00000377209106, -0.00000193980039, -0.00000116336853, 0.00000053099583, 0.00000032915263, -0.00000013320727, -0.00000008500698, 0.00000003042031, 0.00000001992151, -0.00000000627695, -0.00000000420710, 0.00000000116018, 0.00000000079407, -0.00000000019012, -0.00000000013262, 0.00000000002728, 0.00000000001936, -0.00000000000337, -0.00000000000243, 0.00000000000035, 0.00000000000026, -0.00000000000003, -0.00000000000002, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000002, 0.00000000000003, 0.00000000000026, -0.00000000000035, -0.00000000000243, 0.00000000000337, 0.00000000001936, -0.00000000002728, -0.00000000013262, 0.00000000019012, 0.00000000079407, -0.00000000116018, -0.00000000420710, 0.00000000627695, 0.00000001992151, -0.00000003042031, -0.00000008500698, 0.00000013320727, 0.00000032915263, -0.00000053099583, -0.00000116336853, 0.00000193980039, 0.00000377209106, -0.00000653441559, -0.00001126710177, 0.00002041399072, 0.00003112248367, -0.00005943677222, -0.00007986582334, 0.00016180532955, 0.00019175221056, -0.00041206868139, -0.00043554294379, 0.00097907305461, 0.00094989997549, -0.00215849846653, -0.00201892979680, 0.00438452413454, 0.00421808950096, -0.00814664268259, -0.00865469059482, 0.01375609348897, 0.01732475364848, -0.02099633854680, -0.03376590813620, 0.02884567344425, 0.06523194483317, -0.03554988160863, -0.13489723657820, 0.03919056961250, 0.44768169876169, -0.74565221337108, 0.43892440853136, 0.03373249605098, -0.12655584892864, -0.02618853484172, 0.05771035117515, 0.01797342382926, -0.02740743945378, -0.01085781218464, 0.01233614757165, 0.00574290284726, -0.00505189364251, -0.00264175381472, 0.00183566044669, 0.00104793891684, -0.00058001104415, -0.00035458138561, 0.00015621764400, 0.00010088156289, -0.00003505374322, -0.00002367321573, 0.00000636368402, 0.00000446042318, -0.00000089682120, -0.00000064858694, 0.00000009197257, 0.00000006831332, -0.00000000610221, -0.00000000463776, 0.00000000019652, 0.00000000015237])
    elif family=="coif16":
        dec_lo = np.array([-0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000002, 0.00000000000003, -0.00000000000022, -0.00000000000029, 0.00000000000184, 0.00000000000253, -0.00000000001350, -0.00000000001889, 0.00000000008675, 0.00000000012338, -0.00000000049379, -0.00000000071498, 0.00000000251443, 0.00000000371289, -0.00000001154860, -0.00000001742574, 0.00000004817368, 0.00000007445243, -0.00000018358908, -0.00000029142697, 0.00000064247571, 0.00000105104552, -0.00000207372315, -0.00000351116294, 0.00000619673477, 0.00001091902605, -0.00001720027604, -0.00003175372313, 0.00004450165162, 0.00008666386407, -0.00010783305712, -0.00022237656648, 0.00024655483339, 0.00053606828793, -0.00053786084965, -0.00120984533977, 0.00113473859937, 0.00254216301610, -0.00234320018818, -0.00494062819525, 0.00476232019250, 0.00882318718398, -0.00950315885774, -0.01439504906308, 0.01851129491005, 0.02135308989071, -0.03519902731387, -0.02868782834982, 0.06665164158949, 0.03479958514972, -0.13592591498912, -0.03801409965110, 0.44797844765328, 0.74441085107759, 0.43948622906261, -0.03280546383359, -0.12781007706564, 0.02578422990745, 0.05928039565969, -0.01805676077382, -0.02889277993976, 0.01122539739109, 0.01347684513905, -0.00616720474023, -0.00578157642258, 0.00297777748316, 0.00222806275271, -0.00125491040143, -0.00075756111566, 0.00045758095419, 0.00022347487351, -0.00014276520952, -0.00005616532905, 0.00003756395792, 0.00001176562458, -0.00000817465630, -0.00000199600860, 0.00000143207920, 0.00000026320534, -0.00000019408664, -0.00000002529381, 0.00000001909694, 0.00000000157499, -0.00000000121376, -0.00000000004768, 0.00000000003741])
        dec_hi = np.array([-0.00000000003741, -0.00000000004768, 0.00000000121376, 0.00000000157499, -0.00000001909694, -0.00000002529381, 0.00000019408664, 0.00000026320534, -0.00000143207920, -0.00000199600860, 0.00000817465630, 0.00001176562458, -0.00003756395792, -0.00005616532905, 0.00014276520952, 0.00022347487351, -0.00045758095419, -0.00075756111566, 0.00125491040143, 0.00222806275271, -0.00297777748316, -0.00578157642258, 0.00616720474023, 0.01347684513905, -0.01122539739109, -0.02889277993976, 0.01805676077382, 0.05928039565969, -0.02578422990745, -0.12781007706564, 0.03280546383359, 0.43948622906261, -0.74441085107759, 0.44797844765328, 0.03801409965110, -0.13592591498912, -0.03479958514972, 0.06665164158949, 0.02868782834982, -0.03519902731387, -0.02135308989071, 0.01851129491005, 0.01439504906308, -0.00950315885774, -0.00882318718398, 0.00476232019250, 0.00494062819525, -0.00234320018818, -0.00254216301610, 0.00113473859937, 0.00120984533977, -0.00053786084965, -0.00053606828793, 0.00024655483339, 0.00022237656648, -0.00010783305712, -0.00008666386407, 0.00004450165162, 0.00003175372313, -0.00001720027604, -0.00001091902605, 0.00000619673477, 0.00000351116294, -0.00000207372315, -0.00000105104552, 0.00000064247571, 0.00000029142697, -0.00000018358908, -0.00000007445243, 0.00000004817368, 0.00000001742574, -0.00000001154860, -0.00000000371289, 0.00000000251443, 0.00000000071498, -0.00000000049379, -0.00000000012338, 0.00000000008675, 0.00000000001889, -0.00000000001350, -0.00000000000253, 0.00000000000184, 0.00000000000029, -0.00000000000022, -0.00000000000003, 0.00000000000002, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000])
        rec_lo = np.array([0.00000000003741, -0.00000000004768, -0.00000000121376, 0.00000000157499, 0.00000001909694, -0.00000002529381, -0.00000019408664, 0.00000026320534, 0.00000143207920, -0.00000199600860, -0.00000817465630, 0.00001176562458, 0.00003756395792, -0.00005616532905, -0.00014276520952, 0.00022347487351, 0.00045758095419, -0.00075756111566, -0.00125491040143, 0.00222806275271, 0.00297777748316, -0.00578157642258, -0.00616720474023, 0.01347684513905, 0.01122539739109, -0.02889277993976, -0.01805676077382, 0.05928039565969, 0.02578422990745, -0.12781007706564, -0.03280546383359, 0.43948622906261, 0.74441085107759, 0.44797844765328, -0.03801409965110, -0.13592591498912, 0.03479958514972, 0.06665164158949, -0.02868782834982, -0.03519902731387, 0.02135308989071, 0.01851129491005, -0.01439504906308, -0.00950315885774, 0.00882318718398, 0.00476232019250, -0.00494062819525, -0.00234320018818, 0.00254216301610, 0.00113473859937, -0.00120984533977, -0.00053786084965, 0.00053606828793, 0.00024655483339, -0.00022237656648, -0.00010783305712, 0.00008666386407, 0.00004450165162, -0.00003175372313, -0.00001720027604, 0.00001091902605, 0.00000619673477, -0.00000351116294, -0.00000207372315, 0.00000105104552, 0.00000064247571, -0.00000029142697, -0.00000018358908, 0.00000007445243, 0.00000004817368, -0.00000001742574, -0.00000001154860, 0.00000000371289, 0.00000000251443, -0.00000000071498, -0.00000000049379, 0.00000000012338, 0.00000000008675, -0.00000000001889, -0.00000000001350, 0.00000000000253, 0.00000000000184, -0.00000000000029, -0.00000000000022, 0.00000000000003, 0.00000000000002, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000002, -0.00000000000003, -0.00000000000022, 0.00000000000029, 0.00000000000184, -0.00000000000253, -0.00000000001350, 0.00000000001889, 0.00000000008675, -0.00000000012338, -0.00000000049379, 0.00000000071498, 0.00000000251443, -0.00000000371289, -0.00000001154860, 0.00000001742574, 0.00000004817368, -0.00000007445243, -0.00000018358908, 0.00000029142697, 0.00000064247571, -0.00000105104552, -0.00000207372315, 0.00000351116294, 0.00000619673477, -0.00001091902605, -0.00001720027604, 0.00003175372313, 0.00004450165162, -0.00008666386407, -0.00010783305712, 0.00022237656648, 0.00024655483339, -0.00053606828793, -0.00053786084965, 0.00120984533977, 0.00113473859937, -0.00254216301610, -0.00234320018818, 0.00494062819525, 0.00476232019250, -0.00882318718398, -0.00950315885774, 0.01439504906308, 0.01851129491005, -0.02135308989071, -0.03519902731387, 0.02868782834982, 0.06665164158949, -0.03479958514972, -0.13592591498912, 0.03801409965110, 0.44797844765328, -0.74441085107759, 0.43948622906261, 0.03280546383359, -0.12781007706564, -0.02578422990745, 0.05928039565969, 0.01805676077382, -0.02889277993976, -0.01122539739109, 0.01347684513905, 0.00616720474023, -0.00578157642258, -0.00297777748316, 0.00222806275271, 0.00125491040143, -0.00075756111566, -0.00045758095419, 0.00022347487351, 0.00014276520952, -0.00005616532905, -0.00003756395792, 0.00001176562458, 0.00000817465630, -0.00000199600860, -0.00000143207920, 0.00000026320534, 0.00000019408664, -0.00000002529381, -0.00000001909694, 0.00000000157499, 0.00000000121376, -0.00000000004768, -0.00000000003741])
    elif family=="coif17":
        dec_lo = np.array([-0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000002, -0.00000000000002, 0.00000000000017, 0.00000000000023, -0.00000000000133, -0.00000000000183, 0.00000000000915, 0.00000000001272, -0.00000000005572, -0.00000000007865, 0.00000000030370, 0.00000000043594, -0.00000000149428, -0.00000000218480, 0.00000000668292, 0.00000000997066, -0.00000002732772, -0.00000004169074, 0.00000010269527, 0.00000016059153, -0.00000035623547, -0.00000057267800, 0.00000114513038, 0.00000189928068, -0.00000342275731, -0.00000588363073, 0.00000954094608, 0.00001709458331, -0.00002487360746, -0.00004674968584, 0.00006085540590, 0.00012063544337, -0.00014043136492, -0.00029391737298, 0.00030804457557, 0.00067500537229, -0.00064926747401, -0.00145555980508, 0.00133083418196, 0.00293101845437, -0.00267814233529, -0.00547795912484, 0.00530800723460, 0.00944627552284, -0.01032766598728, -0.01495136425606, 0.01963029395841, 0.02162718108412, -0.03651586618685, -0.02849019813686, 0.06792967266468, 0.03408179745380, -0.13683866582966, -0.03693320069683, 0.44823872485150, 0.74328122144114, 0.43998382065204, -0.03195235252062, -0.12892696344405, 0.02539065863933, 0.06069883609932, -0.01810211838528, -0.03026535046257, 0.01154189870877, 0.01456427664808, -0.00655615479827, -0.00650589491282, 0.00330231945343, 0.00263790200953, -0.00146656638866, -0.00095502641247, 0.00057018926878, 0.00030426489109, -0.00019235958525, -0.00008403499222, 0.00005567734714, 0.00001977791991, -0.00001362511749, -0.00000388274193, 0.00000276432850, 0.00000061797282, -0.00000045249840, -0.00000007654856, 0.00000005742465, 0.00000000691968, -0.00000000530140, -0.00000000040587, 0.00000000031674, 0.00000000001159, -0.00000000000919])
        dec_hi = np.array([0.00000000000919, 0.00000000001159, -0.00000000031674, -0.00000000040587, 0.00000000530140, 0.00000000691968, -0.00000005742465, -0.00000007654856, 0.00000045249840, 0.00000061797282, -0.00000276432850, -0.00000388274193, 0.00001362511749, 0.00001977791991, -0.00005567734714, -0.00008403499222, 0.00019235958525, 0.00030426489109, -0.00057018926878, -0.00095502641247, 0.00146656638866, 0.00263790200953, -0.00330231945343, -0.00650589491282, 0.00655615479827, 0.01456427664808, -0.01154189870877, -0.03026535046257, 0.01810211838528, 0.06069883609932, -0.02539065863933, -0.12892696344405, 0.03195235252062, 0.43998382065204, -0.74328122144114, 0.44823872485150, 0.03693320069683, -0.13683866582966, -0.03408179745380, 0.06792967266468, 0.02849019813686, -0.03651586618685, -0.02162718108412, 0.01963029395841, 0.01495136425606, -0.01032766598728, -0.00944627552284, 0.00530800723460, 0.00547795912484, -0.00267814233529, -0.00293101845437, 0.00133083418196, 0.00145555980508, -0.00064926747401, -0.00067500537229, 0.00030804457557, 0.00029391737298, -0.00014043136492, -0.00012063544337, 0.00006085540590, 0.00004674968584, -0.00002487360746, -0.00001709458331, 0.00000954094608, 0.00000588363073, -0.00000342275731, -0.00000189928068, 0.00000114513038, 0.00000057267800, -0.00000035623547, -0.00000016059153, 0.00000010269527, 0.00000004169074, -0.00000002732772, -0.00000000997066, 0.00000000668292, 0.00000000218480, -0.00000000149428, -0.00000000043594, 0.00000000030370, 0.00000000007865, -0.00000000005572, -0.00000000001272, 0.00000000000915, 0.00000000000183, -0.00000000000133, -0.00000000000023, 0.00000000000017, 0.00000000000002, -0.00000000000002, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000])
        rec_lo = np.array([-0.00000000000919, 0.00000000001159, 0.00000000031674, -0.00000000040587, -0.00000000530140, 0.00000000691968, 0.00000005742465, -0.00000007654856, -0.00000045249840, 0.00000061797282, 0.00000276432850, -0.00000388274193, -0.00001362511749, 0.00001977791991, 0.00005567734714, -0.00008403499222, -0.00019235958525, 0.00030426489109, 0.00057018926878, -0.00095502641247, -0.00146656638866, 0.00263790200953, 0.00330231945343, -0.00650589491282, -0.00655615479827, 0.01456427664808, 0.01154189870877, -0.03026535046257, -0.01810211838528, 0.06069883609932, 0.02539065863933, -0.12892696344405, -0.03195235252062, 0.43998382065204, 0.74328122144114, 0.44823872485150, -0.03693320069683, -0.13683866582966, 0.03408179745380, 0.06792967266468, -0.02849019813686, -0.03651586618685, 0.02162718108412, 0.01963029395841, -0.01495136425606, -0.01032766598728, 0.00944627552284, 0.00530800723460, -0.00547795912484, -0.00267814233529, 0.00293101845437, 0.00133083418196, -0.00145555980508, -0.00064926747401, 0.00067500537229, 0.00030804457557, -0.00029391737298, -0.00014043136492, 0.00012063544337, 0.00006085540590, -0.00004674968584, -0.00002487360746, 0.00001709458331, 0.00000954094608, -0.00000588363073, -0.00000342275731, 0.00000189928068, 0.00000114513038, -0.00000057267800, -0.00000035623547, 0.00000016059153, 0.00000010269527, -0.00000004169074, -0.00000002732772, 0.00000000997066, 0.00000000668292, -0.00000000218480, -0.00000000149428, 0.00000000043594, 0.00000000030370, -0.00000000007865, -0.00000000005572, 0.00000000001272, 0.00000000000915, -0.00000000000183, -0.00000000000133, 0.00000000000023, 0.00000000000017, -0.00000000000002, -0.00000000000002, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000002, 0.00000000000002, 0.00000000000017, -0.00000000000023, -0.00000000000133, 0.00000000000183, 0.00000000000915, -0.00000000001272, -0.00000000005572, 0.00000000007865, 0.00000000030370, -0.00000000043594, -0.00000000149428, 0.00000000218480, 0.00000000668292, -0.00000000997066, -0.00000002732772, 0.00000004169074, 0.00000010269527, -0.00000016059153, -0.00000035623547, 0.00000057267800, 0.00000114513038, -0.00000189928068, -0.00000342275731, 0.00000588363073, 0.00000954094608, -0.00001709458331, -0.00002487360746, 0.00004674968584, 0.00006085540590, -0.00012063544337, -0.00014043136492, 0.00029391737298, 0.00030804457557, -0.00067500537229, -0.00064926747401, 0.00145555980508, 0.00133083418196, -0.00293101845437, -0.00267814233529, 0.00547795912484, 0.00530800723460, -0.00944627552284, -0.01032766598728, 0.01495136425606, 0.01963029395841, -0.02162718108412, -0.03651586618685, 0.02849019813686, 0.06792967266468, -0.03408179745380, -0.13683866582966, 0.03693320069683, 0.44823872485150, -0.74328122144114, 0.43998382065204, 0.03195235252062, -0.12892696344405, -0.02539065863933, 0.06069883609932, 0.01810211838528, -0.03026535046257, -0.01154189870877, 0.01456427664808, 0.00655615479827, -0.00650589491282, -0.00330231945343, 0.00263790200953, 0.00146656638866, -0.00095502641247, -0.00057018926878, 0.00030426489109, 0.00019235958525, -0.00008403499222, -0.00005567734714, 0.00001977791991, 0.00001362511749, -0.00000388274193, -0.00000276432850, 0.00000061797282, 0.00000045249840, -0.00000007654856, -0.00000005742465, 0.00000000691968, 0.00000000530140, -0.00000000040587, -0.00000000031674, 0.00000000001159, 0.00000000000919])
    elif family=="db1":
        dec_lo = np.array([0.70710678118655, 0.70710678118655])
        dec_hi = np.array([-0.70710678118655, 0.70710678118655])
        rec_lo = np.array([0.70710678118655, 0.70710678118655])
        rec_hi = np.array([0.70710678118655, -0.70710678118655])
    elif family=="db2":
        dec_lo = np.array([-0.12940952255126, 0.22414386804201, 0.83651630373781, 0.48296291314453])
        dec_hi = np.array([-0.48296291314453, 0.83651630373781, -0.22414386804201, -0.12940952255126])
        rec_lo = np.array([0.48296291314453, 0.83651630373781, 0.22414386804201, -0.12940952255126])
        rec_hi = np.array([-0.12940952255126, -0.22414386804201, 0.83651630373781, -0.48296291314453])
    elif family=="db3":
        dec_lo = np.array([0.03522629188571, -0.08544127388203, -0.13501102001025, 0.45987750211849, 0.80689150931109, 0.33267055295008])
        dec_hi = np.array([-0.33267055295008, 0.80689150931109, -0.45987750211849, -0.13501102001025, 0.08544127388203, 0.03522629188571])
        rec_lo = np.array([0.33267055295008, 0.80689150931109, 0.45987750211849, -0.13501102001025, -0.08544127388203, 0.03522629188571])
        rec_hi = np.array([0.03522629188571, 0.08544127388203, -0.13501102001025, -0.45987750211849, 0.80689150931109, -0.33267055295008])
    elif family=="db4":
        dec_lo = np.array([-0.01059740178507, 0.03288301166689, 0.03084138183556, -0.18703481171909, -0.02798376941686, 0.63088076792986, 0.71484657055292, 0.23037781330890])
        dec_hi = np.array([-0.23037781330890, 0.71484657055292, -0.63088076792986, -0.02798376941686, 0.18703481171909, 0.03084138183556, -0.03288301166689, -0.01059740178507])
        rec_lo = np.array([0.23037781330890, 0.71484657055292, 0.63088076792986, -0.02798376941686, -0.18703481171909, 0.03084138183556, 0.03288301166689, -0.01059740178507])
        rec_hi = np.array([-0.01059740178507, -0.03288301166689, 0.03084138183556, 0.18703481171909, -0.02798376941686, -0.63088076792986, 0.71484657055292, -0.23037781330890])
    elif family=="db5":
        dec_lo = np.array([0.00333572528547, -0.01258075199908, -0.00624149021280, 0.07757149384005, -0.03224486958464, -0.24229488706638, 0.13842814590132, 0.72430852843777, 0.60382926979719, 0.16010239797419])
        dec_hi = np.array([-0.16010239797419, 0.60382926979719, -0.72430852843777, 0.13842814590132, 0.24229488706638, -0.03224486958464, -0.07757149384005, -0.00624149021280, 0.01258075199908, 0.00333572528547])
        rec_lo = np.array([0.16010239797419, 0.60382926979719, 0.72430852843777, 0.13842814590132, -0.24229488706638, -0.03224486958464, 0.07757149384005, -0.00624149021280, -0.01258075199908, 0.00333572528547])
        rec_hi = np.array([0.00333572528547, 0.01258075199908, -0.00624149021280, -0.07757149384005, -0.03224486958464, 0.24229488706638, 0.13842814590132, -0.72430852843777, 0.60382926979719, -0.16010239797419])
    elif family=="db6":
        dec_lo = np.array([-0.00107730108531, 0.00477725751095, 0.00055384220116, -0.03158203931749, 0.02752286553031, 0.09750160558732, -0.12976686756726, -0.22626469396544, 0.31525035170920, 0.75113390802110, 0.49462389039845, 0.11154074335011])
        dec_hi = np.array([-0.11154074335011, 0.49462389039845, -0.75113390802110, 0.31525035170920, 0.22626469396544, -0.12976686756726, -0.09750160558732, 0.02752286553031, 0.03158203931749, 0.00055384220116, -0.00477725751095, -0.00107730108531])
        rec_lo = np.array([0.11154074335011, 0.49462389039845, 0.75113390802110, 0.31525035170920, -0.22626469396544, -0.12976686756726, 0.09750160558732, 0.02752286553031, -0.03158203931749, 0.00055384220116, 0.00477725751095, -0.00107730108531])
        rec_hi = np.array([-0.00107730108531, -0.00477725751095, 0.00055384220116, 0.03158203931749, 0.02752286553031, -0.09750160558732, -0.12976686756726, 0.22626469396544, 0.31525035170920, -0.75113390802110, 0.49462389039845, -0.11154074335011])
    elif family=="db7":
        dec_lo = np.array([0.00035371379997, -0.00180164070405, 0.00042957797292, 0.01255099855610, -0.01657454163067, -0.03802993693501, 0.08061260915108, 0.07130921926683, -0.22403618499387, -0.14390600392856, 0.46978228740519, 0.72913209084624, 0.39653931948192, 0.07785205408501])
        dec_hi = np.array([-0.07785205408501, 0.39653931948192, -0.72913209084624, 0.46978228740519, 0.14390600392856, -0.22403618499387, -0.07130921926683, 0.08061260915108, 0.03802993693501, -0.01657454163067, -0.01255099855610, 0.00042957797292, 0.00180164070405, 0.00035371379997])
        rec_lo = np.array([0.07785205408501, 0.39653931948192, 0.72913209084624, 0.46978228740519, -0.14390600392856, -0.22403618499387, 0.07130921926683, 0.08061260915108, -0.03802993693501, -0.01657454163067, 0.01255099855610, 0.00042957797292, -0.00180164070405, 0.00035371379997])
        rec_hi = np.array([0.00035371379997, 0.00180164070405, 0.00042957797292, -0.01255099855610, -0.01657454163067, 0.03802993693501, 0.08061260915108, -0.07130921926683, -0.22403618499387, 0.14390600392856, 0.46978228740519, -0.72913209084624, 0.39653931948192, -0.07785205408501])
    elif family=="db8":
        dec_lo = np.array([-0.00011747678412, 0.00067544940645, -0.00039174037338, -0.00487035299345, 0.00874609404741, 0.01398102791740, -0.04408825393079, -0.01736930100181, 0.12874742662048, 0.00047248457391, -0.28401554296155, -0.01582910525635, 0.58535468365421, 0.67563073629729, 0.31287159091430, 0.05441584224310])
        dec_hi = np.array([-0.05441584224310, 0.31287159091430, -0.67563073629729, 0.58535468365421, 0.01582910525635, -0.28401554296155, -0.00047248457391, 0.12874742662048, 0.01736930100181, -0.04408825393079, -0.01398102791740, 0.00874609404741, 0.00487035299345, -0.00039174037338, -0.00067544940645, -0.00011747678412])
        rec_lo = np.array([0.05441584224310, 0.31287159091430, 0.67563073629729, 0.58535468365421, -0.01582910525635, -0.28401554296155, 0.00047248457391, 0.12874742662048, -0.01736930100181, -0.04408825393079, 0.01398102791740, 0.00874609404741, -0.00487035299345, -0.00039174037338, 0.00067544940645, -0.00011747678412])
        rec_hi = np.array([-0.00011747678412, -0.00067544940645, -0.00039174037338, 0.00487035299345, 0.00874609404741, -0.01398102791740, -0.04408825393079, 0.01736930100181, 0.12874742662048, -0.00047248457391, -0.28401554296155, 0.01582910525635, 0.58535468365421, -0.67563073629729, 0.31287159091430, -0.05441584224310])
    elif family=="db9":
        dec_lo = np.array([0.00003934732032, -0.00025196318894, 0.00023038576352, 0.00184764688306, -0.00428150368246, -0.00472320475775, 0.02236166212368, 0.00025094711483, -0.06763282906133, 0.03072568147933, 0.14854074933811, -0.09684078322298, -0.29327378327917, 0.13319738582501, 0.65728807805130, 0.60482312369011, 0.24383467461259, 0.03807794736388])
        dec_hi = np.array([-0.03807794736388, 0.24383467461259, -0.60482312369011, 0.65728807805130, -0.13319738582501, -0.29327378327917, 0.09684078322298, 0.14854074933811, -0.03072568147933, -0.06763282906133, -0.00025094711483, 0.02236166212368, 0.00472320475775, -0.00428150368246, -0.00184764688306, 0.00023038576352, 0.00025196318894, 0.00003934732032])
        rec_lo = np.array([0.03807794736388, 0.24383467461259, 0.60482312369011, 0.65728807805130, 0.13319738582501, -0.29327378327917, -0.09684078322298, 0.14854074933811, 0.03072568147933, -0.06763282906133, 0.00025094711483, 0.02236166212368, -0.00472320475775, -0.00428150368246, 0.00184764688306, 0.00023038576352, -0.00025196318894, 0.00003934732032])
        rec_hi = np.array([0.00003934732032, 0.00025196318894, 0.00023038576352, -0.00184764688306, -0.00428150368246, 0.00472320475775, 0.02236166212368, -0.00025094711483, -0.06763282906133, -0.03072568147933, 0.14854074933811, 0.09684078322298, -0.29327378327917, -0.13319738582501, 0.65728807805130, -0.60482312369011, 0.24383467461259, -0.03807794736388])
    elif family=="db10":
        dec_lo = np.array([-0.00001326420289, 0.00009358867032, -0.00011646685513, -0.00068585669496, 0.00199240529519, 0.00139535174705, -0.01073317548333, 0.00360655356696, 0.03321267405934, -0.02945753682188, -0.07139414716640, 0.09305736460357, 0.12736934033579, -0.19594627437738, -0.24984642432732, 0.28117234366058, 0.68845903945360, 0.52720118893173, 0.18817680007769, 0.02667005790056])
        dec_hi = np.array([-0.02667005790056, 0.18817680007769, -0.52720118893173, 0.68845903945360, -0.28117234366058, -0.24984642432732, 0.19594627437738, 0.12736934033579, -0.09305736460357, -0.07139414716640, 0.02945753682188, 0.03321267405934, -0.00360655356696, -0.01073317548333, -0.00139535174705, 0.00199240529519, 0.00068585669496, -0.00011646685513, -0.00009358867032, -0.00001326420289])
        rec_lo = np.array([0.02667005790056, 0.18817680007769, 0.52720118893173, 0.68845903945360, 0.28117234366058, -0.24984642432732, -0.19594627437738, 0.12736934033579, 0.09305736460357, -0.07139414716640, -0.02945753682188, 0.03321267405934, 0.00360655356696, -0.01073317548333, 0.00139535174705, 0.00199240529519, -0.00068585669496, -0.00011646685513, 0.00009358867032, -0.00001326420289])
        rec_hi = np.array([-0.00001326420289, -0.00009358867032, -0.00011646685513, 0.00068585669496, 0.00199240529519, -0.00139535174705, -0.01073317548333, -0.00360655356696, 0.03321267405934, 0.02945753682188, -0.07139414716640, -0.09305736460357, 0.12736934033579, 0.19594627437738, -0.24984642432732, -0.28117234366058, 0.68845903945360, -0.52720118893173, 0.18817680007769, -0.02667005790056])
    elif family=="db11":
        dec_lo = np.array([0.00000449427428, -0.00003463498419, 0.00005443907470, 0.00024915252355, -0.00089302325067, -0.00030859285882, 0.00492841765606, -0.00334085887301, -0.01536482090620, 0.02084090436018, 0.03133509021905, -0.06643878569503, -0.04647995511668, 0.14981201246638, 0.06604358819668, -0.27423084681795, -0.16227524502749, 0.41196436894791, 0.68568677491620, 0.44989976435605, 0.14406702115062, 0.01869429776147])
        dec_hi = np.array([-0.01869429776147, 0.14406702115062, -0.44989976435605, 0.68568677491620, -0.41196436894791, -0.16227524502749, 0.27423084681795, 0.06604358819668, -0.14981201246638, -0.04647995511668, 0.06643878569503, 0.03133509021905, -0.02084090436018, -0.01536482090620, 0.00334085887301, 0.00492841765606, 0.00030859285882, -0.00089302325067, -0.00024915252355, 0.00005443907470, 0.00003463498419, 0.00000449427428])
        rec_lo = np.array([0.01869429776147, 0.14406702115062, 0.44989976435605, 0.68568677491620, 0.41196436894791, -0.16227524502749, -0.27423084681795, 0.06604358819668, 0.14981201246638, -0.04647995511668, -0.06643878569503, 0.03133509021905, 0.02084090436018, -0.01536482090620, -0.00334085887301, 0.00492841765606, -0.00030859285882, -0.00089302325067, 0.00024915252355, 0.00005443907470, -0.00003463498419, 0.00000449427428])
        rec_hi = np.array([0.00000449427428, 0.00003463498419, 0.00005443907470, -0.00024915252355, -0.00089302325067, 0.00030859285882, 0.00492841765606, 0.00334085887301, -0.01536482090620, -0.02084090436018, 0.03133509021905, 0.06643878569503, -0.04647995511668, -0.14981201246638, 0.06604358819668, 0.27423084681795, -0.16227524502749, -0.41196436894791, 0.68568677491620, -0.44989976435605, 0.14406702115062, -0.01869429776147])
    elif family=="db12":
        dec_lo = np.array([-0.00000152907176, 0.00001277695222, -0.00002424154576, -0.00008850410921, 0.00038865306282, 0.00000654512821, -0.00217950361863, 0.00224860724100, 0.00671149900880, -0.01284082519830, -0.01221864906975, 0.04154627749508, 0.01084913025582, -0.09643212009651, 0.00535956967435, 0.18247860592758, -0.02377925725607, -0.31617845375279, -0.04476388565377, 0.51588647842782, 0.65719872257931, 0.37735513521421, 0.10956627282119, 0.01311225795723])
        dec_hi = np.array([-0.01311225795723, 0.10956627282119, -0.37735513521421, 0.65719872257931, -0.51588647842782, -0.04476388565377, 0.31617845375279, -0.02377925725607, -0.18247860592758, 0.00535956967435, 0.09643212009651, 0.01084913025582, -0.04154627749508, -0.01221864906975, 0.01284082519830, 0.00671149900880, -0.00224860724100, -0.00217950361863, -0.00000654512821, 0.00038865306282, 0.00008850410921, -0.00002424154576, -0.00001277695222, -0.00000152907176])
        rec_lo = np.array([0.01311225795723, 0.10956627282119, 0.37735513521421, 0.65719872257931, 0.51588647842782, -0.04476388565377, -0.31617845375279, -0.02377925725607, 0.18247860592758, 0.00535956967435, -0.09643212009651, 0.01084913025582, 0.04154627749508, -0.01221864906975, -0.01284082519830, 0.00671149900880, 0.00224860724100, -0.00217950361863, 0.00000654512821, 0.00038865306282, -0.00008850410921, -0.00002424154576, 0.00001277695222, -0.00000152907176])
        rec_hi = np.array([-0.00000152907176, -0.00001277695222, -0.00002424154576, 0.00008850410921, 0.00038865306282, -0.00000654512821, -0.00217950361863, -0.00224860724100, 0.00671149900880, 0.01284082519830, -0.01221864906975, -0.04154627749508, 0.01084913025582, 0.09643212009651, 0.00535956967435, -0.18247860592758, -0.02377925725607, 0.31617845375279, -0.04476388565377, -0.51588647842782, 0.65719872257931, -0.37735513521421, 0.10956627282119, -0.01311225795723])
    elif family=="db13":
        dec_lo = np.array([0.00000052200351, -0.00000470041648, 0.00001044193057, 0.00003067853758, -0.00016512898856, 0.00004925152513, 0.00093232613087, -0.00131567391189, -0.00276191123466, 0.00725558940162, 0.00392394144880, -0.02383142071032, 0.00237997225406, 0.05613947710028, -0.02648840647534, -0.10580761818793, 0.07294893365678, 0.17947607942934, -0.12457673075082, -0.31497290771139, 0.08698572617965, 0.58888957043122, 0.61105585115879, 0.31199632216044, 0.08286124387290, 0.00920213353896])
        dec_hi = np.array([-0.00920213353896, 0.08286124387290, -0.31199632216044, 0.61105585115879, -0.58888957043122, 0.08698572617965, 0.31497290771139, -0.12457673075082, -0.17947607942934, 0.07294893365678, 0.10580761818793, -0.02648840647534, -0.05613947710028, 0.00237997225406, 0.02383142071032, 0.00392394144880, -0.00725558940162, -0.00276191123466, 0.00131567391189, 0.00093232613087, -0.00004925152513, -0.00016512898856, -0.00003067853758, 0.00001044193057, 0.00000470041648, 0.00000052200351])
        rec_lo = np.array([0.00920213353896, 0.08286124387290, 0.31199632216044, 0.61105585115879, 0.58888957043122, 0.08698572617965, -0.31497290771139, -0.12457673075082, 0.17947607942934, 0.07294893365678, -0.10580761818793, -0.02648840647534, 0.05613947710028, 0.00237997225406, -0.02383142071032, 0.00392394144880, 0.00725558940162, -0.00276191123466, -0.00131567391189, 0.00093232613087, 0.00004925152513, -0.00016512898856, 0.00003067853758, 0.00001044193057, -0.00000470041648, 0.00000052200351])
        rec_hi = np.array([0.00000052200351, 0.00000470041648, 0.00001044193057, -0.00003067853758, -0.00016512898856, -0.00004925152513, 0.00093232613087, 0.00131567391189, -0.00276191123466, -0.00725558940162, 0.00392394144880, 0.02383142071032, 0.00237997225406, -0.05613947710028, -0.02648840647534, 0.10580761818793, 0.07294893365678, -0.17947607942934, -0.12457673075082, 0.31497290771139, 0.08698572617965, -0.58888957043122, 0.61105585115879, -0.31199632216044, 0.08286124387290, -0.00920213353896])
    elif family=="db14":
        dec_lo = np.array([-0.00000017871400, 0.00000172499468, -0.00000438970490, -0.00001033720918, 0.00006875504253, -0.00004177724577, -0.00038683194731, 0.00070802115424, 0.00106169108561, -0.00384963886802, -0.00074621898927, 0.01278949326633, -0.00561504953036, -0.03018535154039, 0.02698140830791, 0.05523712625922, -0.07154895550405, -0.08674841156817, 0.13998901658446, 0.13839521386481, -0.21803352999328, -0.27168855227875, 0.21867068775891, 0.63118784910486, 0.55430561794089, 0.25485026779262, 0.06236475884940, 0.00646115346009])
        dec_hi = np.array([-0.00646115346009, 0.06236475884940, -0.25485026779262, 0.55430561794089, -0.63118784910486, 0.21867068775891, 0.27168855227875, -0.21803352999328, -0.13839521386481, 0.13998901658446, 0.08674841156817, -0.07154895550405, -0.05523712625922, 0.02698140830791, 0.03018535154039, -0.00561504953036, -0.01278949326633, -0.00074621898927, 0.00384963886802, 0.00106169108561, -0.00070802115424, -0.00038683194731, 0.00004177724577, 0.00006875504253, 0.00001033720918, -0.00000438970490, -0.00000172499468, -0.00000017871400])
        rec_lo = np.array([0.00646115346009, 0.06236475884940, 0.25485026779262, 0.55430561794089, 0.63118784910486, 0.21867068775891, -0.27168855227875, -0.21803352999328, 0.13839521386481, 0.13998901658446, -0.08674841156817, -0.07154895550405, 0.05523712625922, 0.02698140830791, -0.03018535154039, -0.00561504953036, 0.01278949326633, -0.00074621898927, -0.00384963886802, 0.00106169108561, 0.00070802115424, -0.00038683194731, -0.00004177724577, 0.00006875504253, -0.00001033720918, -0.00000438970490, 0.00000172499468, -0.00000017871400])
        rec_hi = np.array([-0.00000017871400, -0.00000172499468, -0.00000438970490, 0.00001033720918, 0.00006875504253, 0.00004177724577, -0.00038683194731, -0.00070802115424, 0.00106169108561, 0.00384963886802, -0.00074621898927, -0.01278949326633, -0.00561504953036, 0.03018535154039, 0.02698140830791, -0.05523712625922, -0.07154895550405, 0.08674841156817, 0.13998901658446, -0.13839521386481, -0.21803352999328, 0.27168855227875, 0.21867068775891, -0.63118784910486, 0.55430561794089, -0.25485026779262, 0.06236475884940, -0.00646115346009])
    elif family=="db15":
        dec_lo = np.array([0.00000006133360, -0.00000063168823, 0.00000181127041, 0.00000336298718, -0.00002813329627, 0.00002579269916, 0.00015589648992, -0.00035956524436, -0.00037348235414, 0.00194332398038, -0.00024175649076, -0.00648773456032, 0.00510100036041, 0.01508391802784, -0.02081005016969, -0.02576700732844, 0.05478055058451, 0.03387714392351, -0.11112093603723, -0.03966617655579, 0.19014671400712, 0.06528295284877, -0.28888259656697, -0.19320413960915, 0.33900253545473, 0.64581314035742, 0.49263177170814, 0.20602386398700, 0.04674339489277, 0.00453853736158])
        dec_hi = np.array([-0.00453853736158, 0.04674339489277, -0.20602386398700, 0.49263177170814, -0.64581314035742, 0.33900253545473, 0.19320413960915, -0.28888259656697, -0.06528295284877, 0.19014671400712, 0.03966617655579, -0.11112093603723, -0.03387714392351, 0.05478055058451, 0.02576700732844, -0.02081005016969, -0.01508391802784, 0.00510100036041, 0.00648773456032, -0.00024175649076, -0.00194332398038, -0.00037348235414, 0.00035956524436, 0.00015589648992, -0.00002579269916, -0.00002813329627, -0.00000336298718, 0.00000181127041, 0.00000063168823, 0.00000006133360])
        rec_lo = np.array([0.00453853736158, 0.04674339489277, 0.20602386398700, 0.49263177170814, 0.64581314035742, 0.33900253545473, -0.19320413960915, -0.28888259656697, 0.06528295284877, 0.19014671400712, -0.03966617655579, -0.11112093603723, 0.03387714392351, 0.05478055058451, -0.02576700732844, -0.02081005016969, 0.01508391802784, 0.00510100036041, -0.00648773456032, -0.00024175649076, 0.00194332398038, -0.00037348235414, -0.00035956524436, 0.00015589648992, 0.00002579269916, -0.00002813329627, 0.00000336298718, 0.00000181127041, -0.00000063168823, 0.00000006133360])
        rec_hi = np.array([0.00000006133360, 0.00000063168823, 0.00000181127041, -0.00000336298718, -0.00002813329627, -0.00002579269916, 0.00015589648992, 0.00035956524436, -0.00037348235414, -0.00194332398038, -0.00024175649076, 0.00648773456032, 0.00510100036041, -0.01508391802784, -0.02081005016969, 0.02576700732844, 0.05478055058451, -0.03387714392351, -0.11112093603723, 0.03966617655579, 0.19014671400712, -0.06528295284877, -0.28888259656697, 0.19320413960915, 0.33900253545473, -0.64581314035742, 0.49263177170814, -0.20602386398700, 0.04674339489277, -0.00453853736158])
    elif family=="db16":
        dec_lo = np.array([-0.00000002109340, 0.00000023087841, -0.00000073636568, -0.00000104357134, 0.00001133660866, -0.00001394566899, -0.00006103596621, 0.00017478724523, 0.00011424152004, -0.00094102174936, 0.00040789698085, 0.00312802338121, -0.00364427962150, -0.00699001456341, 0.01399376885983, 0.01029765964096, -0.03688839769173, -0.00758897436886, 0.07592423604428, -0.00623972275247, -0.13238830556381, 0.02734026375272, 0.21119069394710, -0.02791820813303, -0.32706331052792, -0.08975108940249, 0.44029025688636, 0.63735633208379, 0.43031272284600, 0.16506428348885, 0.03490771432367, 0.00318922092535])
        dec_hi = np.array([-0.00318922092535, 0.03490771432367, -0.16506428348885, 0.43031272284600, -0.63735633208379, 0.44029025688636, 0.08975108940249, -0.32706331052792, 0.02791820813303, 0.21119069394710, -0.02734026375272, -0.13238830556381, 0.00623972275247, 0.07592423604428, 0.00758897436886, -0.03688839769173, -0.01029765964096, 0.01399376885983, 0.00699001456341, -0.00364427962150, -0.00312802338121, 0.00040789698085, 0.00094102174936, 0.00011424152004, -0.00017478724523, -0.00006103596621, 0.00001394566899, 0.00001133660866, 0.00000104357134, -0.00000073636568, -0.00000023087841, -0.00000002109340])
        rec_lo = np.array([0.00318922092535, 0.03490771432367, 0.16506428348885, 0.43031272284600, 0.63735633208379, 0.44029025688636, -0.08975108940249, -0.32706331052792, -0.02791820813303, 0.21119069394710, 0.02734026375272, -0.13238830556381, -0.00623972275247, 0.07592423604428, -0.00758897436886, -0.03688839769173, 0.01029765964096, 0.01399376885983, -0.00699001456341, -0.00364427962150, 0.00312802338121, 0.00040789698085, -0.00094102174936, 0.00011424152004, 0.00017478724523, -0.00006103596621, -0.00001394566899, 0.00001133660866, -0.00000104357134, -0.00000073636568, 0.00000023087841, -0.00000002109340])
        rec_hi = np.array([-0.00000002109340, -0.00000023087841, -0.00000073636568, 0.00000104357134, 0.00001133660866, 0.00001394566899, -0.00006103596621, -0.00017478724523, 0.00011424152004, 0.00094102174936, 0.00040789698085, -0.00312802338121, -0.00364427962150, 0.00699001456341, 0.01399376885983, -0.01029765964096, -0.03688839769173, 0.00758897436886, 0.07592423604428, 0.00623972275247, -0.13238830556381, -0.02734026375272, 0.21119069394710, 0.02791820813303, -0.32706331052792, 0.08975108940249, 0.44029025688636, -0.63735633208379, 0.43031272284600, -0.16506428348885, 0.03490771432367, -0.00318922092535])
    elif family=="db17":
        dec_lo = np.array([0.00000000726749, -0.00000008423948, 0.00000029577009, 0.00000030165496, -0.00000450594248, 0.00000699060099, 0.00002318681380, -0.00008204803202, -0.00002561010957, 0.00043946542777, -0.00032813251941, -0.00143684530480, 0.00230120524215, 0.00296799669153, -0.00860292152032, -0.00304298998135, 0.02273367658395, -0.00327095553582, -0.04692243838927, 0.02231233617810, 0.08110598665416, -0.05709141963168, -0.12681569177829, 0.10113548917747, 0.19731058956501, -0.12659975221588, -0.32832074836396, 0.02731497040329, 0.51831576405694, 0.61099661568462, 0.37035072415264, 0.13121490330782, 0.02598539370361, 0.00224180700104])
        dec_hi = np.array([-0.00224180700104, 0.02598539370361, -0.13121490330782, 0.37035072415264, -0.61099661568462, 0.51831576405694, -0.02731497040329, -0.32832074836396, 0.12659975221588, 0.19731058956501, -0.10113548917747, -0.12681569177829, 0.05709141963168, 0.08110598665416, -0.02231233617810, -0.04692243838927, 0.00327095553582, 0.02273367658395, 0.00304298998135, -0.00860292152032, -0.00296799669153, 0.00230120524215, 0.00143684530480, -0.00032813251941, -0.00043946542777, -0.00002561010957, 0.00008204803202, 0.00002318681380, -0.00000699060099, -0.00000450594248, -0.00000030165496, 0.00000029577009, 0.00000008423948, 0.00000000726749])
        rec_lo = np.array([0.00224180700104, 0.02598539370361, 0.13121490330782, 0.37035072415264, 0.61099661568462, 0.51831576405694, 0.02731497040329, -0.32832074836396, -0.12659975221588, 0.19731058956501, 0.10113548917747, -0.12681569177829, -0.05709141963168, 0.08110598665416, 0.02231233617810, -0.04692243838927, -0.00327095553582, 0.02273367658395, -0.00304298998135, -0.00860292152032, 0.00296799669153, 0.00230120524215, -0.00143684530480, -0.00032813251941, 0.00043946542777, -0.00002561010957, -0.00008204803202, 0.00002318681380, 0.00000699060099, -0.00000450594248, 0.00000030165496, 0.00000029577009, -0.00000008423948, 0.00000000726749])
        rec_hi = np.array([0.00000000726749, 0.00000008423948, 0.00000029577009, -0.00000030165496, -0.00000450594248, -0.00000699060099, 0.00002318681380, 0.00008204803202, -0.00002561010957, -0.00043946542777, -0.00032813251941, 0.00143684530480, 0.00230120524215, -0.00296799669153, -0.00860292152032, 0.00304298998135, 0.02273367658395, 0.00327095553582, -0.04692243838927, -0.02231233617810, 0.08110598665416, 0.05709141963168, -0.12681569177829, -0.10113548917747, 0.19731058956501, 0.12659975221588, -0.32832074836396, -0.02731497040329, 0.51831576405694, -0.61099661568462, 0.37035072415264, -0.13121490330782, 0.02598539370361, -0.00224180700104])
    elif family=="db18":
        dec_lo = np.array([-0.00000000250793, 0.00000003068836, -0.00000011760988, -0.00000007691633, 0.00000176871298, -0.00000333263448, -0.00000852060254, 0.00003741237881, -0.00000015359171, -0.00019864855231, 0.00021358156191, 0.00062846568297, -0.00134059629834, -0.00111873266699, 0.00494334360547, 0.00011863003386, -0.01305148094661, 0.00626216795431, 0.02667070592647, -0.02373321039586, -0.04452614190298, 0.05705124773854, 0.06488721621191, -0.10675224665983, -0.09233188415085, 0.16708131276326, 0.14953397556538, -0.21648093400514, -0.29365404073656, 0.14722311196993, 0.57180165488865, 0.57182680776661, 0.31467894133703, 0.10358846582242, 0.01928853172415, 0.00157631021844])
        dec_hi = np.array([-0.00157631021844, 0.01928853172415, -0.10358846582242, 0.31467894133703, -0.57182680776661, 0.57180165488865, -0.14722311196993, -0.29365404073656, 0.21648093400514, 0.14953397556538, -0.16708131276326, -0.09233188415085, 0.10675224665983, 0.06488721621191, -0.05705124773854, -0.04452614190298, 0.02373321039586, 0.02667070592647, -0.00626216795431, -0.01305148094661, -0.00011863003386, 0.00494334360547, 0.00111873266699, -0.00134059629834, -0.00062846568297, 0.00021358156191, 0.00019864855231, -0.00000015359171, -0.00003741237881, -0.00000852060254, 0.00000333263448, 0.00000176871298, 0.00000007691633, -0.00000011760988, -0.00000003068836, -0.00000000250793])
        rec_lo = np.array([0.00157631021844, 0.01928853172415, 0.10358846582242, 0.31467894133703, 0.57182680776661, 0.57180165488865, 0.14722311196993, -0.29365404073656, -0.21648093400514, 0.14953397556538, 0.16708131276326, -0.09233188415085, -0.10675224665983, 0.06488721621191, 0.05705124773854, -0.04452614190298, -0.02373321039586, 0.02667070592647, 0.00626216795431, -0.01305148094661, 0.00011863003386, 0.00494334360547, -0.00111873266699, -0.00134059629834, 0.00062846568297, 0.00021358156191, -0.00019864855231, -0.00000015359171, 0.00003741237881, -0.00000852060254, -0.00000333263448, 0.00000176871298, -0.00000007691633, -0.00000011760988, 0.00000003068836, -0.00000000250793])
        rec_hi = np.array([-0.00000000250793, -0.00000003068836, -0.00000011760988, 0.00000007691633, 0.00000176871298, 0.00000333263448, -0.00000852060254, -0.00003741237881, -0.00000015359171, 0.00019864855231, 0.00021358156191, -0.00062846568297, -0.00134059629834, 0.00111873266699, 0.00494334360547, -0.00011863003386, -0.01305148094661, -0.00626216795431, 0.02667070592647, 0.02373321039586, -0.04452614190298, -0.05705124773854, 0.06488721621191, 0.10675224665983, -0.09233188415085, -0.16708131276326, 0.14953397556538, 0.21648093400514, -0.29365404073656, -0.14722311196993, 0.57180165488865, -0.57182680776661, 0.31467894133703, -0.10358846582242, 0.01928853172415, -0.00157631021844])
    elif family=="db19":
        dec_lo = np.array([0.00000000086668, -0.00000001116402, 0.00000004636938, 0.00000001447088, -0.00000068627557, 0.00000153193148, 0.00000301096432, -0.00001664017630, 0.00000510595049, 0.00008711270467, -0.00012460079173, -0.00026067613568, 0.00073580252051, 0.00034180865346, -0.00268755180070, 0.00076895435926, 0.00704074736711, -0.00586692228101, -0.01398838867854, 0.01937554988918, 0.02162376740959, -0.04567422627723, -0.02650123625012, 0.08690675555581, 0.02758435062563, -0.14278569503874, -0.03351854190230, 0.21234974330628, 0.07465226970810, -0.28583863175583, -0.22809139421548, 0.26089495265104, 0.60170454912754, 0.52443637746465, 0.26438843174090, 0.08127811326546, 0.01428109845076, 0.00110866976318])
        dec_hi = np.array([-0.00110866976318, 0.01428109845076, -0.08127811326546, 0.26438843174090, -0.52443637746465, 0.60170454912754, -0.26089495265104, -0.22809139421548, 0.28583863175583, 0.07465226970810, -0.21234974330628, -0.03351854190230, 0.14278569503874, 0.02758435062563, -0.08690675555581, -0.02650123625012, 0.04567422627723, 0.02162376740959, -0.01937554988918, -0.01398838867854, 0.00586692228101, 0.00704074736711, -0.00076895435926, -0.00268755180070, -0.00034180865346, 0.00073580252051, 0.00026067613568, -0.00012460079173, -0.00008711270467, 0.00000510595049, 0.00001664017630, 0.00000301096432, -0.00000153193148, -0.00000068627557, -0.00000001447088, 0.00000004636938, 0.00000001116402, 0.00000000086668])
        rec_lo = np.array([0.00110866976318, 0.01428109845076, 0.08127811326546, 0.26438843174090, 0.52443637746465, 0.60170454912754, 0.26089495265104, -0.22809139421548, -0.28583863175583, 0.07465226970810, 0.21234974330628, -0.03351854190230, -0.14278569503874, 0.02758435062563, 0.08690675555581, -0.02650123625012, -0.04567422627723, 0.02162376740959, 0.01937554988918, -0.01398838867854, -0.00586692228101, 0.00704074736711, 0.00076895435926, -0.00268755180070, 0.00034180865346, 0.00073580252051, -0.00026067613568, -0.00012460079173, 0.00008711270467, 0.00000510595049, -0.00001664017630, 0.00000301096432, 0.00000153193148, -0.00000068627557, 0.00000001447088, 0.00000004636938, -0.00000001116402, 0.00000000086668])
        rec_hi = np.array([0.00000000086668, 0.00000001116402, 0.00000004636938, -0.00000001447088, -0.00000068627557, -0.00000153193148, 0.00000301096432, 0.00001664017630, 0.00000510595049, -0.00008711270467, -0.00012460079173, 0.00026067613568, 0.00073580252051, -0.00034180865346, -0.00268755180070, -0.00076895435926, 0.00704074736711, 0.00586692228101, -0.01398838867854, -0.01937554988918, 0.02162376740959, 0.04567422627723, -0.02650123625012, -0.08690675555581, 0.02758435062563, 0.14278569503874, -0.03351854190230, -0.21234974330628, 0.07465226970810, 0.28583863175583, -0.22809139421548, -0.26089495265104, 0.60170454912754, -0.52443637746465, 0.26438843174090, -0.08127811326546, 0.01428109845076, -0.00110866976318])
    elif family=="db20":
        dec_lo = np.array([-0.00000000029988, 0.00000000405613, -0.00000001814843, 0.00000000020143, 0.00000026339242, -0.00000068470796, -0.00000101199401, 0.00000724124829, -0.00000437614386, -0.00003710586183, 0.00006774280828, 0.00010153288974, -0.00038510474870, -0.00005349759844, 0.00139255961932, -0.00083156217282, -0.00358149425961, 0.00442054238705, 0.00672162730226, -0.01381052613715, -0.00878932492390, 0.03229429953077, 0.00587468181181, -0.06172289962468, 0.00563224685731, 0.10229171917444, -0.02471682733861, -0.15545875070727, 0.03985024645777, 0.22829105081992, -0.01672708830908, -0.32678680043403, -0.13921208801148, 0.36150229873933, 0.61049323893859, 0.47269618531090, 0.21994211355140, 0.06342378045908, 0.01054939462495, 0.00077995361367])
        dec_hi = np.array([-0.00077995361367, 0.01054939462495, -0.06342378045908, 0.21994211355140, -0.47269618531090, 0.61049323893859, -0.36150229873933, -0.13921208801148, 0.32678680043403, -0.01672708830908, -0.22829105081992, 0.03985024645777, 0.15545875070727, -0.02471682733861, -0.10229171917444, 0.00563224685731, 0.06172289962468, 0.00587468181181, -0.03229429953077, -0.00878932492390, 0.01381052613715, 0.00672162730226, -0.00442054238705, -0.00358149425961, 0.00083156217282, 0.00139255961932, 0.00005349759844, -0.00038510474870, -0.00010153288974, 0.00006774280828, 0.00003710586183, -0.00000437614386, -0.00000724124829, -0.00000101199401, 0.00000068470796, 0.00000026339242, -0.00000000020143, -0.00000001814843, -0.00000000405613, -0.00000000029988])
        rec_lo = np.array([0.00077995361367, 0.01054939462495, 0.06342378045908, 0.21994211355140, 0.47269618531090, 0.61049323893859, 0.36150229873933, -0.13921208801148, -0.32678680043403, -0.01672708830908, 0.22829105081992, 0.03985024645777, -0.15545875070727, -0.02471682733861, 0.10229171917444, 0.00563224685731, -0.06172289962468, 0.00587468181181, 0.03229429953077, -0.00878932492390, -0.01381052613715, 0.00672162730226, 0.00442054238705, -0.00358149425961, -0.00083156217282, 0.00139255961932, -0.00005349759844, -0.00038510474870, 0.00010153288974, 0.00006774280828, -0.00003710586183, -0.00000437614386, 0.00000724124829, -0.00000101199401, -0.00000068470796, 0.00000026339242, 0.00000000020143, -0.00000001814843, 0.00000000405613, -0.00000000029988])
        rec_hi = np.array([-0.00000000029988, -0.00000000405613, -0.00000001814843, -0.00000000020143, 0.00000026339242, 0.00000068470796, -0.00000101199401, -0.00000724124829, -0.00000437614386, 0.00003710586183, 0.00006774280828, -0.00010153288974, -0.00038510474870, 0.00005349759844, 0.00139255961932, 0.00083156217282, -0.00358149425961, -0.00442054238705, 0.00672162730226, 0.01381052613715, -0.00878932492390, -0.03229429953077, 0.00587468181181, 0.06172289962468, 0.00563224685731, -0.10229171917444, -0.02471682733861, 0.15545875070727, 0.03985024645777, -0.22829105081992, -0.01672708830908, 0.32678680043403, -0.13921208801148, -0.36150229873933, 0.61049323893859, -0.47269618531090, 0.21994211355140, -0.06342378045908, 0.01054939462495, -0.00077995361367])
    elif family=="db21":
        dec_lo = np.array([0.00000000010388, -0.00000000147195, 0.00000000705803, -0.00000000225401, -0.00000010004009, 0.00000029921366, 0.00000031660954, -0.00000309001716, 0.00000279033054, 0.00001535482509, -0.00003499665985, -0.00003635520250, 0.00019366465042, -0.00003196406278, -0.00069067111708, 0.00063941850051, 0.00171660704063, -0.00295837403893, -0.00289133434859, 0.00898882438197, 0.00240347092081, -0.02089205367798, 0.00335775639034, 0.03972683542785, -0.01865385920212, -0.06497750489373, 0.04572340574923, 0.09660039032372, -0.08177594298086, -0.13994042493255, 0.11523329843969, 0.21156452768087, -0.11239707156845, -0.33566408953053, -0.03572291961726, 0.44459045192760, 0.60150609493500, 0.41968794493936, 0.18135962544038, 0.04924777153818, 0.00777663905235, 0.00054882250985])
        dec_hi = np.array([-0.00054882250985, 0.00777663905235, -0.04924777153818, 0.18135962544038, -0.41968794493936, 0.60150609493500, -0.44459045192760, -0.03572291961726, 0.33566408953053, -0.11239707156845, -0.21156452768087, 0.11523329843969, 0.13994042493255, -0.08177594298086, -0.09660039032372, 0.04572340574923, 0.06497750489373, -0.01865385920212, -0.03972683542785, 0.00335775639034, 0.02089205367798, 0.00240347092081, -0.00898882438197, -0.00289133434859, 0.00295837403893, 0.00171660704063, -0.00063941850051, -0.00069067111708, 0.00003196406278, 0.00019366465042, 0.00003635520250, -0.00003499665985, -0.00001535482509, 0.00000279033054, 0.00000309001716, 0.00000031660954, -0.00000029921366, -0.00000010004009, 0.00000000225401, 0.00000000705803, 0.00000000147195, 0.00000000010388])
        rec_lo = np.array([0.00054882250985, 0.00777663905235, 0.04924777153818, 0.18135962544038, 0.41968794493936, 0.60150609493500, 0.44459045192760, -0.03572291961726, -0.33566408953053, -0.11239707156845, 0.21156452768087, 0.11523329843969, -0.13994042493255, -0.08177594298086, 0.09660039032372, 0.04572340574923, -0.06497750489373, -0.01865385920212, 0.03972683542785, 0.00335775639034, -0.02089205367798, 0.00240347092081, 0.00898882438197, -0.00289133434859, -0.00295837403893, 0.00171660704063, 0.00063941850051, -0.00069067111708, -0.00003196406278, 0.00019366465042, -0.00003635520250, -0.00003499665985, 0.00001535482509, 0.00000279033054, -0.00000309001716, 0.00000031660954, 0.00000029921366, -0.00000010004009, -0.00000000225401, 0.00000000705803, -0.00000000147195, 0.00000000010388])
        rec_hi = np.array([0.00000000010388, 0.00000000147195, 0.00000000705803, 0.00000000225401, -0.00000010004009, -0.00000029921366, 0.00000031660954, 0.00000309001716, 0.00000279033054, -0.00001535482509, -0.00003499665985, 0.00003635520250, 0.00019366465042, 0.00003196406278, -0.00069067111708, -0.00063941850051, 0.00171660704063, 0.00295837403893, -0.00289133434859, -0.00898882438197, 0.00240347092081, 0.02089205367798, 0.00335775639034, -0.03972683542785, -0.01865385920212, 0.06497750489373, 0.04572340574923, -0.09660039032372, -0.08177594298086, 0.13994042493255, 0.11523329843969, -0.21156452768087, -0.11239707156845, 0.33566408953053, -0.03572291961726, -0.44459045192760, 0.60150609493500, -0.41968794493936, 0.18135962544038, -0.04924777153818, 0.00777663905235, -0.00054882250985])
    elif family=="db22":
        dec_lo = np.array([-0.00000000003602, 0.00000000053359, -0.00000000272962, 0.00000000168017, 0.00000003761229, -0.00000012833362, -0.00000008779880, 0.00000129518206, -0.00000156517913, -0.00000616672932, 0.00001737375696, 0.00001137434966, -0.00009405223635, 0.00004345899905, 0.00032860941421, -0.00042378739984, -0.00077069098812, 0.00182701049566, 0.00104426073919, -0.00545569198616, 0.00030013739851, 0.01256472521834, -0.00621378284936, -0.02348000134449, 0.02058670762757, 0.03697084662070, -0.04653081182751, -0.05136425429744, 0.08455737636683, 0.06807631439273, -0.13176813768668, -0.09711079840911, 0.17997318799289, 0.16409318810677, -0.20056840610489, -0.31272658042830, 0.07372450118363, 0.50790109062216, 0.57843273100952, 0.36772868344604, 0.14836754089011, 0.03806993723641, 0.00572185463133, 0.00038626323149])
        dec_hi = np.array([-0.00038626323149, 0.00572185463133, -0.03806993723641, 0.14836754089011, -0.36772868344604, 0.57843273100952, -0.50790109062216, 0.07372450118363, 0.31272658042830, -0.20056840610489, -0.16409318810677, 0.17997318799289, 0.09711079840911, -0.13176813768668, -0.06807631439273, 0.08455737636683, 0.05136425429744, -0.04653081182751, -0.03697084662070, 0.02058670762757, 0.02348000134449, -0.00621378284936, -0.01256472521834, 0.00030013739851, 0.00545569198616, 0.00104426073919, -0.00182701049566, -0.00077069098812, 0.00042378739984, 0.00032860941421, -0.00004345899905, -0.00009405223635, -0.00001137434966, 0.00001737375696, 0.00000616672932, -0.00000156517913, -0.00000129518206, -0.00000008779880, 0.00000012833362, 0.00000003761229, -0.00000000168017, -0.00000000272962, -0.00000000053359, -0.00000000003602])
        rec_lo = np.array([0.00038626323149, 0.00572185463133, 0.03806993723641, 0.14836754089011, 0.36772868344604, 0.57843273100952, 0.50790109062216, 0.07372450118363, -0.31272658042830, -0.20056840610489, 0.16409318810677, 0.17997318799289, -0.09711079840911, -0.13176813768668, 0.06807631439273, 0.08455737636683, -0.05136425429744, -0.04653081182751, 0.03697084662070, 0.02058670762757, -0.02348000134449, -0.00621378284936, 0.01256472521834, 0.00030013739851, -0.00545569198616, 0.00104426073919, 0.00182701049566, -0.00077069098812, -0.00042378739984, 0.00032860941421, 0.00004345899905, -0.00009405223635, 0.00001137434966, 0.00001737375696, -0.00000616672932, -0.00000156517913, 0.00000129518206, -0.00000008779880, -0.00000012833362, 0.00000003761229, 0.00000000168017, -0.00000000272962, 0.00000000053359, -0.00000000003602])
        rec_hi = np.array([-0.00000000003602, -0.00000000053359, -0.00000000272962, -0.00000000168017, 0.00000003761229, 0.00000012833362, -0.00000008779880, -0.00000129518206, -0.00000156517913, 0.00000616672932, 0.00001737375696, -0.00001137434966, -0.00009405223635, -0.00004345899905, 0.00032860941421, 0.00042378739984, -0.00077069098812, -0.00182701049566, 0.00104426073919, 0.00545569198616, 0.00030013739851, -0.01256472521834, -0.00621378284936, 0.02348000134449, 0.02058670762757, -0.03697084662070, -0.04653081182751, 0.05136425429744, 0.08455737636683, -0.06807631439273, -0.13176813768668, 0.09711079840911, 0.17997318799289, -0.16409318810677, -0.20056840610489, 0.31272658042830, 0.07372450118363, -0.50790109062216, 0.57843273100952, -0.36772868344604, 0.14836754089011, -0.03806993723641, 0.00572185463133, -0.00038626323149])
    elif family=="db23":
        dec_lo = np.array([0.00000000001250, -0.00000000019324, 0.00000000105045, -0.00000000094729, -0.00000001399935, 0.00000005417549, 0.00000001853092, -0.00000053390054, 0.00000081475748, 0.00000239756955, -0.00000834787557, -0.00000263520789, 0.00004426071203, -0.00003378894834, -0.00015002185035, 0.00025676245201, 0.00031942049271, -0.00106123122889, -0.00024650140052, 0.00312287644982, -0.00113486547336, -0.00707531927371, 0.00603184065002, 0.01275194393153, -0.01753710100304, -0.01852351365016, 0.03849533252257, 0.02176585683450, -0.07020739157490, -0.02112621235623, 0.11229704361811, 0.02028307457565, -0.16401132153188, -0.03303744709429, 0.22357365824204, 0.09212540708242, -0.27140209860784, -0.26139214803064, 0.18139262536384, 0.55101851724192, 0.54493114787352, 0.31845081385287, 0.12051553178397, 0.02931000365788, 0.00420274889318, 0.00027190419413])
        dec_hi = np.array([-0.00027190419413, 0.00420274889318, -0.02931000365788, 0.12051553178397, -0.31845081385287, 0.54493114787352, -0.55101851724192, 0.18139262536384, 0.26139214803064, -0.27140209860784, -0.09212540708242, 0.22357365824204, 0.03303744709429, -0.16401132153188, -0.02028307457565, 0.11229704361811, 0.02112621235623, -0.07020739157490, -0.02176585683450, 0.03849533252257, 0.01852351365016, -0.01753710100304, -0.01275194393153, 0.00603184065002, 0.00707531927371, -0.00113486547336, -0.00312287644982, -0.00024650140052, 0.00106123122889, 0.00031942049271, -0.00025676245201, -0.00015002185035, 0.00003378894834, 0.00004426071203, 0.00000263520789, -0.00000834787557, -0.00000239756955, 0.00000081475748, 0.00000053390054, 0.00000001853092, -0.00000005417549, -0.00000001399935, 0.00000000094729, 0.00000000105045, 0.00000000019324, 0.00000000001250])
        rec_lo = np.array([0.00027190419413, 0.00420274889318, 0.02931000365788, 0.12051553178397, 0.31845081385287, 0.54493114787352, 0.55101851724192, 0.18139262536384, -0.26139214803064, -0.27140209860784, 0.09212540708242, 0.22357365824204, -0.03303744709429, -0.16401132153188, 0.02028307457565, 0.11229704361811, -0.02112621235623, -0.07020739157490, 0.02176585683450, 0.03849533252257, -0.01852351365016, -0.01753710100304, 0.01275194393153, 0.00603184065002, -0.00707531927371, -0.00113486547336, 0.00312287644982, -0.00024650140052, -0.00106123122889, 0.00031942049271, 0.00025676245201, -0.00015002185035, -0.00003378894834, 0.00004426071203, -0.00000263520789, -0.00000834787557, 0.00000239756955, 0.00000081475748, -0.00000053390054, 0.00000001853092, 0.00000005417549, -0.00000001399935, -0.00000000094729, 0.00000000105045, -0.00000000019324, 0.00000000001250])
        rec_hi = np.array([0.00000000001250, 0.00000000019324, 0.00000000105045, 0.00000000094729, -0.00000001399935, -0.00000005417549, 0.00000001853092, 0.00000053390054, 0.00000081475748, -0.00000239756955, -0.00000834787557, 0.00000263520789, 0.00004426071203, 0.00003378894834, -0.00015002185035, -0.00025676245201, 0.00031942049271, 0.00106123122889, -0.00024650140052, -0.00312287644982, -0.00113486547336, 0.00707531927371, 0.00603184065002, -0.01275194393153, -0.01753710100304, 0.01852351365016, 0.03849533252257, -0.02176585683450, -0.07020739157490, 0.02112621235623, 0.11229704361811, -0.02028307457565, -0.16401132153188, 0.03303744709429, 0.22357365824204, -0.09212540708242, -0.27140209860784, 0.26139214803064, 0.18139262536384, -0.55101851724192, 0.54493114787352, -0.31845081385287, 0.12051553178397, -0.02931000365788, 0.00420274889318, -0.00027190419413])
    elif family=="db24":
        dec_lo = np.array([-0.00000000000434, 0.00000000006992, -0.00000000040247, 0.00000000047484, 0.00000000515778, -0.00000002255740, -0.00000000050576, 0.00000021663397, -0.00000040325078, -0.00000089802531, 0.00000390110034, 0.00000001341158, -0.00002022888293, 0.00002183241460, 0.00006559388639, -0.00014600798178, -0.00011812332380, 0.00058612705932, -0.00004416184856, -0.00169645681897, 0.00115376493684, 0.00373604617828, -0.00474656878632, -0.00629143537002, 0.01304997087109, 0.00766172188165, -0.02821310709490, -0.00494470942813, 0.05130162003998, -0.00457843624182, -0.08216165420800, 0.02098011370914, 0.12101630346922, -0.03877717357792, -0.17117535137035, 0.04252872964148, 0.23923738878031, 0.00477661368434, -0.31794307899936, -0.18727140688516, 0.28098555323371, 0.57493922109554, 0.50437104083993, 0.27290891606773, 0.09726223583363, 0.02248233994972, 0.00308208171491, 0.00019143580095])
        dec_hi = np.array([-0.00019143580095, 0.00308208171491, -0.02248233994972, 0.09726223583363, -0.27290891606773, 0.50437104083993, -0.57493922109554, 0.28098555323371, 0.18727140688516, -0.31794307899936, -0.00477661368434, 0.23923738878031, -0.04252872964148, -0.17117535137035, 0.03877717357792, 0.12101630346922, -0.02098011370914, -0.08216165420800, 0.00457843624182, 0.05130162003998, 0.00494470942813, -0.02821310709490, -0.00766172188165, 0.01304997087109, 0.00629143537002, -0.00474656878632, -0.00373604617828, 0.00115376493684, 0.00169645681897, -0.00004416184856, -0.00058612705932, -0.00011812332380, 0.00014600798178, 0.00006559388639, -0.00002183241460, -0.00002022888293, -0.00000001341158, 0.00000390110034, 0.00000089802531, -0.00000040325078, -0.00000021663397, -0.00000000050576, 0.00000002255740, 0.00000000515778, -0.00000000047484, -0.00000000040247, -0.00000000006992, -0.00000000000434])
        rec_lo = np.array([0.00019143580095, 0.00308208171491, 0.02248233994972, 0.09726223583363, 0.27290891606773, 0.50437104083993, 0.57493922109554, 0.28098555323371, -0.18727140688516, -0.31794307899936, 0.00477661368434, 0.23923738878031, 0.04252872964148, -0.17117535137035, -0.03877717357792, 0.12101630346922, 0.02098011370914, -0.08216165420800, -0.00457843624182, 0.05130162003998, -0.00494470942813, -0.02821310709490, 0.00766172188165, 0.01304997087109, -0.00629143537002, -0.00474656878632, 0.00373604617828, 0.00115376493684, -0.00169645681897, -0.00004416184856, 0.00058612705932, -0.00011812332380, -0.00014600798178, 0.00006559388639, 0.00002183241460, -0.00002022888293, 0.00000001341158, 0.00000390110034, -0.00000089802531, -0.00000040325078, 0.00000021663397, -0.00000000050576, -0.00000002255740, 0.00000000515778, 0.00000000047484, -0.00000000040247, 0.00000000006992, -0.00000000000434])
        rec_hi = np.array([-0.00000000000434, -0.00000000006992, -0.00000000040247, -0.00000000047484, 0.00000000515778, 0.00000002255740, -0.00000000050576, -0.00000021663397, -0.00000040325078, 0.00000089802531, 0.00000390110034, -0.00000001341158, -0.00002022888293, -0.00002183241460, 0.00006559388639, 0.00014600798178, -0.00011812332380, -0.00058612705932, -0.00004416184856, 0.00169645681897, 0.00115376493684, -0.00373604617828, -0.00474656878632, 0.00629143537002, 0.01304997087109, -0.00766172188165, -0.02821310709490, 0.00494470942813, 0.05130162003998, 0.00457843624182, -0.08216165420800, -0.02098011370914, 0.12101630346922, 0.03877717357792, -0.17117535137035, -0.04252872964148, 0.23923738878031, -0.00477661368434, -0.31794307899936, 0.18727140688516, 0.28098555323371, -0.57493922109554, 0.50437104083993, -0.27290891606773, 0.09726223583363, -0.02248233994972, 0.00308208171491, -0.00019143580095])
    elif family=="db25":
        dec_lo = np.array([0.00000000000151, -0.00000000002528, 0.00000000015359, -0.00000000022285, -0.00000000188042, 0.00000000927922, -0.00000000261160, -0.00000008656942, 0.00000019228068, 0.00000032120375, -0.00000177920133, 0.00000052328277, 0.00000899066139, -0.00001277195293, -0.00002733048120, 0.00007904640004, 0.00003543714523, -0.00030988009910, 0.00011532124405, 0.00087725819367, -0.00089997742375, -0.00184248429020, 0.00332270777397, 0.00272693625874, -0.00886070261805, -0.00198942578220, 0.01892280447663, -0.00307983679485, -0.03404232046065, 0.01554260592910, 0.05361790939878, -0.03717396286112, -0.07708411105657, 0.06675216449402, 0.10663380501848, -0.09850861528996, -0.15056021375058, 0.11815528671996, 0.22453781974510, -0.08758761458765, -0.33647307964175, -0.09717464096464, 0.36788507480295, 0.58163689674606, 0.45968341514609, 0.23169350788602, 0.07803586287213, 0.01718674125404, 0.00225695959185, 0.00013480297935])
        dec_hi = np.array([-0.00013480297935, 0.00225695959185, -0.01718674125404, 0.07803586287213, -0.23169350788602, 0.45968341514609, -0.58163689674606, 0.36788507480295, 0.09717464096464, -0.33647307964175, 0.08758761458765, 0.22453781974510, -0.11815528671996, -0.15056021375058, 0.09850861528996, 0.10663380501848, -0.06675216449402, -0.07708411105657, 0.03717396286112, 0.05361790939878, -0.01554260592910, -0.03404232046065, 0.00307983679485, 0.01892280447663, 0.00198942578220, -0.00886070261805, -0.00272693625874, 0.00332270777397, 0.00184248429020, -0.00089997742375, -0.00087725819367, 0.00011532124405, 0.00030988009910, 0.00003543714523, -0.00007904640004, -0.00002733048120, 0.00001277195293, 0.00000899066139, -0.00000052328277, -0.00000177920133, -0.00000032120375, 0.00000019228068, 0.00000008656942, -0.00000000261160, -0.00000000927922, -0.00000000188042, 0.00000000022285, 0.00000000015359, 0.00000000002528, 0.00000000000151])
        rec_lo = np.array([0.00013480297935, 0.00225695959185, 0.01718674125404, 0.07803586287213, 0.23169350788602, 0.45968341514609, 0.58163689674606, 0.36788507480295, -0.09717464096464, -0.33647307964175, -0.08758761458765, 0.22453781974510, 0.11815528671996, -0.15056021375058, -0.09850861528996, 0.10663380501848, 0.06675216449402, -0.07708411105657, -0.03717396286112, 0.05361790939878, 0.01554260592910, -0.03404232046065, -0.00307983679485, 0.01892280447663, -0.00198942578220, -0.00886070261805, 0.00272693625874, 0.00332270777397, -0.00184248429020, -0.00089997742375, 0.00087725819367, 0.00011532124405, -0.00030988009910, 0.00003543714523, 0.00007904640004, -0.00002733048120, -0.00001277195293, 0.00000899066139, 0.00000052328277, -0.00000177920133, 0.00000032120375, 0.00000019228068, -0.00000008656942, -0.00000000261160, 0.00000000927922, -0.00000000188042, -0.00000000022285, 0.00000000015359, -0.00000000002528, 0.00000000000151])
        rec_hi = np.array([0.00000000000151, 0.00000000002528, 0.00000000015359, 0.00000000022285, -0.00000000188042, -0.00000000927922, -0.00000000261160, 0.00000008656942, 0.00000019228068, -0.00000032120375, -0.00000177920133, -0.00000052328277, 0.00000899066139, 0.00001277195293, -0.00002733048120, -0.00007904640004, 0.00003543714523, 0.00030988009910, 0.00011532124405, -0.00087725819367, -0.00089997742375, 0.00184248429020, 0.00332270777397, -0.00272693625874, -0.00886070261805, 0.00198942578220, 0.01892280447663, 0.00307983679485, -0.03404232046065, -0.01554260592910, 0.05361790939878, 0.03717396286112, -0.07708411105657, -0.06675216449402, 0.10663380501848, 0.09850861528996, -0.15056021375058, -0.11815528671996, 0.22453781974510, 0.08758761458765, -0.33647307964175, 0.09717464096464, 0.36788507480295, -0.58163689674606, 0.45968341514609, -0.23169350788602, 0.07803586287213, -0.01718674125404, 0.00225695959185, -0.00013480297935])
    elif family=="db26":
        dec_lo = np.array([-0.00000000000053, 0.00000000000913, -0.00000000005840, 0.00000000010023, 0.00000000067800, -0.00000000377601, 0.00000000216933, 0.00000003407796, -0.00000008904466, -0.00000010790042, 0.00000079392106, -0.00000046504632, -0.00000388740016, 0.00000700007868, 0.00001074221541, -0.00004109673996, -0.00000527779549, 0.00015747952386, -0.00010605747483, -0.00043195570743, 0.00061613822046, 0.00083834880565, -0.00214553028157, -0.00093905825047, 0.00560194723942, -0.00052873839926, -0.01178549790619, 0.00582958055532, 0.02073492017996, -0.01776090356836, -0.03137811036307, 0.03853571597111, 0.04223218579637, -0.06865475960404, -0.05344856168148, 0.10648240524981, 0.06982318611329, -0.14797719327525, -0.10432390028593, 0.18275540958967, 0.18129183231112, -0.17483996128939, -0.32638459369178, 0.00177407678099, 0.43915831178917, 0.57366904303422, 0.41329296227836, 0.19503943871677, 0.06227474402515, 0.01309755429256, 0.00165052023353, 0.00009493795751])
        dec_hi = np.array([-0.00009493795751, 0.00165052023353, -0.01309755429256, 0.06227474402515, -0.19503943871677, 0.41329296227836, -0.57366904303422, 0.43915831178917, -0.00177407678099, -0.32638459369178, 0.17483996128939, 0.18129183231112, -0.18275540958967, -0.10432390028593, 0.14797719327525, 0.06982318611329, -0.10648240524981, -0.05344856168148, 0.06865475960404, 0.04223218579637, -0.03853571597111, -0.03137811036307, 0.01776090356836, 0.02073492017996, -0.00582958055532, -0.01178549790619, 0.00052873839926, 0.00560194723942, 0.00093905825047, -0.00214553028157, -0.00083834880565, 0.00061613822046, 0.00043195570743, -0.00010605747483, -0.00015747952386, -0.00000527779549, 0.00004109673996, 0.00001074221541, -0.00000700007868, -0.00000388740016, 0.00000046504632, 0.00000079392106, 0.00000010790042, -0.00000008904466, -0.00000003407796, 0.00000000216933, 0.00000000377601, 0.00000000067800, -0.00000000010023, -0.00000000005840, -0.00000000000913, -0.00000000000053])
        rec_lo = np.array([0.00009493795751, 0.00165052023353, 0.01309755429256, 0.06227474402515, 0.19503943871677, 0.41329296227836, 0.57366904303422, 0.43915831178917, 0.00177407678099, -0.32638459369178, -0.17483996128939, 0.18129183231112, 0.18275540958967, -0.10432390028593, -0.14797719327525, 0.06982318611329, 0.10648240524981, -0.05344856168148, -0.06865475960404, 0.04223218579637, 0.03853571597111, -0.03137811036307, -0.01776090356836, 0.02073492017996, 0.00582958055532, -0.01178549790619, -0.00052873839926, 0.00560194723942, -0.00093905825047, -0.00214553028157, 0.00083834880565, 0.00061613822046, -0.00043195570743, -0.00010605747483, 0.00015747952386, -0.00000527779549, -0.00004109673996, 0.00001074221541, 0.00000700007868, -0.00000388740016, -0.00000046504632, 0.00000079392106, -0.00000010790042, -0.00000008904466, 0.00000003407796, 0.00000000216933, -0.00000000377601, 0.00000000067800, 0.00000000010023, -0.00000000005840, 0.00000000000913, -0.00000000000053])
        rec_hi = np.array([-0.00000000000053, -0.00000000000913, -0.00000000005840, -0.00000000010023, 0.00000000067800, 0.00000000377601, 0.00000000216933, -0.00000003407796, -0.00000008904466, 0.00000010790042, 0.00000079392106, 0.00000046504632, -0.00000388740016, -0.00000700007868, 0.00001074221541, 0.00004109673996, -0.00000527779549, -0.00015747952386, -0.00010605747483, 0.00043195570743, 0.00061613822046, -0.00083834880565, -0.00214553028157, 0.00093905825047, 0.00560194723942, 0.00052873839926, -0.01178549790619, -0.00582958055532, 0.02073492017996, 0.01776090356836, -0.03137811036307, -0.03853571597111, 0.04223218579637, 0.06865475960404, -0.05344856168148, -0.10648240524981, 0.06982318611329, 0.14797719327525, -0.10432390028593, -0.18275540958967, 0.18129183231112, 0.17483996128939, -0.32638459369178, -0.00177407678099, 0.43915831178917, -0.57366904303422, 0.41329296227836, -0.19503943871677, 0.06227474402515, -0.01309755429256, 0.00165052023353, -0.00009493795751])
    elif family=="db27":
        dec_lo = np.array([0.00000000000018, -0.00000000000330, 0.00000000002214, -0.00000000004375, -0.00000000024155, 0.00000000152161, -0.00000000130947, -0.00000001321332, 0.00000004026255, 0.00000003286559, -0.00000034724681, 0.00000030508807, 0.00000163436962, -0.00000365750091, -0.00000390116407, 0.00002063442648, -0.00000351748361, -0.00007711145518, 0.00007660058387, 0.00020197198797, -0.00038790185741, -0.00034183512269, 0.00130117745024, 0.00014575296259, -0.00333285446952, 0.00134262687730, 0.00685663560968, -0.00586209634546, -0.01157718645898, 0.01566559564892, 0.01614696692240, -0.03273906663102, -0.01851249356200, 0.05796940573472, 0.01731101826549, -0.09102290652957, -0.01406275155581, 0.13119797171716, 0.01579939746024, -0.17803174095901, -0.03878641863180, 0.22727328841417, 0.11482301951779, -0.24826458190326, -0.28971680331459, 0.10284085506182, 0.49340612267800, 0.55384986099048, 0.36711021412539, 0.16292202750239, 0.04945259998290, 0.00995258878088, 0.00120553123167, 0.00006687131385])
        dec_hi = np.array([-0.00006687131385, 0.00120553123167, -0.00995258878088, 0.04945259998290, -0.16292202750239, 0.36711021412539, -0.55384986099048, 0.49340612267800, -0.10284085506182, -0.28971680331459, 0.24826458190326, 0.11482301951779, -0.22727328841417, -0.03878641863180, 0.17803174095901, 0.01579939746024, -0.13119797171716, -0.01406275155581, 0.09102290652957, 0.01731101826549, -0.05796940573472, -0.01851249356200, 0.03273906663102, 0.01614696692240, -0.01566559564892, -0.01157718645898, 0.00586209634546, 0.00685663560968, -0.00134262687730, -0.00333285446952, -0.00014575296259, 0.00130117745024, 0.00034183512269, -0.00038790185741, -0.00020197198797, 0.00007660058387, 0.00007711145518, -0.00000351748361, -0.00002063442648, -0.00000390116407, 0.00000365750091, 0.00000163436962, -0.00000030508807, -0.00000034724681, -0.00000003286559, 0.00000004026255, 0.00000001321332, -0.00000000130947, -0.00000000152161, -0.00000000024155, 0.00000000004375, 0.00000000002214, 0.00000000000330, 0.00000000000018])
        rec_lo = np.array([0.00006687131385, 0.00120553123167, 0.00995258878088, 0.04945259998290, 0.16292202750239, 0.36711021412539, 0.55384986099048, 0.49340612267800, 0.10284085506182, -0.28971680331459, -0.24826458190326, 0.11482301951779, 0.22727328841417, -0.03878641863180, -0.17803174095901, 0.01579939746024, 0.13119797171716, -0.01406275155581, -0.09102290652957, 0.01731101826549, 0.05796940573472, -0.01851249356200, -0.03273906663102, 0.01614696692240, 0.01566559564892, -0.01157718645898, -0.00586209634546, 0.00685663560968, 0.00134262687730, -0.00333285446952, 0.00014575296259, 0.00130117745024, -0.00034183512269, -0.00038790185741, 0.00020197198797, 0.00007660058387, -0.00007711145518, -0.00000351748361, 0.00002063442648, -0.00000390116407, -0.00000365750091, 0.00000163436962, 0.00000030508807, -0.00000034724681, 0.00000003286559, 0.00000004026255, -0.00000001321332, -0.00000000130947, 0.00000000152161, -0.00000000024155, -0.00000000004375, 0.00000000002214, -0.00000000000330, 0.00000000000018])
        rec_hi = np.array([0.00000000000018, 0.00000000000330, 0.00000000002214, 0.00000000004375, -0.00000000024155, -0.00000000152161, -0.00000000130947, 0.00000001321332, 0.00000004026255, -0.00000003286559, -0.00000034724681, -0.00000030508807, 0.00000163436962, 0.00000365750091, -0.00000390116407, -0.00002063442648, -0.00000351748361, 0.00007711145518, 0.00007660058387, -0.00020197198797, -0.00038790185741, 0.00034183512269, 0.00130117745024, -0.00014575296259, -0.00333285446952, -0.00134262687730, 0.00685663560968, 0.00586209634546, -0.01157718645898, -0.01566559564892, 0.01614696692240, 0.03273906663102, -0.01851249356200, -0.05796940573472, 0.01731101826549, 0.09102290652957, -0.01406275155581, -0.13119797171716, 0.01579939746024, 0.17803174095901, -0.03878641863180, -0.22727328841417, 0.11482301951779, 0.24826458190326, -0.28971680331459, -0.10284085506182, 0.49340612267800, -0.55384986099048, 0.36711021412539, -0.16292202750239, 0.04945259998290, -0.00995258878088, 0.00120553123167, -0.00006687131385])
    elif family=="db28":
        dec_lo = np.array([-0.00000000000006, 0.00000000000119, -0.00000000000837, 0.00000000001867, 0.00000000008492, -0.00000000060770, 0.00000000069445, 0.00000000504405, -0.00000001784139, -0.00000000826239, 0.00000014906600, -0.00000017574612, -0.00000066702155, 0.00000184036373, 0.00000124790032, -0.00001004326041, 0.00000463866498, 0.00003641401211, -0.00004907713416, -0.00008903901490, 0.00022957909822, 0.00011546560637, -0.00074867495591, 0.00014156723931, 0.00187599866820, -0.00136037384564, -0.00372546124707, 0.00478486311245, 0.00583881662775, -0.01206359196822, -0.00681554976455, 0.02468806001015, 0.00443173291006, -0.04333336861609, 0.00344801895554, 0.06774789550191, -0.01734192283131, -0.09768535580565, 0.03447863127510, 0.13462756791023, -0.04683823374455, -0.18287733073298, 0.03690688531571, 0.24580815137376, 0.03285787916339, -0.30132780953264, -0.23049895404758, 0.20017614404598, 0.53051629344149, 0.52499823163034, 0.32256336128552, 0.13513791425364, 0.03909260811541, 0.00754265037765, 0.00087949851598, 0.00004710807775])
        dec_hi = np.array([-0.00004710807775, 0.00087949851598, -0.00754265037765, 0.03909260811541, -0.13513791425364, 0.32256336128552, -0.52499823163034, 0.53051629344149, -0.20017614404598, -0.23049895404758, 0.30132780953264, 0.03285787916339, -0.24580815137376, 0.03690688531571, 0.18287733073298, -0.04683823374455, -0.13462756791023, 0.03447863127510, 0.09768535580565, -0.01734192283131, -0.06774789550191, 0.00344801895554, 0.04333336861609, 0.00443173291006, -0.02468806001015, -0.00681554976455, 0.01206359196822, 0.00583881662775, -0.00478486311245, -0.00372546124707, 0.00136037384564, 0.00187599866820, -0.00014156723931, -0.00074867495591, -0.00011546560637, 0.00022957909822, 0.00008903901490, -0.00004907713416, -0.00003641401211, 0.00000463866498, 0.00001004326041, 0.00000124790032, -0.00000184036373, -0.00000066702155, 0.00000017574612, 0.00000014906600, 0.00000000826239, -0.00000001784139, -0.00000000504405, 0.00000000069445, 0.00000000060770, 0.00000000008492, -0.00000000001867, -0.00000000000837, -0.00000000000119, -0.00000000000006])
        rec_lo = np.array([0.00004710807775, 0.00087949851598, 0.00754265037765, 0.03909260811541, 0.13513791425364, 0.32256336128552, 0.52499823163034, 0.53051629344149, 0.20017614404598, -0.23049895404758, -0.30132780953264, 0.03285787916339, 0.24580815137376, 0.03690688531571, -0.18287733073298, -0.04683823374455, 0.13462756791023, 0.03447863127510, -0.09768535580565, -0.01734192283131, 0.06774789550191, 0.00344801895554, -0.04333336861609, 0.00443173291006, 0.02468806001015, -0.00681554976455, -0.01206359196822, 0.00583881662775, 0.00478486311245, -0.00372546124707, -0.00136037384564, 0.00187599866820, 0.00014156723931, -0.00074867495591, 0.00011546560637, 0.00022957909822, -0.00008903901490, -0.00004907713416, 0.00003641401211, 0.00000463866498, -0.00001004326041, 0.00000124790032, 0.00000184036373, -0.00000066702155, -0.00000017574612, 0.00000014906600, -0.00000000826239, -0.00000001784139, 0.00000000504405, 0.00000000069445, -0.00000000060770, 0.00000000008492, 0.00000000001867, -0.00000000000837, 0.00000000000119, -0.00000000000006])
        rec_hi = np.array([-0.00000000000006, -0.00000000000119, -0.00000000000837, -0.00000000001867, 0.00000000008492, 0.00000000060770, 0.00000000069445, -0.00000000504405, -0.00000001784139, 0.00000000826239, 0.00000014906600, 0.00000017574612, -0.00000066702155, -0.00000184036373, 0.00000124790032, 0.00001004326041, 0.00000463866498, -0.00003641401211, -0.00004907713416, 0.00008903901490, 0.00022957909822, -0.00011546560637, -0.00074867495591, -0.00014156723931, 0.00187599866820, 0.00136037384564, -0.00372546124707, -0.00478486311245, 0.00583881662775, 0.01206359196822, -0.00681554976455, -0.02468806001015, 0.00443173291006, 0.04333336861609, 0.00344801895554, -0.06774789550191, -0.01734192283131, 0.09768535580565, 0.03447863127510, -0.13462756791023, -0.04683823374455, 0.18287733073298, 0.03690688531571, -0.24580815137376, 0.03285787916339, 0.30132780953264, -0.23049895404758, -0.20017614404598, 0.53051629344149, -0.52499823163034, 0.32256336128552, -0.13513791425364, 0.03909260811541, -0.00754265037765, 0.00087949851598, -0.00004710807775])
    elif family=="db29":
        dec_lo = np.array([0.00000000000002, -0.00000000000043, 0.00000000000315, -0.00000000000783, -0.00000000002941, 0.00000000024071, -0.00000000034268, -0.00000000189400, 0.00000000776898, 0.00000000107659, -0.00000006286157, 0.00000009387197, 0.00000026338984, -0.00000089757018, -0.00000030290546, 0.00000475060925, -0.00000359364480, -0.00001657328395, 0.00002913344750, 0.00003645026069, -0.00012930448401, -0.00002292018041, 0.00041112834547, -0.00020007113631, -0.00100077832709, 0.00108705394223, 0.00187712092572, -0.00347379898968, -0.00255080712779, 0.00846972549356, 0.00173788033272, -0.01704122457361, 0.00264832730768, 0.02947043187175, -0.01291714255427, -0.04518798127779, 0.03053154327270, 0.06347916458421, -0.05502748952533, -0.08512549261564, 0.08322074716245, 0.11447229589382, -0.10784594993872, -0.16087798859419, 0.11241917487319, 0.23610523615303, -0.05570680007294, -0.33004094891759, -0.15402873445990, 0.28910523833583, 0.55137443275838, 0.48975880476220, 0.28065345597098, 0.11137011695174, 0.03077358022141, 0.00570212651777, 0.00064095168030, 0.00003318966280])
        dec_hi = np.array([-0.00003318966280, 0.00064095168030, -0.00570212651777, 0.03077358022141, -0.11137011695174, 0.28065345597098, -0.48975880476220, 0.55137443275838, -0.28910523833583, -0.15402873445990, 0.33004094891759, -0.05570680007294, -0.23610523615303, 0.11241917487319, 0.16087798859419, -0.10784594993872, -0.11447229589382, 0.08322074716245, 0.08512549261564, -0.05502748952533, -0.06347916458421, 0.03053154327270, 0.04518798127779, -0.01291714255427, -0.02947043187175, 0.00264832730768, 0.01704122457361, 0.00173788033272, -0.00846972549356, -0.00255080712779, 0.00347379898968, 0.00187712092572, -0.00108705394223, -0.00100077832709, 0.00020007113631, 0.00041112834547, 0.00002292018041, -0.00012930448401, -0.00003645026069, 0.00002913344750, 0.00001657328395, -0.00000359364480, -0.00000475060925, -0.00000030290546, 0.00000089757018, 0.00000026338984, -0.00000009387197, -0.00000006286157, -0.00000000107659, 0.00000000776898, 0.00000000189400, -0.00000000034268, -0.00000000024071, -0.00000000002941, 0.00000000000783, 0.00000000000315, 0.00000000000043, 0.00000000000002])
        rec_lo = np.array([0.00003318966280, 0.00064095168030, 0.00570212651777, 0.03077358022141, 0.11137011695174, 0.28065345597098, 0.48975880476220, 0.55137443275838, 0.28910523833583, -0.15402873445990, -0.33004094891759, -0.05570680007294, 0.23610523615303, 0.11241917487319, -0.16087798859419, -0.10784594993872, 0.11447229589382, 0.08322074716245, -0.08512549261564, -0.05502748952533, 0.06347916458421, 0.03053154327270, -0.04518798127779, -0.01291714255427, 0.02947043187175, 0.00264832730768, -0.01704122457361, 0.00173788033272, 0.00846972549356, -0.00255080712779, -0.00347379898968, 0.00187712092572, 0.00108705394223, -0.00100077832709, -0.00020007113631, 0.00041112834547, -0.00002292018041, -0.00012930448401, 0.00003645026069, 0.00002913344750, -0.00001657328395, -0.00000359364480, 0.00000475060925, -0.00000030290546, -0.00000089757018, 0.00000026338984, 0.00000009387197, -0.00000006286157, 0.00000000107659, 0.00000000776898, -0.00000000189400, -0.00000000034268, 0.00000000024071, -0.00000000002941, -0.00000000000783, 0.00000000000315, -0.00000000000043, 0.00000000000002])
        rec_hi = np.array([0.00000000000002, 0.00000000000043, 0.00000000000315, 0.00000000000783, -0.00000000002941, -0.00000000024071, -0.00000000034268, 0.00000000189400, 0.00000000776898, -0.00000000107659, -0.00000006286157, -0.00000009387197, 0.00000026338984, 0.00000089757018, -0.00000030290546, -0.00000475060925, -0.00000359364480, 0.00001657328395, 0.00002913344750, -0.00003645026069, -0.00012930448401, 0.00002292018041, 0.00041112834547, 0.00020007113631, -0.00100077832709, -0.00108705394223, 0.00187712092572, 0.00347379898968, -0.00255080712779, -0.00846972549356, 0.00173788033272, 0.01704122457361, 0.00264832730768, -0.02947043187175, -0.01291714255427, 0.04518798127779, 0.03053154327270, -0.06347916458421, -0.05502748952533, 0.08512549261564, 0.08322074716245, -0.11447229589382, -0.10784594993872, 0.16087798859419, 0.11241917487319, -0.23610523615303, -0.05570680007294, 0.33004094891759, -0.15402873445990, -0.28910523833583, 0.55137443275838, -0.48975880476220, 0.28065345597098, -0.11137011695174, 0.03077358022141, -0.00570212651777, 0.00064095168030, -0.00003318966280])
    elif family=="db30":
        dec_lo = np.array([-0.00000000000001, 0.00000000000015, -0.00000000000119, 0.00000000000324, 0.00000000001000, -0.00000000009461, 0.00000000016136, 0.00000000069849, -0.00000000333111, 0.00000000055534, 0.00000002605443, -0.00000004764380, -0.00000010004147, 0.00000042616623, 0.00000001099474, -0.00000218726768, 0.00000232754910, 0.00000725214554, -0.00001636152479, -0.00001339716863, 0.00006982008371, -0.00000854830547, -0.00021617183012, 0.00017248258424, 0.00050509482390, -0.00076787825044, -0.00086092769681, 0.00232452009406, 0.00084338458666, -0.00553073014819, 0.00061967175650, 0.01091563165830, -0.00529685966613, -0.01839974386812, 0.01528796076986, 0.02707861959529, -0.03226375891935, -0.03567339749676, 0.05671236574474, 0.04380166467142, -0.08765869003638, -0.05380646545826, 0.12274774604501, 0.07277865897036, -0.15723681795999, -0.11455821943271, 0.17782987324484, 0.19946212158066, -0.14196851333008, -0.33296697502086, -0.06618367077594, 0.36624268337163, 0.55757223291284, 0.45048782185332, 0.24202067094021, 0.09123830406702, 0.02413083267159, 0.00430079716505, 0.00046663795043, 0.00002338616173])
        dec_hi = np.array([-0.00002338616173, 0.00046663795043, -0.00430079716505, 0.02413083267159, -0.09123830406702, 0.24202067094021, -0.45048782185332, 0.55757223291284, -0.36624268337163, -0.06618367077594, 0.33296697502086, -0.14196851333008, -0.19946212158066, 0.17782987324484, 0.11455821943271, -0.15723681795999, -0.07277865897036, 0.12274774604501, 0.05380646545826, -0.08765869003638, -0.04380166467142, 0.05671236574474, 0.03567339749676, -0.03226375891935, -0.02707861959529, 0.01528796076986, 0.01839974386812, -0.00529685966613, -0.01091563165830, 0.00061967175650, 0.00553073014819, 0.00084338458666, -0.00232452009406, -0.00086092769681, 0.00076787825044, 0.00050509482390, -0.00017248258424, -0.00021617183012, 0.00000854830547, 0.00006982008371, 0.00001339716863, -0.00001636152479, -0.00000725214554, 0.00000232754910, 0.00000218726768, 0.00000001099474, -0.00000042616623, -0.00000010004147, 0.00000004764380, 0.00000002605443, -0.00000000055534, -0.00000000333111, -0.00000000069849, 0.00000000016136, 0.00000000009461, 0.00000000001000, -0.00000000000324, -0.00000000000119, -0.00000000000015, -0.00000000000001])
        rec_lo = np.array([0.00002338616173, 0.00046663795043, 0.00430079716505, 0.02413083267159, 0.09123830406702, 0.24202067094021, 0.45048782185332, 0.55757223291284, 0.36624268337163, -0.06618367077594, -0.33296697502086, -0.14196851333008, 0.19946212158066, 0.17782987324484, -0.11455821943271, -0.15723681795999, 0.07277865897036, 0.12274774604501, -0.05380646545826, -0.08765869003638, 0.04380166467142, 0.05671236574474, -0.03567339749676, -0.03226375891935, 0.02707861959529, 0.01528796076986, -0.01839974386812, -0.00529685966613, 0.01091563165830, 0.00061967175650, -0.00553073014819, 0.00084338458666, 0.00232452009406, -0.00086092769681, -0.00076787825044, 0.00050509482390, 0.00017248258424, -0.00021617183012, -0.00000854830547, 0.00006982008371, -0.00001339716863, -0.00001636152479, 0.00000725214554, 0.00000232754910, -0.00000218726768, 0.00000001099474, 0.00000042616623, -0.00000010004147, -0.00000004764380, 0.00000002605443, 0.00000000055534, -0.00000000333111, 0.00000000069849, 0.00000000016136, -0.00000000009461, 0.00000000001000, 0.00000000000324, -0.00000000000119, 0.00000000000015, -0.00000000000001])
        rec_hi = np.array([-0.00000000000001, -0.00000000000015, -0.00000000000119, -0.00000000000324, 0.00000000001000, 0.00000000009461, 0.00000000016136, -0.00000000069849, -0.00000000333111, -0.00000000055534, 0.00000002605443, 0.00000004764380, -0.00000010004147, -0.00000042616623, 0.00000001099474, 0.00000218726768, 0.00000232754910, -0.00000725214554, -0.00001636152479, 0.00001339716863, 0.00006982008371, 0.00000854830547, -0.00021617183012, -0.00017248258424, 0.00050509482390, 0.00076787825044, -0.00086092769681, -0.00232452009406, 0.00084338458666, 0.00553073014819, 0.00061967175650, -0.01091563165830, -0.00529685966613, 0.01839974386812, 0.01528796076986, -0.02707861959529, -0.03226375891935, 0.03567339749676, 0.05671236574474, -0.04380166467142, -0.08765869003638, 0.05380646545826, 0.12274774604501, -0.07277865897036, -0.15723681795999, 0.11455821943271, 0.17782987324484, -0.19946212158066, -0.14196851333008, 0.33296697502086, -0.06618367077594, -0.36624268337163, 0.55757223291284, -0.45048782185332, 0.24202067094021, -0.09123830406702, 0.02413083267159, -0.00430079716505, 0.00046663795043, -0.00002338616173])
    elif family=="db31":
        dec_lo = np.array([0.00000000000000, -0.00000000000006, 0.00000000000044, -0.00000000000132, -0.00000000000333, 0.00000000003692, -0.00000000007349, -0.00000000025240, 0.00000000140857, -0.00000000064743, -0.00000001061530, 0.00000002328310, 0.00000003616827, -0.00000019759251, 0.00000005327251, 0.00000098100154, -0.00000136906023, -0.00000303514237, 0.00000879530134, 0.00000403452024, -0.00003631255158, 0.00001501335727, 0.00010895843504, -0.00012434116173, -0.00023965834694, 0.00049988161756, 0.00034313982969, -0.00145904174199, -0.00006397901106, 0.00339306677672, -0.00142826422322, -0.00652085237587, 0.00551616357331, 0.01051763948737, -0.01390055293927, -0.01427627527776, 0.02804761936676, 0.01615417156599, -0.04861907546485, -0.01488002661810, 0.07535361174328, 0.01094129745236, -0.10761277332350, -0.00813983227347, 0.14508950093199, 0.01543698842949, -0.18696236089572, -0.04992634916047, 0.22496671147374, 0.14017828876527, -0.21797848552356, -0.31095511831951, 0.02716921249737, 0.42946880820614, 0.55113984091428, 0.40919220003743, 0.20701287448524, 0.07433609301165, 0.01885369161299, 0.00323688406863, 0.00033941220378, 0.00001648013386])
        dec_hi = np.array([-0.00001648013386, 0.00033941220378, -0.00323688406863, 0.01885369161299, -0.07433609301165, 0.20701287448524, -0.40919220003743, 0.55113984091428, -0.42946880820614, 0.02716921249737, 0.31095511831951, -0.21797848552356, -0.14017828876527, 0.22496671147374, 0.04992634916047, -0.18696236089572, -0.01543698842949, 0.14508950093199, 0.00813983227347, -0.10761277332350, -0.01094129745236, 0.07535361174328, 0.01488002661810, -0.04861907546485, -0.01615417156599, 0.02804761936676, 0.01427627527776, -0.01390055293927, -0.01051763948737, 0.00551616357331, 0.00652085237587, -0.00142826422322, -0.00339306677672, -0.00006397901106, 0.00145904174199, 0.00034313982969, -0.00049988161756, -0.00023965834694, 0.00012434116173, 0.00010895843504, -0.00001501335727, -0.00003631255158, -0.00000403452024, 0.00000879530134, 0.00000303514237, -0.00000136906023, -0.00000098100154, 0.00000005327251, 0.00000019759251, 0.00000003616827, -0.00000002328310, -0.00000001061530, 0.00000000064743, 0.00000000140857, 0.00000000025240, -0.00000000007349, -0.00000000003692, -0.00000000000333, 0.00000000000132, 0.00000000000044, 0.00000000000006, 0.00000000000000])
        rec_lo = np.array([0.00001648013386, 0.00033941220378, 0.00323688406863, 0.01885369161299, 0.07433609301165, 0.20701287448524, 0.40919220003743, 0.55113984091428, 0.42946880820614, 0.02716921249737, -0.31095511831951, -0.21797848552356, 0.14017828876527, 0.22496671147374, -0.04992634916047, -0.18696236089572, 0.01543698842949, 0.14508950093199, -0.00813983227347, -0.10761277332350, 0.01094129745236, 0.07535361174328, -0.01488002661810, -0.04861907546485, 0.01615417156599, 0.02804761936676, -0.01427627527776, -0.01390055293927, 0.01051763948737, 0.00551616357331, -0.00652085237587, -0.00142826422322, 0.00339306677672, -0.00006397901106, -0.00145904174199, 0.00034313982969, 0.00049988161756, -0.00023965834694, -0.00012434116173, 0.00010895843504, 0.00001501335727, -0.00003631255158, 0.00000403452024, 0.00000879530134, -0.00000303514237, -0.00000136906023, 0.00000098100154, 0.00000005327251, -0.00000019759251, 0.00000003616827, 0.00000002328310, -0.00000001061530, -0.00000000064743, 0.00000000140857, -0.00000000025240, -0.00000000007349, 0.00000000003692, -0.00000000000333, -0.00000000000132, 0.00000000000044, -0.00000000000006, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, 0.00000000000006, 0.00000000000044, 0.00000000000132, -0.00000000000333, -0.00000000003692, -0.00000000007349, 0.00000000025240, 0.00000000140857, 0.00000000064743, -0.00000001061530, -0.00000002328310, 0.00000003616827, 0.00000019759251, 0.00000005327251, -0.00000098100154, -0.00000136906023, 0.00000303514237, 0.00000879530134, -0.00000403452024, -0.00003631255158, -0.00001501335727, 0.00010895843504, 0.00012434116173, -0.00023965834694, -0.00049988161756, 0.00034313982969, 0.00145904174199, -0.00006397901106, -0.00339306677672, -0.00142826422322, 0.00652085237587, 0.00551616357331, -0.01051763948737, -0.01390055293927, 0.01427627527776, 0.02804761936676, -0.01615417156599, -0.04861907546485, 0.01488002661810, 0.07535361174328, -0.01094129745236, -0.10761277332350, 0.00813983227347, 0.14508950093199, -0.01543698842949, -0.18696236089572, 0.04992634916047, 0.22496671147374, -0.14017828876527, -0.21797848552356, 0.31095511831951, 0.02716921249737, -0.42946880820614, 0.55113984091428, -0.40919220003743, 0.20701287448524, -0.07433609301165, 0.01885369161299, -0.00323688406863, 0.00033941220378, -0.00001648013386])
    elif family=="db32":
        dec_lo = np.array([-0.00000000000000, 0.00000000000002, -0.00000000000017, 0.00000000000054, 0.00000000000108, -0.00000000001431, 0.00000000003263, 0.00000000008905, -0.00000000058811, 0.00000000043844, 0.00000000425042, -0.00000001104383, -0.00000001219924, 0.00000008965966, -0.00000005003362, -0.00000042859707, 0.00000075600476, 0.00000120288904, -0.00000455830958, -0.00000063617815, 0.00001824268402, -0.00001294045779, -0.00005259809283, 0.00008103678329, 0.00010539154617, -0.00030596544238, -0.00010245373106, 0.00086730585185, -0.00022116787296, -0.00196474055582, 0.00146895510047, 0.00362722464069, -0.00464921675118, -0.00541156825728, 0.01101740071541, 0.00616752731069, -0.02166282283639, -0.00414590766083, 0.03705145792354, -0.00238026446493, -0.05692631406248, 0.01410615151611, 0.08087414063848, -0.02962787250845, -0.10945611311609, 0.04440490819994, 0.14523207947529, -0.04899511718467, -0.19210234470855, 0.02466244483970, 0.24831064235688, 0.06471335480552, -0.27742158155843, -0.26669818147668, 0.12063053826562, 0.47780916373395, 0.53431791934095, 0.36750962859735, 0.17575078363944, 0.06025749912034, 0.01468104638142, 0.00243126191957, 0.00024665669064, 0.00001161463302])
        dec_hi = np.array([-0.00001161463302, 0.00024665669064, -0.00243126191957, 0.01468104638142, -0.06025749912034, 0.17575078363944, -0.36750962859735, 0.53431791934095, -0.47780916373395, 0.12063053826562, 0.26669818147668, -0.27742158155843, -0.06471335480552, 0.24831064235688, -0.02466244483970, -0.19210234470855, 0.04899511718467, 0.14523207947529, -0.04440490819994, -0.10945611311609, 0.02962787250845, 0.08087414063848, -0.01410615151611, -0.05692631406248, 0.00238026446493, 0.03705145792354, 0.00414590766083, -0.02166282283639, -0.00616752731069, 0.01101740071541, 0.00541156825728, -0.00464921675118, -0.00362722464069, 0.00146895510047, 0.00196474055582, -0.00022116787296, -0.00086730585185, -0.00010245373106, 0.00030596544238, 0.00010539154617, -0.00008103678329, -0.00005259809283, 0.00001294045779, 0.00001824268402, 0.00000063617815, -0.00000455830958, -0.00000120288904, 0.00000075600476, 0.00000042859707, -0.00000005003362, -0.00000008965966, -0.00000001219924, 0.00000001104383, 0.00000000425042, -0.00000000043844, -0.00000000058811, -0.00000000008905, 0.00000000003263, 0.00000000001431, 0.00000000000108, -0.00000000000054, -0.00000000000017, -0.00000000000002, -0.00000000000000])
        rec_lo = np.array([0.00001161463302, 0.00024665669064, 0.00243126191957, 0.01468104638142, 0.06025749912034, 0.17575078363944, 0.36750962859735, 0.53431791934095, 0.47780916373395, 0.12063053826562, -0.26669818147668, -0.27742158155843, 0.06471335480552, 0.24831064235688, 0.02466244483970, -0.19210234470855, -0.04899511718467, 0.14523207947529, 0.04440490819994, -0.10945611311609, -0.02962787250845, 0.08087414063848, 0.01410615151611, -0.05692631406248, -0.00238026446493, 0.03705145792354, -0.00414590766083, -0.02166282283639, 0.00616752731069, 0.01101740071541, -0.00541156825728, -0.00464921675118, 0.00362722464069, 0.00146895510047, -0.00196474055582, -0.00022116787296, 0.00086730585185, -0.00010245373106, -0.00030596544238, 0.00010539154617, 0.00008103678329, -0.00005259809283, -0.00001294045779, 0.00001824268402, -0.00000063617815, -0.00000455830958, 0.00000120288904, 0.00000075600476, -0.00000042859707, -0.00000005003362, 0.00000008965966, -0.00000001219924, -0.00000001104383, 0.00000000425042, 0.00000000043844, -0.00000000058811, 0.00000000008905, 0.00000000003263, -0.00000000001431, 0.00000000000108, 0.00000000000054, -0.00000000000017, 0.00000000000002, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, -0.00000000000002, -0.00000000000017, -0.00000000000054, 0.00000000000108, 0.00000000001431, 0.00000000003263, -0.00000000008905, -0.00000000058811, -0.00000000043844, 0.00000000425042, 0.00000001104383, -0.00000001219924, -0.00000008965966, -0.00000005003362, 0.00000042859707, 0.00000075600476, -0.00000120288904, -0.00000455830958, 0.00000063617815, 0.00001824268402, 0.00001294045779, -0.00005259809283, -0.00008103678329, 0.00010539154617, 0.00030596544238, -0.00010245373106, -0.00086730585185, -0.00022116787296, 0.00196474055582, 0.00146895510047, -0.00362722464069, -0.00464921675118, 0.00541156825728, 0.01101740071541, -0.00616752731069, -0.02166282283639, 0.00414590766083, 0.03705145792354, 0.00238026446493, -0.05692631406248, -0.01410615151611, 0.08087414063848, 0.02962787250845, -0.10945611311609, -0.04440490819994, 0.14523207947529, 0.04899511718467, -0.19210234470855, -0.02466244483970, 0.24831064235688, -0.06471335480552, -0.27742158155843, 0.26669818147668, 0.12063053826562, -0.47780916373395, 0.53431791934095, -0.36750962859735, 0.17575078363944, -0.06025749912034, 0.01468104638142, -0.00243126191957, 0.00024665669064, -0.00001161463302])
    elif family=="db33":
        dec_lo = np.array([0.00000000000000, -0.00000000000001, 0.00000000000006, -0.00000000000022, -0.00000000000033, 0.00000000000551, -0.00000000001420, -0.00000000003050, 0.00000000024268, -0.00000000024964, -0.00000000167139, 0.00000000511121, 0.00000000367286, -0.00000003987838, 0.00000003377973, 0.00000018224433, -0.00000039857913, -0.00000044269234, 0.00000228837128, -0.00000036075161, -0.00000886612137, 0.00000907080576, 0.00002423335399, -0.00004929564423, -0.00004160438516, 0.00017804318983, 0.00000439316625, -0.00049083290076, 0.00027273058473, 0.00107438069635, -0.00120430925760, -0.00186071821446, 0.00348080095341, 0.00238906240817, -0.00795354038706, -0.00159428878241, 0.01531695411586, -0.00216775861735, -0.02572876175473, 0.01070326582002, 0.03868706076024, -0.02524858297748, -0.05347125133582, 0.04573456189390, 0.07019114394100, -0.07030248505406, -0.09114696835133, 0.09478808805062, 0.12196785640373, -0.11084413311671, -0.17142809905186, 0.09985155868034, 0.24542061211928, -0.01927833943695, -0.31599741076656, -0.20420262239854, 0.20958235071306, 0.51125477058327, 0.50937617251494, 0.32671813011771, 0.14818631318005, 0.04861466653172, 0.01139594337458, 0.00182270943516, 0.00017910161537, 0.00000818635831])
        dec_hi = np.array([-0.00000818635831, 0.00017910161537, -0.00182270943516, 0.01139594337458, -0.04861466653172, 0.14818631318005, -0.32671813011771, 0.50937617251494, -0.51125477058327, 0.20958235071306, 0.20420262239854, -0.31599741076656, 0.01927833943695, 0.24542061211928, -0.09985155868034, -0.17142809905186, 0.11084413311671, 0.12196785640373, -0.09478808805062, -0.09114696835133, 0.07030248505406, 0.07019114394100, -0.04573456189390, -0.05347125133582, 0.02524858297748, 0.03868706076024, -0.01070326582002, -0.02572876175473, 0.00216775861735, 0.01531695411586, 0.00159428878241, -0.00795354038706, -0.00238906240817, 0.00348080095341, 0.00186071821446, -0.00120430925760, -0.00107438069635, 0.00027273058473, 0.00049083290076, 0.00000439316625, -0.00017804318983, -0.00004160438516, 0.00004929564423, 0.00002423335399, -0.00000907080576, -0.00000886612137, 0.00000036075161, 0.00000228837128, 0.00000044269234, -0.00000039857913, -0.00000018224433, 0.00000003377973, 0.00000003987838, 0.00000000367286, -0.00000000511121, -0.00000000167139, 0.00000000024964, 0.00000000024268, 0.00000000003050, -0.00000000001420, -0.00000000000551, -0.00000000000033, 0.00000000000022, 0.00000000000006, 0.00000000000001, 0.00000000000000])
        rec_lo = np.array([0.00000818635831, 0.00017910161537, 0.00182270943516, 0.01139594337458, 0.04861466653172, 0.14818631318005, 0.32671813011771, 0.50937617251494, 0.51125477058327, 0.20958235071306, -0.20420262239854, -0.31599741076656, -0.01927833943695, 0.24542061211928, 0.09985155868034, -0.17142809905186, -0.11084413311671, 0.12196785640373, 0.09478808805062, -0.09114696835133, -0.07030248505406, 0.07019114394100, 0.04573456189390, -0.05347125133582, -0.02524858297748, 0.03868706076024, 0.01070326582002, -0.02572876175473, -0.00216775861735, 0.01531695411586, -0.00159428878241, -0.00795354038706, 0.00238906240817, 0.00348080095341, -0.00186071821446, -0.00120430925760, 0.00107438069635, 0.00027273058473, -0.00049083290076, 0.00000439316625, 0.00017804318983, -0.00004160438516, -0.00004929564423, 0.00002423335399, 0.00000907080576, -0.00000886612137, -0.00000036075161, 0.00000228837128, -0.00000044269234, -0.00000039857913, 0.00000018224433, 0.00000003377973, -0.00000003987838, 0.00000000367286, 0.00000000511121, -0.00000000167139, -0.00000000024964, 0.00000000024268, -0.00000000003050, -0.00000000001420, 0.00000000000551, -0.00000000000033, -0.00000000000022, 0.00000000000006, -0.00000000000001, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, 0.00000000000001, 0.00000000000006, 0.00000000000022, -0.00000000000033, -0.00000000000551, -0.00000000001420, 0.00000000003050, 0.00000000024268, 0.00000000024964, -0.00000000167139, -0.00000000511121, 0.00000000367286, 0.00000003987838, 0.00000003377973, -0.00000018224433, -0.00000039857913, 0.00000044269234, 0.00000228837128, 0.00000036075161, -0.00000886612137, -0.00000907080576, 0.00002423335399, 0.00004929564423, -0.00004160438516, -0.00017804318983, 0.00000439316625, 0.00049083290076, 0.00027273058473, -0.00107438069635, -0.00120430925760, 0.00186071821446, 0.00348080095341, -0.00238906240817, -0.00795354038706, 0.00159428878241, 0.01531695411586, 0.00216775861735, -0.02572876175473, -0.01070326582002, 0.03868706076024, 0.02524858297748, -0.05347125133582, -0.04573456189390, 0.07019114394100, 0.07030248505406, -0.09114696835133, -0.09478808805062, 0.12196785640373, 0.11084413311671, -0.17142809905186, -0.09985155868034, 0.24542061211928, 0.01927833943695, -0.31599741076656, 0.20420262239854, 0.20958235071306, -0.51125477058327, 0.50937617251494, -0.32671813011771, 0.14818631318005, -0.04861466653172, 0.01139594337458, -0.00182270943516, 0.00017910161537, -0.00000818635831])
    elif family=="db34":
        dec_lo = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000002, 0.00000000000009, 0.00000000000010, -0.00000000000211, 0.00000000000608, 0.00000000001004, -0.00000000009905, 0.00000000013004, 0.00000000064464, -0.00000000231650, -0.00000000086657, 0.00000001740423, -0.00000001990347, -0.00000007526702, 0.00000020259907, 0.00000014481957, -0.00000111630653, 0.00000049797181, 0.00000416987176, -0.00000571082651, -0.00001057657494, 0.00002844951420, 0.00001353117227, -0.00009914697771, 0.00002660050018, 0.00026507723976, -0.00023267321402, -0.00055273557621, 0.00087519990641, 0.00085899598744, -0.00239945394354, -0.00076921279751, 0.00533495076876, -0.00061947488452, -0.01004550670836, 0.00471364926100, 0.01640937419987, -0.01314398001666, -0.02367173792283, 0.02722835075635, 0.03073974657396, -0.04743855964528, -0.03701283841786, 0.07318523543680, 0.04357609464963, -0.10294759699281, -0.05448296806414, 0.13412596027114, 0.07799184693795, -0.16092492717787, -0.12733735822380, 0.16660175041221, 0.21690722018743, -0.10389191551564, -0.33152530150839, -0.12824684217444, 0.29036632950727, 0.53055509965646, 0.47847874627937, 0.28776505923371, 0.12415248211138, 0.03904884135179, 0.00881988940388, 0.00136406139006, 0.00012994762007, 0.00000577051063])
        dec_hi = np.array([-0.00000577051063, 0.00012994762007, -0.00136406139006, 0.00881988940388, -0.03904884135179, 0.12415248211138, -0.28776505923371, 0.47847874627937, -0.53055509965646, 0.29036632950727, 0.12824684217444, -0.33152530150839, 0.10389191551564, 0.21690722018743, -0.16660175041221, -0.12733735822380, 0.16092492717787, 0.07799184693795, -0.13412596027114, -0.05448296806414, 0.10294759699281, 0.04357609464963, -0.07318523543680, -0.03701283841786, 0.04743855964528, 0.03073974657396, -0.02722835075635, -0.02367173792283, 0.01314398001666, 0.01640937419987, -0.00471364926100, -0.01004550670836, 0.00061947488452, 0.00533495076876, 0.00076921279751, -0.00239945394354, -0.00085899598744, 0.00087519990641, 0.00055273557621, -0.00023267321402, -0.00026507723976, 0.00002660050018, 0.00009914697771, 0.00001353117227, -0.00002844951420, -0.00001057657494, 0.00000571082651, 0.00000416987176, -0.00000049797181, -0.00000111630653, -0.00000014481957, 0.00000020259907, 0.00000007526702, -0.00000001990347, -0.00000001740423, -0.00000000086657, 0.00000000231650, 0.00000000064464, -0.00000000013004, -0.00000000009905, -0.00000000001004, 0.00000000000608, 0.00000000000211, 0.00000000000010, -0.00000000000009, -0.00000000000002, -0.00000000000000, -0.00000000000000])
        rec_lo = np.array([0.00000577051063, 0.00012994762007, 0.00136406139006, 0.00881988940388, 0.03904884135179, 0.12415248211138, 0.28776505923371, 0.47847874627937, 0.53055509965646, 0.29036632950727, -0.12824684217444, -0.33152530150839, -0.10389191551564, 0.21690722018743, 0.16660175041221, -0.12733735822380, -0.16092492717787, 0.07799184693795, 0.13412596027114, -0.05448296806414, -0.10294759699281, 0.04357609464963, 0.07318523543680, -0.03701283841786, -0.04743855964528, 0.03073974657396, 0.02722835075635, -0.02367173792283, -0.01314398001666, 0.01640937419987, 0.00471364926100, -0.01004550670836, -0.00061947488452, 0.00533495076876, -0.00076921279751, -0.00239945394354, 0.00085899598744, 0.00087519990641, -0.00055273557621, -0.00023267321402, 0.00026507723976, 0.00002660050018, -0.00009914697771, 0.00001353117227, 0.00002844951420, -0.00001057657494, -0.00000571082651, 0.00000416987176, 0.00000049797181, -0.00000111630653, 0.00000014481957, 0.00000020259907, -0.00000007526702, -0.00000001990347, 0.00000001740423, -0.00000000086657, -0.00000000231650, 0.00000000064464, 0.00000000013004, -0.00000000009905, 0.00000000001004, 0.00000000000608, -0.00000000000211, 0.00000000000010, 0.00000000000009, -0.00000000000002, 0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, -0.00000000000000, -0.00000000000002, -0.00000000000009, 0.00000000000010, 0.00000000000211, 0.00000000000608, -0.00000000001004, -0.00000000009905, -0.00000000013004, 0.00000000064464, 0.00000000231650, -0.00000000086657, -0.00000001740423, -0.00000001990347, 0.00000007526702, 0.00000020259907, -0.00000014481957, -0.00000111630653, -0.00000049797181, 0.00000416987176, 0.00000571082651, -0.00001057657494, -0.00002844951420, 0.00001353117227, 0.00009914697771, 0.00002660050018, -0.00026507723976, -0.00023267321402, 0.00055273557621, 0.00087519990641, -0.00085899598744, -0.00239945394354, 0.00076921279751, 0.00533495076876, 0.00061947488452, -0.01004550670836, -0.00471364926100, 0.01640937419987, 0.01314398001666, -0.02367173792283, -0.02722835075635, 0.03073974657396, 0.04743855964528, -0.03701283841786, -0.07318523543680, 0.04357609464963, 0.10294759699281, -0.05448296806414, -0.13412596027114, 0.07799184693795, 0.16092492717787, -0.12733735822380, -0.16660175041221, 0.21690722018743, 0.10389191551564, -0.33152530150839, 0.12824684217444, 0.29036632950727, -0.53055509965646, 0.47847874627937, -0.28776505923371, 0.12415248211138, -0.03904884135179, 0.00881988940388, -0.00136406139006, 0.00012994762007, -0.00000577051063])
    elif family=="db35":
        dec_lo = np.array([0.00000000000000, -0.00000000000000, 0.00000000000001, -0.00000000000003, -0.00000000000003, 0.00000000000080, -0.00000000000257, -0.00000000000313, 0.00000000004001, -0.00000000006408, -0.00000000024335, 0.00000000103082, 0.00000000005898, -0.00000000745812, 0.00000001084903, 0.00000003008189, -0.00000009990397, -0.00000003700308, 0.00000053023686, -0.00000039039317, -0.00000189592962, 0.00000335334586, 0.00000430804786, -0.00001572442077, -0.00000243700153, 0.00005304143123, -0.00002976995963, -0.00013658830723, 0.00017000122837, 0.00026483288200, -0.00058648103190, -0.00033466921643, 0.00154963746970, 0.00000761596944, -0.00335764438092, 0.00142808879407, 0.00613775458674, -0.00508599164923, -0.00957779789924, 0.01228943600812, 0.01276645671566, -0.02416949780166, -0.01436683978422, 0.04125469306471, 0.01322854958504, -0.06335603744044, -0.00931855894990, 0.08991354757073, 0.00473422917264, -0.12058552264339, -0.00475268083411, 0.15529248039624, 0.01930954466602, -0.19191958929859, -0.06526287131068, 0.21729928932109, 0.16604135749078, -0.18178697676673, -0.32382286491212, -0.04388388187393, 0.36034564051805, 0.53700842750917, 0.44359273922404, 0.25130737899449, 0.10340445586148, 0.03123628851149, 0.00680729288432, 0.00101912268038, 0.00009421469476, 0.00000406793406])
        dec_hi = np.array([-0.00000406793406, 0.00009421469476, -0.00101912268038, 0.00680729288432, -0.03123628851149, 0.10340445586148, -0.25130737899449, 0.44359273922404, -0.53700842750917, 0.36034564051805, 0.04388388187393, -0.32382286491212, 0.18178697676673, 0.16604135749078, -0.21729928932109, -0.06526287131068, 0.19191958929859, 0.01930954466602, -0.15529248039624, -0.00475268083411, 0.12058552264339, 0.00473422917264, -0.08991354757073, -0.00931855894990, 0.06335603744044, 0.01322854958504, -0.04125469306471, -0.01436683978422, 0.02416949780166, 0.01276645671566, -0.01228943600812, -0.00957779789924, 0.00508599164923, 0.00613775458674, -0.00142808879407, -0.00335764438092, -0.00000761596944, 0.00154963746970, 0.00033466921643, -0.00058648103190, -0.00026483288200, 0.00017000122837, 0.00013658830723, -0.00002976995963, -0.00005304143123, -0.00000243700153, 0.00001572442077, 0.00000430804786, -0.00000335334586, -0.00000189592962, 0.00000039039317, 0.00000053023686, 0.00000003700308, -0.00000009990397, -0.00000003008189, 0.00000001084903, 0.00000000745812, 0.00000000005898, -0.00000000103082, -0.00000000024335, 0.00000000006408, 0.00000000004001, 0.00000000000313, -0.00000000000257, -0.00000000000080, -0.00000000000003, 0.00000000000003, 0.00000000000001, 0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000406793406, 0.00009421469476, 0.00101912268038, 0.00680729288432, 0.03123628851149, 0.10340445586148, 0.25130737899449, 0.44359273922404, 0.53700842750917, 0.36034564051805, -0.04388388187393, -0.32382286491212, -0.18178697676673, 0.16604135749078, 0.21729928932109, -0.06526287131068, -0.19191958929859, 0.01930954466602, 0.15529248039624, -0.00475268083411, -0.12058552264339, 0.00473422917264, 0.08991354757073, -0.00931855894990, -0.06335603744044, 0.01322854958504, 0.04125469306471, -0.01436683978422, -0.02416949780166, 0.01276645671566, 0.01228943600812, -0.00957779789924, -0.00508599164923, 0.00613775458674, 0.00142808879407, -0.00335764438092, 0.00000761596944, 0.00154963746970, -0.00033466921643, -0.00058648103190, 0.00026483288200, 0.00017000122837, -0.00013658830723, -0.00002976995963, 0.00005304143123, -0.00000243700153, -0.00001572442077, 0.00000430804786, 0.00000335334586, -0.00000189592962, -0.00000039039317, 0.00000053023686, -0.00000003700308, -0.00000009990397, 0.00000003008189, 0.00000001084903, -0.00000000745812, 0.00000000005898, 0.00000000103082, -0.00000000024335, -0.00000000006408, 0.00000000004001, -0.00000000000313, -0.00000000000257, 0.00000000000080, -0.00000000000003, -0.00000000000003, 0.00000000000001, -0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, 0.00000000000000, 0.00000000000001, 0.00000000000003, -0.00000000000003, -0.00000000000080, -0.00000000000257, 0.00000000000313, 0.00000000004001, 0.00000000006408, -0.00000000024335, -0.00000000103082, 0.00000000005898, 0.00000000745812, 0.00000001084903, -0.00000003008189, -0.00000009990397, 0.00000003700308, 0.00000053023686, 0.00000039039317, -0.00000189592962, -0.00000335334586, 0.00000430804786, 0.00001572442077, -0.00000243700153, -0.00005304143123, -0.00002976995963, 0.00013658830723, 0.00017000122837, -0.00026483288200, -0.00058648103190, 0.00033466921643, 0.00154963746970, -0.00000761596944, -0.00335764438092, -0.00142808879407, 0.00613775458674, 0.00508599164923, -0.00957779789924, -0.01228943600812, 0.01276645671566, 0.02416949780166, -0.01436683978422, -0.04125469306471, 0.01322854958504, 0.06335603744044, -0.00931855894990, -0.08991354757073, 0.00473422917264, 0.12058552264339, -0.00475268083411, -0.15529248039624, 0.01930954466602, 0.19191958929859, -0.06526287131068, -0.21729928932109, 0.16604135749078, 0.18178697676673, -0.32382286491212, 0.04388388187393, 0.36034564051805, -0.53700842750917, 0.44359273922404, -0.25130737899449, 0.10340445586148, -0.03123628851149, 0.00680729288432, -0.00101912268038, 0.00009421469476, -0.00000406793406])
    elif family=="db36":
        dec_lo = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000001, 0.00000000000001, -0.00000000000030, 0.00000000000107, 0.00000000000089, -0.00000000001600, 0.00000000003037, 0.00000000008962, -0.00000000045125, 0.00000000010908, 0.00000000313884, -0.00000000561278, -0.00000001156094, 0.00000004799043, 0.00000000275325, -0.00000024553777, 0.00000025484235, 0.00000083114213, -0.00000187081160, -0.00000158614578, 0.00000837221820, -0.00000118347106, -0.00002731390825, 0.00002375106684, 0.00006694741197, -0.00011318994681, -0.00011551188958, 0.00036935072850, 0.00008614565759, -0.00094634038233, 0.00027768127957, 0.00199079377185, -0.00150307406630, -0.00348454144540, 0.00441348483535, 0.00502298910667, -0.00999026347328, -0.00565781324506, 0.01906359478063, 0.00398404019872, -0.03198072067764, 0.00142497266177, 0.04851308354781, -0.01131910031682, -0.06820901663682, 0.02503872144957, 0.09115678225802, -0.03988085357551, -0.11880375431014, 0.05027618007354, 0.15410623662764, -0.04586140074639, -0.19933720560865, 0.00727851509579, 0.24653727760897, 0.09811420416311, -0.24680703697813, -0.29442103958911, 0.04397519752935, 0.41787533560097, 0.53226689526073, 0.40643369770826, 0.21775695309790, 0.08565209259526, 0.02489056564483, 0.00524029737741, 0.00076021510997, 0.00006826028679, 0.00000286792518])
        dec_hi = np.array([-0.00000286792518, 0.00006826028679, -0.00076021510997, 0.00524029737741, -0.02489056564483, 0.08565209259526, -0.21775695309790, 0.40643369770826, -0.53226689526073, 0.41787533560097, -0.04397519752935, -0.29442103958911, 0.24680703697813, 0.09811420416311, -0.24653727760897, 0.00727851509579, 0.19933720560865, -0.04586140074639, -0.15410623662764, 0.05027618007354, 0.11880375431014, -0.03988085357551, -0.09115678225802, 0.02503872144957, 0.06820901663682, -0.01131910031682, -0.04851308354781, 0.00142497266177, 0.03198072067764, 0.00398404019872, -0.01906359478063, -0.00565781324506, 0.00999026347328, 0.00502298910667, -0.00441348483535, -0.00348454144540, 0.00150307406630, 0.00199079377185, -0.00027768127957, -0.00094634038233, -0.00008614565759, 0.00036935072850, 0.00011551188958, -0.00011318994681, -0.00006694741197, 0.00002375106684, 0.00002731390825, -0.00000118347106, -0.00000837221820, -0.00000158614578, 0.00000187081160, 0.00000083114213, -0.00000025484235, -0.00000024553777, -0.00000000275325, 0.00000004799043, 0.00000001156094, -0.00000000561278, -0.00000000313884, 0.00000000010908, 0.00000000045125, 0.00000000008962, -0.00000000003037, -0.00000000001600, -0.00000000000089, 0.00000000000107, 0.00000000000030, 0.00000000000001, -0.00000000000001, -0.00000000000000, -0.00000000000000, -0.00000000000000])
        rec_lo = np.array([0.00000286792518, 0.00006826028679, 0.00076021510997, 0.00524029737741, 0.02489056564483, 0.08565209259526, 0.21775695309790, 0.40643369770826, 0.53226689526073, 0.41787533560097, 0.04397519752935, -0.29442103958911, -0.24680703697813, 0.09811420416311, 0.24653727760897, 0.00727851509579, -0.19933720560865, -0.04586140074639, 0.15410623662764, 0.05027618007354, -0.11880375431014, -0.03988085357551, 0.09115678225802, 0.02503872144957, -0.06820901663682, -0.01131910031682, 0.04851308354781, 0.00142497266177, -0.03198072067764, 0.00398404019872, 0.01906359478063, -0.00565781324506, -0.00999026347328, 0.00502298910667, 0.00441348483535, -0.00348454144540, -0.00150307406630, 0.00199079377185, 0.00027768127957, -0.00094634038233, 0.00008614565759, 0.00036935072850, -0.00011551188958, -0.00011318994681, 0.00006694741197, 0.00002375106684, -0.00002731390825, -0.00000118347106, 0.00000837221820, -0.00000158614578, -0.00000187081160, 0.00000083114213, 0.00000025484235, -0.00000024553777, 0.00000000275325, 0.00000004799043, -0.00000001156094, -0.00000000561278, 0.00000000313884, 0.00000000010908, -0.00000000045125, 0.00000000008962, 0.00000000003037, -0.00000000001600, 0.00000000000089, 0.00000000000107, -0.00000000000030, 0.00000000000001, 0.00000000000001, -0.00000000000000, 0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, -0.00000000000000, -0.00000000000000, -0.00000000000001, 0.00000000000001, 0.00000000000030, 0.00000000000107, -0.00000000000089, -0.00000000001600, -0.00000000003037, 0.00000000008962, 0.00000000045125, 0.00000000010908, -0.00000000313884, -0.00000000561278, 0.00000001156094, 0.00000004799043, -0.00000000275325, -0.00000024553777, -0.00000025484235, 0.00000083114213, 0.00000187081160, -0.00000158614578, -0.00000837221820, -0.00000118347106, 0.00002731390825, 0.00002375106684, -0.00006694741197, -0.00011318994681, 0.00011551188958, 0.00036935072850, -0.00008614565759, -0.00094634038233, -0.00027768127957, 0.00199079377185, 0.00150307406630, -0.00348454144540, -0.00441348483535, 0.00502298910667, 0.00999026347328, -0.00565781324506, -0.01906359478063, 0.00398404019872, 0.03198072067764, 0.00142497266177, -0.04851308354781, -0.01131910031682, 0.06820901663682, 0.02503872144957, -0.09115678225802, -0.03988085357551, 0.11880375431014, 0.05027618007354, -0.15410623662764, -0.04586140074639, 0.19933720560865, 0.00727851509579, -0.24653727760897, 0.09811420416311, 0.24680703697813, -0.29442103958911, -0.04397519752935, 0.41787533560097, -0.53226689526073, 0.40643369770826, -0.21775695309790, 0.08565209259526, -0.02489056564483, 0.00524029737741, -0.00076021510997, 0.00006826028679, -0.00000286792518])
    elif family=="db37":
        dec_lo = np.array([0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000001, -0.00000000000000, 0.00000000000011, -0.00000000000044, -0.00000000000021, 0.00000000000633, -0.00000000001398, -0.00000000003203, 0.00000000019462, -0.00000000010314, -0.00000000129721, 0.00000000279397, 0.00000000422449, -0.00000002252194, 0.00000000535066, 0.00000011090312, -0.00000015098854, -0.00000034949486, 0.00000100212140, 0.00000048547314, -0.00000430994156, 0.00000184994500, 0.00001354327718, -0.00001639162496, -0.00003098662928, 0.00007055138782, 0.00004336726126, -0.00022089440325, 0.00001534439023, 0.00054905327734, -0.00032807884709, -0.00111148486532, 0.00126393425812, 0.00181687134380, -0.00339452327641, -0.00224805318700, 0.00738775745286, 0.00151930577883, -0.01376398196289, 0.00169047238348, 0.02261865154460, -0.00883349389041, -0.03352358406410, 0.02097280059260, 0.04580794415127, -0.03825382947938, -0.05925681563266, 0.05956741087153, 0.07504761994836, -0.08233021190656, -0.09660754061668, 0.10178029683881, 0.12992964695985, -0.10845171382330, -0.18196229177861, 0.08180602838722, 0.25152325436027, 0.01967150045236, -0.29437591526266, -0.24618042976108, 0.13087896323302, 0.46220755366161, 0.51816704085562, 0.36844097240031, 0.18732633186206, 0.07058482597718, 0.01976228615388, 0.00402414036826, 0.00056624183771, 0.00004942343751, 0.00000202206086])
        dec_hi = np.array([-0.00000202206086, 0.00004942343751, -0.00056624183771, 0.00402414036826, -0.01976228615388, 0.07058482597718, -0.18732633186206, 0.36844097240031, -0.51816704085562, 0.46220755366161, -0.13087896323302, -0.24618042976108, 0.29437591526266, 0.01967150045236, -0.25152325436027, 0.08180602838722, 0.18196229177861, -0.10845171382330, -0.12992964695985, 0.10178029683881, 0.09660754061668, -0.08233021190656, -0.07504761994836, 0.05956741087153, 0.05925681563266, -0.03825382947938, -0.04580794415127, 0.02097280059260, 0.03352358406410, -0.00883349389041, -0.02261865154460, 0.00169047238348, 0.01376398196289, 0.00151930577883, -0.00738775745286, -0.00224805318700, 0.00339452327641, 0.00181687134380, -0.00126393425812, -0.00111148486532, 0.00032807884709, 0.00054905327734, -0.00001534439023, -0.00022089440325, -0.00004336726126, 0.00007055138782, 0.00003098662928, -0.00001639162496, -0.00001354327718, 0.00000184994500, 0.00000430994156, 0.00000048547314, -0.00000100212140, -0.00000034949486, 0.00000015098854, 0.00000011090312, -0.00000000535066, -0.00000002252194, -0.00000000422449, 0.00000000279397, 0.00000000129721, -0.00000000010314, -0.00000000019462, -0.00000000003203, 0.00000000001398, 0.00000000000633, 0.00000000000021, -0.00000000000044, -0.00000000000011, -0.00000000000000, 0.00000000000001, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000202206086, 0.00004942343751, 0.00056624183771, 0.00402414036826, 0.01976228615388, 0.07058482597718, 0.18732633186206, 0.36844097240031, 0.51816704085562, 0.46220755366161, 0.13087896323302, -0.24618042976108, -0.29437591526266, 0.01967150045236, 0.25152325436027, 0.08180602838722, -0.18196229177861, -0.10845171382330, 0.12992964695985, 0.10178029683881, -0.09660754061668, -0.08233021190656, 0.07504761994836, 0.05956741087153, -0.05925681563266, -0.03825382947938, 0.04580794415127, 0.02097280059260, -0.03352358406410, -0.00883349389041, 0.02261865154460, 0.00169047238348, -0.01376398196289, 0.00151930577883, 0.00738775745286, -0.00224805318700, -0.00339452327641, 0.00181687134380, 0.00126393425812, -0.00111148486532, -0.00032807884709, 0.00054905327734, 0.00001534439023, -0.00022089440325, 0.00004336726126, 0.00007055138782, -0.00003098662928, -0.00001639162496, 0.00001354327718, 0.00000184994500, -0.00000430994156, 0.00000048547314, 0.00000100212140, -0.00000034949486, -0.00000015098854, 0.00000011090312, 0.00000000535066, -0.00000002252194, 0.00000000422449, 0.00000000279397, -0.00000000129721, -0.00000000010314, 0.00000000019462, -0.00000000003203, -0.00000000001398, 0.00000000000633, -0.00000000000021, -0.00000000000044, 0.00000000000011, -0.00000000000000, -0.00000000000001, 0.00000000000000, -0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000001, -0.00000000000000, -0.00000000000011, -0.00000000000044, 0.00000000000021, 0.00000000000633, 0.00000000001398, -0.00000000003203, -0.00000000019462, -0.00000000010314, 0.00000000129721, 0.00000000279397, -0.00000000422449, -0.00000002252194, -0.00000000535066, 0.00000011090312, 0.00000015098854, -0.00000034949486, -0.00000100212140, 0.00000048547314, 0.00000430994156, 0.00000184994500, -0.00001354327718, -0.00001639162496, 0.00003098662928, 0.00007055138782, -0.00004336726126, -0.00022089440325, -0.00001534439023, 0.00054905327734, 0.00032807884709, -0.00111148486532, -0.00126393425812, 0.00181687134380, 0.00339452327641, -0.00224805318700, -0.00738775745286, 0.00151930577883, 0.01376398196289, 0.00169047238348, -0.02261865154460, -0.00883349389041, 0.03352358406410, 0.02097280059260, -0.04580794415127, -0.03825382947938, 0.05925681563266, 0.05956741087153, -0.07504761994836, -0.08233021190656, 0.09660754061668, 0.10178029683881, -0.12992964695985, -0.10845171382330, 0.18196229177861, 0.08180602838722, -0.25152325436027, 0.01967150045236, 0.29437591526266, -0.24618042976108, -0.13087896323302, 0.46220755366161, -0.51816704085562, 0.36844097240031, -0.18732633186206, 0.07058482597718, -0.01976228615388, 0.00402414036826, -0.00056624183771, 0.00004942343751, -0.00000202206086])
    elif family=="db38":
        dec_lo = np.array([-0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, -0.00000000000004, 0.00000000000018, 0.00000000000003, -0.00000000000248, 0.00000000000629, 0.00000000001102, -0.00000000008278, 0.00000000006732, 0.00000000052611, -0.00000000134920, -0.00000000143633, 0.00000001034705, -0.00000000542427, -0.00000004884758, 0.00000008400351, 0.00000013963775, -0.00000051877337, -0.00000008487088, 0.00000214996027, -0.00000155084435, -0.00000645673043, 0.00001037359184, 0.00001334176150, -0.00004175141649, -0.00001155409104, 0.00012620433502, -0.00004555682697, -0.00030310204607, 0.00028176392504, 0.00058107597505, -0.00094246140772, -0.00084486266655, 0.00240069778189, 0.00071698218211, -0.00507131450922, 0.00056257157484, 0.00921478503220, -0.00413130665603, -0.01470188206540, 0.01129049727869, 0.02090464525566, -0.02311413402055, -0.02689149388089, 0.04005498110512, 0.03198987753154, -0.06176620870841, -0.03660510340287, 0.08720439826204, 0.04309589543305, -0.11473117071074, -0.05658645863073, 0.14141473407338, 0.08563812155615, -0.15991256515824, -0.14179568597306, 0.14998511961872, 0.23212596383535, -0.06226650604782, -0.32167563780900, -0.18286766770834, 0.21305057135558, 0.49335607851710, 0.49659117531172, 0.33077578141101, 0.16007199356411, 0.05788994361286, 0.01563724934757, 0.00308308811925, 0.00042117026647, 0.00003576251994, 0.00000142577664])
        dec_hi = np.array([-0.00000142577664, 0.00003576251994, -0.00042117026647, 0.00308308811925, -0.01563724934757, 0.05788994361286, -0.16007199356411, 0.33077578141101, -0.49659117531172, 0.49335607851710, -0.21305057135558, -0.18286766770834, 0.32167563780900, -0.06226650604782, -0.23212596383535, 0.14998511961872, 0.14179568597306, -0.15991256515824, -0.08563812155615, 0.14141473407338, 0.05658645863073, -0.11473117071074, -0.04309589543305, 0.08720439826204, 0.03660510340287, -0.06176620870841, -0.03198987753154, 0.04005498110512, 0.02689149388089, -0.02311413402055, -0.02090464525566, 0.01129049727869, 0.01470188206540, -0.00413130665603, -0.00921478503220, 0.00056257157484, 0.00507131450922, 0.00071698218211, -0.00240069778189, -0.00084486266655, 0.00094246140772, 0.00058107597505, -0.00028176392504, -0.00030310204607, 0.00004555682697, 0.00012620433502, 0.00001155409104, -0.00004175141649, -0.00001334176150, 0.00001037359184, 0.00000645673043, -0.00000155084435, -0.00000214996027, -0.00000008487088, 0.00000051877337, 0.00000013963775, -0.00000008400351, -0.00000004884758, 0.00000000542427, 0.00000001034705, 0.00000000143633, -0.00000000134920, -0.00000000052611, 0.00000000006732, 0.00000000008278, 0.00000000001102, -0.00000000000629, -0.00000000000248, -0.00000000000003, 0.00000000000018, 0.00000000000004, -0.00000000000000, -0.00000000000000, -0.00000000000000, -0.00000000000000, -0.00000000000000])
        rec_lo = np.array([0.00000142577664, 0.00003576251994, 0.00042117026647, 0.00308308811925, 0.01563724934757, 0.05788994361286, 0.16007199356411, 0.33077578141101, 0.49659117531172, 0.49335607851710, 0.21305057135558, -0.18286766770834, -0.32167563780900, -0.06226650604782, 0.23212596383535, 0.14998511961872, -0.14179568597306, -0.15991256515824, 0.08563812155615, 0.14141473407338, -0.05658645863073, -0.11473117071074, 0.04309589543305, 0.08720439826204, -0.03660510340287, -0.06176620870841, 0.03198987753154, 0.04005498110512, -0.02689149388089, -0.02311413402055, 0.02090464525566, 0.01129049727869, -0.01470188206540, -0.00413130665603, 0.00921478503220, 0.00056257157484, -0.00507131450922, 0.00071698218211, 0.00240069778189, -0.00084486266655, -0.00094246140772, 0.00058107597505, 0.00028176392504, -0.00030310204607, -0.00004555682697, 0.00012620433502, -0.00001155409104, -0.00004175141649, 0.00001334176150, 0.00001037359184, -0.00000645673043, -0.00000155084435, 0.00000214996027, -0.00000008487088, -0.00000051877337, 0.00000013963775, 0.00000008400351, -0.00000004884758, -0.00000000542427, 0.00000001034705, -0.00000000143633, -0.00000000134920, 0.00000000052611, 0.00000000006732, -0.00000000008278, 0.00000000001102, 0.00000000000629, -0.00000000000248, 0.00000000000003, 0.00000000000018, -0.00000000000004, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000])
        rec_hi = np.array([-0.00000000000000, -0.00000000000000, -0.00000000000000, -0.00000000000000, -0.00000000000000, 0.00000000000004, 0.00000000000018, -0.00000000000003, -0.00000000000248, -0.00000000000629, 0.00000000001102, 0.00000000008278, 0.00000000006732, -0.00000000052611, -0.00000000134920, 0.00000000143633, 0.00000001034705, 0.00000000542427, -0.00000004884758, -0.00000008400351, 0.00000013963775, 0.00000051877337, -0.00000008487088, -0.00000214996027, -0.00000155084435, 0.00000645673043, 0.00001037359184, -0.00001334176150, -0.00004175141649, 0.00001155409104, 0.00012620433502, 0.00004555682697, -0.00030310204607, -0.00028176392504, 0.00058107597505, 0.00094246140772, -0.00084486266655, -0.00240069778189, 0.00071698218211, 0.00507131450922, 0.00056257157484, -0.00921478503220, -0.00413130665603, 0.01470188206540, 0.01129049727869, -0.02090464525566, -0.02311413402055, 0.02689149388089, 0.04005498110512, -0.03198987753154, -0.06176620870841, 0.03660510340287, 0.08720439826204, -0.04309589543305, -0.11473117071074, 0.05658645863073, 0.14141473407338, -0.08563812155615, -0.15991256515824, 0.14179568597306, 0.14998511961872, -0.23212596383535, -0.06226650604782, 0.32167563780900, -0.18286766770834, -0.21305057135558, 0.49335607851710, -0.49659117531172, 0.33077578141101, -0.16007199356411, 0.05788994361286, -0.01563724934757, 0.00308308811925, -0.00042117026647, 0.00003576251994, -0.00000142577664])
    elif family=="dmey":
        dec_lo = np.array([0.00000000000000, -0.00000000000101, 0.00000000851946, -0.00000001111945, -0.00000001079882, 0.00000006066976, -0.00000010866517, 0.00000008200681, 0.00000011783004, -0.00000055063406, 0.00000113079470, -0.00000148954922, 0.00000073675729, 0.00000320544191, -0.00001631269973, 0.00006554305931, -0.00060115023435, -0.00270467212464, 0.00220253410091, 0.00604581409732, -0.00638771831850, -0.01106149639251, 0.01527001513093, 0.01742343410373, -0.03213079399021, -0.02434874590608, 0.06373902432280, 0.03065509196082, -0.13284520043623, -0.03508755565626, 0.44459300275758, 0.74458559231881, 0.44459300275758, -0.03508755565626, -0.13284520043623, 0.03065509196082, 0.06373902432280, -0.02434874590608, -0.03213079399021, 0.01742343410373, 0.01527001513093, -0.01106149639251, -0.00638771831850, 0.00604581409732, 0.00220253410091, -0.00270467212464, -0.00060115023435, 0.00006554305931, -0.00001631269973, 0.00000320544191, 0.00000073675729, -0.00000148954922, 0.00000113079470, -0.00000055063406, 0.00000011783004, 0.00000008200681, -0.00000010866517, 0.00000006066976, -0.00000001079882, -0.00000001111945, 0.00000000851946, -0.00000000000101])
        dec_hi = np.array([0.00000000000101, 0.00000000851946, 0.00000001111945, -0.00000001079882, -0.00000006066976, -0.00000010866517, -0.00000008200681, 0.00000011783004, 0.00000055063406, 0.00000113079470, 0.00000148954922, 0.00000073675729, -0.00000320544191, -0.00001631269973, -0.00006554305931, -0.00060115023435, 0.00270467212464, 0.00220253410091, -0.00604581409732, -0.00638771831850, 0.01106149639251, 0.01527001513093, -0.01742343410373, -0.03213079399021, 0.02434874590608, 0.06373902432280, -0.03065509196082, -0.13284520043623, 0.03508755565626, 0.44459300275758, -0.74458559231881, 0.44459300275758, 0.03508755565626, -0.13284520043623, -0.03065509196082, 0.06373902432280, 0.02434874590608, -0.03213079399021, -0.01742343410373, 0.01527001513093, 0.01106149639251, -0.00638771831850, -0.00604581409732, 0.00220253410091, 0.00270467212464, -0.00060115023435, -0.00006554305931, -0.00001631269973, -0.00000320544191, 0.00000073675729, 0.00000148954922, 0.00000113079470, 0.00000055063406, 0.00000011783004, -0.00000008200681, -0.00000010866517, -0.00000006066976, -0.00000001079882, 0.00000001111945, 0.00000000851946, 0.00000000000101, 0.00000000000000])
        rec_lo = np.array([-0.00000000000101, 0.00000000851946, -0.00000001111945, -0.00000001079882, 0.00000006066976, -0.00000010866517, 0.00000008200681, 0.00000011783004, -0.00000055063406, 0.00000113079470, -0.00000148954922, 0.00000073675729, 0.00000320544191, -0.00001631269973, 0.00006554305931, -0.00060115023435, -0.00270467212464, 0.00220253410091, 0.00604581409732, -0.00638771831850, -0.01106149639251, 0.01527001513093, 0.01742343410373, -0.03213079399021, -0.02434874590608, 0.06373902432280, 0.03065509196082, -0.13284520043623, -0.03508755565626, 0.44459300275758, 0.74458559231881, 0.44459300275758, -0.03508755565626, -0.13284520043623, 0.03065509196082, 0.06373902432280, -0.02434874590608, -0.03213079399021, 0.01742343410373, 0.01527001513093, -0.01106149639251, -0.00638771831850, 0.00604581409732, 0.00220253410091, -0.00270467212464, -0.00060115023435, 0.00006554305931, -0.00001631269973, 0.00000320544191, 0.00000073675729, -0.00000148954922, 0.00000113079470, -0.00000055063406, 0.00000011783004, 0.00000008200681, -0.00000010866517, 0.00000006066976, -0.00000001079882, -0.00000001111945, 0.00000000851946, -0.00000000000101, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, 0.00000000000101, 0.00000000851946, 0.00000001111945, -0.00000001079882, -0.00000006066976, -0.00000010866517, -0.00000008200681, 0.00000011783004, 0.00000055063406, 0.00000113079470, 0.00000148954922, 0.00000073675729, -0.00000320544191, -0.00001631269973, -0.00006554305931, -0.00060115023435, 0.00270467212464, 0.00220253410091, -0.00604581409732, -0.00638771831850, 0.01106149639251, 0.01527001513093, -0.01742343410373, -0.03213079399021, 0.02434874590608, 0.06373902432280, -0.03065509196082, -0.13284520043623, 0.03508755565626, 0.44459300275758, -0.74458559231881, 0.44459300275758, 0.03508755565626, -0.13284520043623, -0.03065509196082, 0.06373902432280, 0.02434874590608, -0.03213079399021, -0.01742343410373, 0.01527001513093, 0.01106149639251, -0.00638771831850, -0.00604581409732, 0.00220253410091, 0.00270467212464, -0.00060115023435, -0.00006554305931, -0.00001631269973, -0.00000320544191, 0.00000073675729, 0.00000148954922, 0.00000113079470, 0.00000055063406, 0.00000011783004, -0.00000008200681, -0.00000010866517, -0.00000006066976, -0.00000001079882, 0.00000001111945, 0.00000000851946, 0.00000000000101])
    elif family=="haar":
        dec_lo = np.array([0.70710678118655, 0.70710678118655])
        dec_hi = np.array([-0.70710678118655, 0.70710678118655])
        rec_lo = np.array([0.70710678118655, 0.70710678118655])
        rec_hi = np.array([0.70710678118655, -0.70710678118655])
    elif family=="rbio1.1":
        dec_lo = np.array([0.70710678118655, 0.70710678118655])
        dec_hi = np.array([-0.70710678118655, 0.70710678118655])
        rec_lo = np.array([0.70710678118655, 0.70710678118655])
        rec_hi = np.array([0.70710678118655, -0.70710678118655])
    elif family=="rbio1.3":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.70710678118655, 0.70710678118655, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([0.08838834764832, 0.08838834764832, -0.70710678118655, 0.70710678118655, -0.08838834764832, -0.08838834764832])
        rec_lo = np.array([-0.08838834764832, 0.08838834764832, 0.70710678118655, 0.70710678118655, 0.08838834764832, -0.08838834764832])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.70710678118655, -0.70710678118655, 0.00000000000000, -0.00000000000000])
    elif family=="rbio1.5":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.70710678118655, 0.70710678118655, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([-0.01657281518406, -0.01657281518406, 0.12153397801644, 0.12153397801644, -0.70710678118655, 0.70710678118655, -0.12153397801644, -0.12153397801644, 0.01657281518406, 0.01657281518406])
        rec_lo = np.array([0.01657281518406, -0.01657281518406, -0.12153397801644, 0.12153397801644, 0.70710678118655, 0.70710678118655, 0.12153397801644, -0.12153397801644, -0.01657281518406, 0.01657281518406])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.70710678118655, -0.70710678118655, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000])
    elif family=="rbio2.2":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.35355339059327, 0.70710678118655, 0.35355339059327, 0.00000000000000])
        dec_hi = np.array([0.17677669529664, 0.35355339059327, -1.06066017177982, 0.35355339059327, 0.17677669529664, 0.00000000000000])
        rec_lo = np.array([-0.17677669529664, 0.35355339059327, 1.06066017177982, 0.35355339059327, -0.17677669529664, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.35355339059327, -0.70710678118655, 0.35355339059327, -0.00000000000000])
    elif family=="rbio2.4":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.35355339059327, 0.70710678118655, 0.35355339059327, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([-0.03314563036812, -0.06629126073624, 0.17677669529664, 0.41984465132951, -0.99436891104358, 0.41984465132951, 0.17677669529664, -0.06629126073624, -0.03314563036812, 0.00000000000000])
        rec_lo = np.array([0.03314563036812, -0.06629126073624, -0.17677669529664, 0.41984465132951, 0.99436891104358, 0.41984465132951, -0.17677669529664, -0.06629126073624, 0.03314563036812, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.35355339059327, -0.70710678118655, 0.35355339059327, -0.00000000000000, 0.00000000000000, -0.00000000000000])
    elif family=="rbio2.6":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.35355339059327, 0.70710678118655, 0.35355339059327, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([0.00690533966002, 0.01381067932005, -0.04695630968817, -0.10772329869639, 0.16987135563661, 0.44746600996961, -0.96674755240348, 0.44746600996961, 0.16987135563661, -0.10772329869639, -0.04695630968817, 0.01381067932005, 0.00690533966002, 0.00000000000000])
        rec_lo = np.array([-0.00690533966002, 0.01381067932005, 0.04695630968817, -0.10772329869639, -0.16987135563661, 0.44746600996961, 0.96674755240348, 0.44746600996961, -0.16987135563661, -0.10772329869639, 0.04695630968817, 0.01381067932005, -0.00690533966002, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.35355339059327, -0.70710678118655, 0.35355339059327, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000])
    elif family=="rbio2.8":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.35355339059327, 0.70710678118655, 0.35355339059327, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([-0.00151054305063, -0.00302108610126, 0.01294751186255, 0.02891610982635, -0.05299848189069, -0.13491307360774, 0.16382918343409, 0.46257144047592, -0.95164212189718, 0.46257144047592, 0.16382918343409, -0.13491307360774, -0.05299848189069, 0.02891610982635, 0.01294751186255, -0.00302108610126, -0.00151054305063, 0.00000000000000])
        rec_lo = np.array([0.00151054305063, -0.00302108610126, -0.01294751186255, 0.02891610982635, 0.05299848189069, -0.13491307360774, -0.16382918343409, 0.46257144047592, 0.95164212189718, 0.46257144047592, -0.16382918343409, -0.13491307360774, 0.05299848189069, 0.02891610982635, -0.01294751186255, -0.00302108610126, 0.00151054305063, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.35355339059327, -0.70710678118655, 0.35355339059327, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000])
    elif family=="rbio3.1":
        dec_lo = np.array([0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664])
        dec_hi = np.array([0.35355339059327, 1.06066017177982, -1.06066017177982, -0.35355339059327])
        rec_lo = np.array([-0.35355339059327, 1.06066017177982, 1.06066017177982, -0.35355339059327])
        rec_hi = np.array([0.17677669529664, -0.53033008588991, 0.53033008588991, -0.17677669529664])
    elif family=="rbio3.3":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([-0.06629126073624, -0.19887378220872, 0.15467960838456, 0.99436891104358, -0.99436891104358, -0.15467960838456, 0.19887378220872, 0.06629126073624])
        rec_lo = np.array([0.06629126073624, -0.19887378220872, -0.15467960838456, 0.99436891104358, 0.99436891104358, -0.15467960838456, -0.19887378220872, 0.06629126073624])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.17677669529664, -0.53033008588991, 0.53033008588991, -0.17677669529664, 0.00000000000000, -0.00000000000000])
    elif family=="rbio3.5":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([0.01381067932005, 0.04143203796015, -0.05248058141619, -0.26792717880897, 0.07181553246426, 0.96674755240348, -0.96674755240348, -0.07181553246426, 0.26792717880897, 0.05248058141619, -0.04143203796015, -0.01381067932005])
        rec_lo = np.array([-0.01381067932005, 0.04143203796015, 0.05248058141619, -0.26792717880897, -0.07181553246426, 0.96674755240348, 0.96674755240348, -0.07181553246426, -0.26792717880897, 0.05248058141619, 0.04143203796015, -0.01381067932005])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.17677669529664, -0.53033008588991, 0.53033008588991, -0.17677669529664, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000])
    elif family=="rbio3.7":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([-0.00302108610126, -0.00906325830378, 0.01683176542131, 0.07466398507402, -0.03133297870736, -0.30115912592284, 0.02649924094535, 0.95164212189718, -0.95164212189718, -0.02649924094535, 0.30115912592284, 0.03133297870736, -0.07466398507402, -0.01683176542131, 0.00906325830378, 0.00302108610126])
        rec_lo = np.array([0.00302108610126, -0.00906325830378, -0.01683176542131, 0.07466398507402, 0.03133297870736, -0.30115912592284, -0.02649924094535, 0.95164212189718, 0.95164212189718, -0.02649924094535, -0.30115912592284, 0.03133297870736, 0.07466398507402, -0.01683176542131, -0.00906325830378, 0.00302108610126])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.17677669529664, -0.53033008588991, 0.53033008588991, -0.17677669529664, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000])
    elif family=="rbio3.9":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.17677669529664, 0.53033008588991, 0.53033008588991, 0.17677669529664, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([0.00067974437278, 0.00203923311835, -0.00506031921961, -0.02061891264111, 0.01411278793018, 0.09913478249423, -0.01230013626942, -0.32019196836078, -0.00205002271157, 0.94212570067821, -0.94212570067821, 0.00205002271157, 0.32019196836078, 0.01230013626942, -0.09913478249423, -0.01411278793018, 0.02061891264111, 0.00506031921961, -0.00203923311835, -0.00067974437278])
        rec_lo = np.array([-0.00067974437278, 0.00203923311835, 0.00506031921961, -0.02061891264111, -0.01411278793018, 0.09913478249423, 0.01230013626942, -0.32019196836078, 0.00205002271157, 0.94212570067821, 0.94212570067821, 0.00205002271157, -0.32019196836078, 0.01230013626942, 0.09913478249423, -0.01411278793018, -0.02061891264111, 0.00506031921961, 0.00203923311835, -0.00067974437278])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.17677669529664, -0.53033008588991, 0.53033008588991, -0.17677669529664, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000])
    elif family=="rbio4.4":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, -0.06453888262870, -0.04068941760916, 0.41809227322162, 0.78848561640558, 0.41809227322162, -0.04068941760916, -0.06453888262870, 0.00000000000000])
        dec_hi = np.array([-0.03782845550726, -0.02384946501956, 0.11062440441844, 0.37740285561283, -0.85269867900889, 0.37740285561283, 0.11062440441844, -0.02384946501956, -0.03782845550726, 0.00000000000000])
        rec_lo = np.array([0.03782845550726, -0.02384946501956, -0.11062440441844, 0.37740285561283, 0.85269867900889, 0.37740285561283, -0.11062440441844, -0.02384946501956, 0.03782845550726, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, -0.06453888262870, 0.04068941760916, 0.41809227322162, -0.78848561640558, 0.41809227322162, 0.04068941760916, -0.06453888262870, -0.00000000000000])
    elif family=="rbio5.5":
        dec_lo = np.array([0.00000000000000, 0.01345670945912, -0.00269496688011, -0.13670658466433, -0.09350469740094, 0.47680326579848, 0.89950610974865, 0.47680326579848, -0.09350469740094, -0.13670658466433, -0.00269496688011, 0.01345670945912])
        dec_hi = np.array([-0.00000000000000, 0.03968708834741, -0.00794810863724, -0.05446378846824, -0.34560528195603, 0.73666018142821, -0.34560528195603, -0.05446378846824, -0.00794810863724, 0.03968708834741, -0.00000000000000, 0.00000000000000])
        rec_lo = np.array([0.00000000000000, 0.03968708834741, 0.00794810863724, -0.05446378846824, 0.34560528195603, 0.73666018142821, 0.34560528195603, -0.05446378846824, 0.00794810863724, 0.03968708834741, 0.00000000000000, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.01345670945912, -0.00269496688011, 0.13670658466433, -0.09350469740094, -0.47680326579848, 0.89950610974865, -0.47680326579848, -0.09350469740094, 0.13670658466433, -0.00269496688011, -0.01345670945912])
    elif family=="rbio6.8":
        dec_lo = np.array([0.00000000000000, 0.00000000000000, 0.00000000000000, 0.00000000000000, 0.01442628250562, 0.01446750489679, -0.07872200106263, -0.04036797903034, 0.41784910915027, 0.75890772945365, 0.41784910915027, -0.04036797903034, -0.07872200106263, 0.01446750489679, 0.01442628250562, 0.00000000000000, 0.00000000000000, 0.00000000000000])
        dec_hi = np.array([-0.00190883173648, -0.00191428612909, 0.01699063986760, 0.01193456527973, -0.04973290349094, -0.07726317316720, 0.09405920349574, 0.42079628460983, -0.82592299745840, 0.42079628460983, 0.09405920349574, -0.07726317316720, -0.04973290349094, 0.01193456527973, 0.01699063986760, -0.00191428612909, -0.00190883173648, 0.00000000000000])
        rec_lo = np.array([0.00190883173648, -0.00191428612909, -0.01699063986760, 0.01193456527973, 0.04973290349094, -0.07726317316720, -0.09405920349574, 0.42079628460983, 0.82592299745840, 0.42079628460983, -0.09405920349574, -0.07726317316720, 0.04973290349094, 0.01193456527973, -0.01699063986760, -0.00191428612909, 0.00190883173648, 0.00000000000000])
        rec_hi = np.array([0.00000000000000, -0.00000000000000, 0.00000000000000, -0.00000000000000, 0.01442628250562, -0.01446750489679, -0.07872200106263, 0.04036797903034, 0.41784910915027, -0.75890772945365, 0.41784910915027, 0.04036797903034, -0.07872200106263, -0.01446750489679, 0.01442628250562, -0.00000000000000, 0.00000000000000, -0.00000000000000])
    elif family=="sym2":
        dec_lo = np.array([-0.12940952255092, 0.22414386804186, 0.83651630373747, 0.48296291314469])
        dec_hi = np.array([-0.48296291314469, 0.83651630373747, -0.22414386804186, -0.12940952255092])
        rec_lo = np.array([0.48296291314469, 0.83651630373747, 0.22414386804186, -0.12940952255092])
        rec_hi = np.array([-0.12940952255092, -0.22414386804186, 0.83651630373747, -0.48296291314469])
    elif family=="sym3":
        dec_lo = np.array([0.03522629188210, -0.08544127388224, -0.13501102001039, 0.45987750211933, 0.80689150931334, 0.33267055295096])
        dec_hi = np.array([-0.33267055295096, 0.80689150931334, -0.45987750211933, -0.13501102001039, 0.08544127388224, 0.03522629188210])
        rec_lo = np.array([0.33267055295096, 0.80689150931334, 0.45987750211933, -0.13501102001039, -0.08544127388224, 0.03522629188210])
        rec_hi = np.array([0.03522629188210, 0.08544127388224, -0.13501102001039, -0.45987750211933, 0.80689150931334, -0.33267055295096])
    elif family=="sym4":
        dec_lo = np.array([-0.07576571478927, -0.02963552764600, 0.49761866763202, 0.80373875180592, 0.29785779560528, -0.09921954357685, -0.01260396726204, 0.03222310060404])
        dec_hi = np.array([-0.03222310060404, -0.01260396726204, 0.09921954357685, 0.29785779560528, -0.80373875180592, 0.49761866763202, 0.02963552764600, -0.07576571478927])
        rec_lo = np.array([0.03222310060404, -0.01260396726204, -0.09921954357685, 0.29785779560528, 0.80373875180592, 0.49761866763202, -0.02963552764600, -0.07576571478927])
        rec_hi = np.array([-0.07576571478927, 0.02963552764600, 0.49761866763202, -0.80373875180592, 0.29785779560528, 0.09921954357685, -0.01260396726204, -0.03222310060404])
    elif family=="sym5":
        dec_lo = np.array([0.02733306834508, 0.02951949092577, -0.03913424930238, 0.19939753397739, 0.72340769040242, 0.63397896345821, 0.01660210576452, -0.17532808990845, -0.02110183402476, 0.01953888273529])
        dec_hi = np.array([-0.01953888273529, -0.02110183402476, 0.17532808990845, 0.01660210576452, -0.63397896345821, 0.72340769040242, -0.19939753397739, -0.03913424930238, -0.02951949092577, 0.02733306834508])
        rec_lo = np.array([0.01953888273529, -0.02110183402476, -0.17532808990845, 0.01660210576452, 0.63397896345821, 0.72340769040242, 0.19939753397739, -0.03913424930238, 0.02951949092577, 0.02733306834508])
        rec_hi = np.array([0.02733306834508, -0.02951949092577, -0.03913424930238, -0.19939753397739, 0.72340769040242, -0.63397896345821, 0.01660210576452, 0.17532808990845, -0.02110183402476, -0.01953888273529])
    elif family=="sym6":
        dec_lo = np.array([0.01540410932703, 0.00349071208422, -0.11799011114819, -0.04831174258563, 0.49105594192675, 0.78764114103019, 0.33792942172762, -0.07263752278646, -0.02106029251230, 0.04472490177067, 0.00176771186424, -0.00780070832503])
        dec_hi = np.array([0.00780070832503, 0.00176771186424, -0.04472490177067, -0.02106029251230, 0.07263752278646, 0.33792942172762, -0.78764114103019, 0.49105594192675, 0.04831174258563, -0.11799011114819, -0.00349071208422, 0.01540410932703])
        rec_lo = np.array([-0.00780070832503, 0.00176771186424, 0.04472490177067, -0.02106029251230, -0.07263752278646, 0.33792942172762, 0.78764114103019, 0.49105594192675, -0.04831174258563, -0.11799011114819, 0.00349071208422, 0.01540410932703])
        rec_hi = np.array([0.01540410932703, -0.00349071208422, -0.11799011114819, 0.04831174258563, 0.49105594192675, -0.78764114103019, 0.33792942172762, 0.07263752278646, -0.02106029251230, -0.04472490177067, 0.00176771186424, 0.00780070832503])
    elif family=="sym7":
        dec_lo = np.array([0.00268181456826, -0.00104738488868, -0.01263630340325, 0.03051551316596, 0.06789269350137, -0.04955283493713, 0.01744125508686, 0.53610191709176, 0.76776431700316, 0.28862963175151, -0.14004724044296, -0.10780823770382, 0.00401024487153, 0.01026817670851])
        dec_hi = np.array([-0.01026817670851, 0.00401024487153, 0.10780823770382, -0.14004724044296, -0.28862963175151, 0.76776431700316, -0.53610191709176, 0.01744125508686, 0.04955283493713, 0.06789269350137, -0.03051551316596, -0.01263630340325, 0.00104738488868, 0.00268181456826])
        rec_lo = np.array([0.01026817670851, 0.00401024487153, -0.10780823770382, -0.14004724044296, 0.28862963175151, 0.76776431700316, 0.53610191709176, 0.01744125508686, -0.04955283493713, 0.06789269350137, 0.03051551316596, -0.01263630340325, -0.00104738488868, 0.00268181456826])
        rec_hi = np.array([0.00268181456826, 0.00104738488868, -0.01263630340325, -0.03051551316596, 0.06789269350137, 0.04955283493713, 0.01744125508686, -0.53610191709176, 0.76776431700316, -0.28862963175151, -0.14004724044296, 0.10780823770382, 0.00401024487153, -0.01026817670851])
    elif family=="sym8":
        dec_lo = np.array([-0.00338241595101, -0.00054213233179, 0.03169508781149, 0.00760748732492, -0.14329423835081, -0.06127335906766, 0.48135965125837, 0.77718575170052, 0.36444189483533, -0.05194583810771, -0.02721902991706, 0.04913717967361, 0.00380875201389, -0.01495225833705, -0.00030292051472, 0.00188995033276])
        dec_hi = np.array([-0.00188995033276, -0.00030292051472, 0.01495225833705, 0.00380875201389, -0.04913717967361, -0.02721902991706, 0.05194583810771, 0.36444189483533, -0.77718575170052, 0.48135965125837, 0.06127335906766, -0.14329423835081, -0.00760748732492, 0.03169508781149, 0.00054213233179, -0.00338241595101])
        rec_lo = np.array([0.00188995033276, -0.00030292051472, -0.01495225833705, 0.00380875201389, 0.04913717967361, -0.02721902991706, -0.05194583810771, 0.36444189483533, 0.77718575170052, 0.48135965125837, -0.06127335906766, -0.14329423835081, 0.00760748732492, 0.03169508781149, -0.00054213233179, -0.00338241595101])
        rec_hi = np.array([-0.00338241595101, 0.00054213233179, 0.03169508781149, -0.00760748732492, -0.14329423835081, 0.06127335906766, 0.48135965125837, -0.77718575170052, 0.36444189483533, 0.05194583810771, -0.02721902991706, -0.04913717967361, 0.00380875201389, 0.01495225833705, -0.00030292051472, -0.00188995033276])
    elif family=="sym9":
        dec_lo = np.array([0.00140091552591, 0.00061978088899, -0.01327196778182, -0.01152821020768, 0.03022487885828, 0.00058346274613, -0.05456895843083, 0.23876091460730, 0.71789708276441, 0.61733844914094, 0.03527248803527, -0.19155083129729, -0.01823377077940, 0.06207778930289, 0.00885926749340, -0.01026406402763, -0.00047315449868, 0.00106949003291])
        dec_hi = np.array([-0.00106949003291, -0.00047315449868, 0.01026406402763, 0.00885926749340, -0.06207778930289, -0.01823377077940, 0.19155083129729, 0.03527248803527, -0.61733844914094, 0.71789708276441, -0.23876091460730, -0.05456895843083, -0.00058346274613, 0.03022487885828, 0.01152821020768, -0.01327196778182, -0.00061978088899, 0.00140091552591])
        rec_lo = np.array([0.00106949003291, -0.00047315449868, -0.01026406402763, 0.00885926749340, 0.06207778930289, -0.01823377077940, -0.19155083129729, 0.03527248803527, 0.61733844914094, 0.71789708276441, 0.23876091460730, -0.05456895843083, 0.00058346274613, 0.03022487885828, -0.01152821020768, -0.01327196778182, 0.00061978088899, 0.00140091552591])
        rec_hi = np.array([0.00140091552591, -0.00061978088899, -0.01327196778182, 0.01152821020768, 0.03022487885828, -0.00058346274613, -0.05456895843083, -0.23876091460730, 0.71789708276441, -0.61733844914094, 0.03527248803527, 0.19155083129729, -0.01823377077940, -0.06207778930289, 0.00885926749340, 0.01026406402763, -0.00047315449868, -0.00106949003291])
    elif family=="sym10":
        dec_lo = np.array([0.00077015980911, 0.00009563267072, -0.00864129927702, -0.00146538258131, 0.04592723923109, 0.01160989390371, -0.15949427888492, -0.07088053578324, 0.47169066693844, 0.76951003702111, 0.38382676106709, -0.03553674047382, -0.03199005688243, 0.04999497207738, 0.00576491203358, -0.02035493981231, -0.00080435893202, 0.00459317358531, 0.00005703608362, -0.00045932942100])
        dec_hi = np.array([0.00045932942100, 0.00005703608362, -0.00459317358531, -0.00080435893202, 0.02035493981231, 0.00576491203358, -0.04999497207738, -0.03199005688243, 0.03553674047382, 0.38382676106709, -0.76951003702111, 0.47169066693844, 0.07088053578324, -0.15949427888492, -0.01160989390371, 0.04592723923109, 0.00146538258131, -0.00864129927702, -0.00009563267072, 0.00077015980911])
        rec_lo = np.array([-0.00045932942100, 0.00005703608362, 0.00459317358531, -0.00080435893202, -0.02035493981231, 0.00576491203358, 0.04999497207738, -0.03199005688243, -0.03553674047382, 0.38382676106709, 0.76951003702111, 0.47169066693844, -0.07088053578324, -0.15949427888492, 0.01160989390371, 0.04592723923109, -0.00146538258131, -0.00864129927702, 0.00009563267072, 0.00077015980911])
        rec_hi = np.array([0.00077015980911, -0.00009563267072, -0.00864129927702, 0.00146538258131, 0.04592723923109, -0.01160989390371, -0.15949427888492, 0.07088053578324, 0.47169066693844, -0.76951003702111, 0.38382676106709, 0.03553674047382, -0.03199005688243, -0.04999497207738, 0.00576491203358, 0.02035493981231, -0.00080435893202, -0.00459317358531, 0.00005703608362, 0.00045932942100])
    elif family=="sym11":
        dec_lo = np.array([0.00017172195070, -0.00003879565574, -0.00173436626730, 0.00058835273540, 0.00651249567477, -0.00985793482879, -0.02408084159586, 0.03703741597886, 0.06997679961073, -0.02283265102256, 0.09719839445891, 0.57202297801009, 0.73034354908840, 0.23768990904925, -0.20465479449580, -0.14460234370532, 0.03526675956447, 0.04300019068155, -0.00200347190011, -0.00638960366645, 0.00011053509764, 0.00048926361026])
        dec_hi = np.array([-0.00048926361026, 0.00011053509764, 0.00638960366645, -0.00200347190011, -0.04300019068155, 0.03526675956447, 0.14460234370532, -0.20465479449580, -0.23768990904925, 0.73034354908840, -0.57202297801009, 0.09719839445891, 0.02283265102256, 0.06997679961073, -0.03703741597886, -0.02408084159586, 0.00985793482879, 0.00651249567477, -0.00058835273540, -0.00173436626730, 0.00003879565574, 0.00017172195070])
        rec_lo = np.array([0.00048926361026, 0.00011053509764, -0.00638960366645, -0.00200347190011, 0.04300019068155, 0.03526675956447, -0.14460234370532, -0.20465479449580, 0.23768990904925, 0.73034354908840, 0.57202297801009, 0.09719839445891, -0.02283265102256, 0.06997679961073, 0.03703741597886, -0.02408084159586, -0.00985793482879, 0.00651249567477, 0.00058835273540, -0.00173436626730, -0.00003879565574, 0.00017172195070])
        rec_hi = np.array([0.00017172195070, 0.00003879565574, -0.00173436626730, -0.00058835273540, 0.00651249567477, 0.00985793482879, -0.02408084159586, -0.03703741597886, 0.06997679961073, 0.02283265102256, 0.09719839445891, -0.57202297801009, 0.73034354908840, -0.23768990904925, -0.20465479449580, 0.14460234370532, 0.03526675956447, -0.04300019068155, -0.00200347190011, 0.00638960366645, 0.00011053509764, -0.00048926361026])
    elif family=="sym12":
        dec_lo = np.array([0.00011196719425, -0.00001135392804, -0.00134975575557, 0.00018021409009, 0.00741496551765, -0.00140890924433, -0.02422072267501, 0.00755378061168, 0.04917931829966, -0.03584883073695, -0.02216230617034, 0.39888597239022, 0.76347909778366, 0.46274103121927, -0.07833262231634, -0.17037069723886, 0.01530174062248, 0.05780417944551, -0.00260439103133, -0.01458983644923, 0.00030764779631, 0.00235029761418, -0.00001815807886, -0.00017906658698])
        dec_hi = np.array([0.00017906658698, -0.00001815807886, -0.00235029761418, 0.00030764779631, 0.01458983644923, -0.00260439103133, -0.05780417944551, 0.01530174062248, 0.17037069723886, -0.07833262231634, -0.46274103121927, 0.76347909778366, -0.39888597239022, -0.02216230617034, 0.03584883073695, 0.04917931829966, -0.00755378061168, -0.02422072267501, 0.00140890924433, 0.00741496551765, -0.00018021409009, -0.00134975575557, 0.00001135392804, 0.00011196719425])
        rec_lo = np.array([-0.00017906658698, -0.00001815807886, 0.00235029761418, 0.00030764779631, -0.01458983644923, -0.00260439103133, 0.05780417944551, 0.01530174062248, -0.17037069723886, -0.07833262231634, 0.46274103121927, 0.76347909778366, 0.39888597239022, -0.02216230617034, -0.03584883073695, 0.04917931829966, 0.00755378061168, -0.02422072267501, -0.00140890924433, 0.00741496551765, 0.00018021409009, -0.00134975575557, -0.00001135392804, 0.00011196719425])
        rec_hi = np.array([0.00011196719425, 0.00001135392804, -0.00134975575557, -0.00018021409009, 0.00741496551765, 0.00140890924433, -0.02422072267501, -0.00755378061168, 0.04917931829966, 0.03584883073695, -0.02216230617034, -0.39888597239022, 0.76347909778366, -0.46274103121927, -0.07833262231634, 0.17037069723886, 0.01530174062248, -0.05780417944551, -0.00260439103133, 0.01458983644923, 0.00030764779631, -0.00235029761418, -0.00001815807886, 0.00017906658698])
    elif family=="sym13":
        dec_lo = np.array([0.00006820325263, -0.00003573862365, -0.00113606343893, -0.00017094285853, 0.00752622538997, 0.00529635973873, -0.02021676813339, -0.01721164272630, 0.01386249743585, -0.05975062771794, -0.12436246075153, 0.19770481877118, 0.69573915056150, 0.64456438390119, 0.11023022302137, -0.14049009311363, 0.00881975767042, 0.09292603089914, 0.01761829688065, -0.02074968632552, -0.00149244727426, 0.00567485376012, 0.00041326119884, -0.00072136438514, 0.00003690537342, 0.00007042986691])
        dec_hi = np.array([-0.00007042986691, 0.00003690537342, 0.00072136438514, 0.00041326119884, -0.00567485376012, -0.00149244727426, 0.02074968632552, 0.01761829688065, -0.09292603089914, 0.00881975767042, 0.14049009311363, 0.11023022302137, -0.64456438390119, 0.69573915056150, -0.19770481877118, -0.12436246075153, 0.05975062771794, 0.01386249743585, 0.01721164272630, -0.02021676813339, -0.00529635973873, 0.00752622538997, 0.00017094285853, -0.00113606343893, 0.00003573862365, 0.00006820325263])
        rec_lo = np.array([0.00007042986691, 0.00003690537342, -0.00072136438514, 0.00041326119884, 0.00567485376012, -0.00149244727426, -0.02074968632552, 0.01761829688065, 0.09292603089914, 0.00881975767042, -0.14049009311363, 0.11023022302137, 0.64456438390119, 0.69573915056150, 0.19770481877118, -0.12436246075153, -0.05975062771794, 0.01386249743585, -0.01721164272630, -0.02021676813339, 0.00529635973873, 0.00752622538997, -0.00017094285853, -0.00113606343893, -0.00003573862365, 0.00006820325263])
        rec_hi = np.array([0.00006820325263, 0.00003573862365, -0.00113606343893, 0.00017094285853, 0.00752622538997, -0.00529635973873, -0.02021676813339, 0.01721164272630, 0.01386249743585, 0.05975062771794, -0.12436246075153, -0.19770481877118, 0.69573915056150, -0.64456438390119, 0.11023022302137, 0.14049009311363, 0.00881975767042, -0.09292603089914, 0.01761829688065, 0.02074968632552, -0.00149244727426, -0.00567485376012, 0.00041326119884, 0.00072136438514, 0.00003690537342, -0.00007042986691])
    elif family=="sym14":
        dec_lo = np.array([-0.00002587909027, 0.00001121086581, 0.00039843567298, -0.00006286542481, -0.00257944172593, 0.00036647657366, 0.01003769371767, -0.00275377479122, -0.02919621776404, 0.00428052049902, 0.03743308836285, -0.05763449835133, -0.03531811211498, 0.39320152196209, 0.75997624196109, 0.47533576263421, -0.05811182331772, -0.15999741114652, 0.02589858753105, 0.06982761636181, -0.00236504883674, -0.01943931426363, 0.00101314198718, 0.00453267747195, -0.00007321421357, -0.00060576018247, 0.00001932901697, 0.00004461897799])
        dec_hi = np.array([-0.00004461897799, 0.00001932901697, 0.00060576018247, -0.00007321421357, -0.00453267747195, 0.00101314198718, 0.01943931426363, -0.00236504883674, -0.06982761636181, 0.02589858753105, 0.15999741114652, -0.05811182331772, -0.47533576263421, 0.75997624196109, -0.39320152196209, -0.03531811211498, 0.05763449835133, 0.03743308836285, -0.00428052049902, -0.02919621776404, 0.00275377479122, 0.01003769371767, -0.00036647657366, -0.00257944172593, 0.00006286542481, 0.00039843567298, -0.00001121086581, -0.00002587909027])
        rec_lo = np.array([0.00004461897799, 0.00001932901697, -0.00060576018247, -0.00007321421357, 0.00453267747195, 0.00101314198718, -0.01943931426363, -0.00236504883674, 0.06982761636181, 0.02589858753105, -0.15999741114652, -0.05811182331772, 0.47533576263421, 0.75997624196109, 0.39320152196209, -0.03531811211498, -0.05763449835133, 0.03743308836285, 0.00428052049902, -0.02919621776404, -0.00275377479122, 0.01003769371767, 0.00036647657366, -0.00257944172593, -0.00006286542481, 0.00039843567298, 0.00001121086581, -0.00002587909027])
        rec_hi = np.array([-0.00002587909027, -0.00001121086581, 0.00039843567298, 0.00006286542481, -0.00257944172593, -0.00036647657366, 0.01003769371767, 0.00275377479122, -0.02919621776404, -0.00428052049902, 0.03743308836285, 0.05763449835133, -0.03531811211498, -0.39320152196209, 0.75997624196109, -0.47533576263421, -0.05811182331772, 0.15999741114652, 0.02589858753105, -0.06982761636181, -0.00236504883674, 0.01943931426363, 0.00101314198718, -0.00453267747195, -0.00007321421357, 0.00060576018247, 0.00001932901697, -0.00004461897799])
    elif family=="sym15":
        dec_lo = np.array([0.00000971241974, -0.00000735966680, -0.00016066186637, 0.00005512254786, 0.00107056721946, -0.00026731644647, -0.00359016544737, 0.00342345073635, 0.01007997708791, -0.01940501143093, -0.03887671687683, 0.02193764271975, 0.04073547969681, -0.04108266663538, 0.11153369514262, 0.57864041521503, 0.72184302963618, 0.24396270543217, -0.19662635876624, -0.13405629845625, 0.06839331006048, 0.06796982904488, -0.00874478888648, -0.01717125278164, 0.00152613827818, 0.00348102873706, -0.00010815440169, -0.00040216853760, 0.00002171789015, 0.00002866070853])
        dec_hi = np.array([-0.00002866070853, 0.00002171789015, 0.00040216853760, -0.00010815440169, -0.00348102873706, 0.00152613827818, 0.01717125278164, -0.00874478888648, -0.06796982904488, 0.06839331006048, 0.13405629845625, -0.19662635876624, -0.24396270543217, 0.72184302963618, -0.57864041521503, 0.11153369514262, 0.04108266663538, 0.04073547969681, -0.02193764271975, -0.03887671687683, 0.01940501143093, 0.01007997708791, -0.00342345073635, -0.00359016544737, 0.00026731644647, 0.00107056721946, -0.00005512254786, -0.00016066186637, 0.00000735966680, 0.00000971241974])
        rec_lo = np.array([0.00002866070853, 0.00002171789015, -0.00040216853760, -0.00010815440169, 0.00348102873706, 0.00152613827818, -0.01717125278164, -0.00874478888648, 0.06796982904488, 0.06839331006048, -0.13405629845625, -0.19662635876624, 0.24396270543217, 0.72184302963618, 0.57864041521503, 0.11153369514262, -0.04108266663538, 0.04073547969681, 0.02193764271975, -0.03887671687683, -0.01940501143093, 0.01007997708791, 0.00342345073635, -0.00359016544737, -0.00026731644647, 0.00107056721946, 0.00005512254786, -0.00016066186637, -0.00000735966680, 0.00000971241974])
        rec_hi = np.array([0.00000971241974, 0.00000735966680, -0.00016066186637, -0.00005512254786, 0.00107056721946, 0.00026731644647, -0.00359016544737, -0.00342345073635, 0.01007997708791, 0.01940501143093, -0.03887671687683, -0.02193764271975, 0.04073547969681, 0.04108266663538, 0.11153369514262, -0.57864041521503, 0.72184302963618, -0.24396270543217, -0.19662635876624, 0.13405629845625, 0.06839331006048, -0.06796982904488, -0.00874478888648, 0.01717125278164, 0.00152613827818, -0.00348102873706, -0.00010815440169, 0.00040216853760, 0.00002171789015, -0.00002866070853])
    elif family=="sym16":
        dec_lo = np.array([0.00000623000670, -0.00000311355641, -0.00010943147930, 0.00002807858213, 0.00085235471080, -0.00010844562231, -0.00388091225260, 0.00071821197883, 0.01266673165986, -0.00312651717227, -0.03105120284355, 0.00486927440490, 0.03233309161066, -0.06698304907022, -0.03457422841697, 0.39712293362064, 0.75652498787570, 0.47534280601152, -0.05404060138761, -0.15959219218521, 0.03072113906330, 0.07803785290342, -0.00351027506837, -0.02495275804629, 0.00135984474248, 0.00693776113080, -0.00022211647621, -0.00133872060669, 0.00003656592483, 0.00016545679579, -0.00000539648318, -0.00001079798210])
        dec_hi = np.array([0.00001079798210, -0.00000539648318, -0.00016545679579, 0.00003656592483, 0.00133872060669, -0.00022211647621, -0.00693776113080, 0.00135984474248, 0.02495275804629, -0.00351027506837, -0.07803785290342, 0.03072113906330, 0.15959219218521, -0.05404060138761, -0.47534280601152, 0.75652498787570, -0.39712293362064, -0.03457422841697, 0.06698304907022, 0.03233309161066, -0.00486927440490, -0.03105120284355, 0.00312651717227, 0.01266673165986, -0.00071821197883, -0.00388091225260, 0.00010844562231, 0.00085235471080, -0.00002807858213, -0.00010943147930, 0.00000311355641, 0.00000623000670])
        rec_lo = np.array([-0.00001079798210, -0.00000539648318, 0.00016545679579, 0.00003656592483, -0.00133872060669, -0.00022211647621, 0.00693776113080, 0.00135984474248, -0.02495275804629, -0.00351027506837, 0.07803785290342, 0.03072113906330, -0.15959219218521, -0.05404060138761, 0.47534280601152, 0.75652498787570, 0.39712293362064, -0.03457422841697, -0.06698304907022, 0.03233309161066, 0.00486927440490, -0.03105120284355, -0.00312651717227, 0.01266673165986, 0.00071821197883, -0.00388091225260, -0.00010844562231, 0.00085235471080, 0.00002807858213, -0.00010943147930, -0.00000311355641, 0.00000623000670])
        rec_hi = np.array([0.00000623000670, 0.00000311355641, -0.00010943147930, -0.00002807858213, 0.00085235471080, 0.00010844562231, -0.00388091225260, -0.00071821197883, 0.01266673165986, 0.00312651717227, -0.03105120284355, -0.00486927440490, 0.03233309161066, 0.06698304907022, -0.03457422841697, -0.39712293362064, 0.75652498787570, -0.47534280601152, -0.05404060138761, 0.15959219218521, 0.03072113906330, -0.07803785290342, -0.00351027506837, 0.02495275804629, 0.00135984474248, -0.00693776113080, -0.00022211647621, 0.00133872060669, 0.00003656592483, -0.00016545679579, -0.00000539648318, 0.00001079798210])
    elif family=="sym17":
        dec_lo = np.array([0.00000429734333, 0.00000278012669, -0.00006293702598, -0.00001350638340, 0.00047599638026, -0.00013864230268, -0.00274167597568, 0.00085677007019, 0.01048236693303, -0.00481921280318, -0.03329138349236, 0.01790395221434, 0.10475461484223, 0.01727117821052, -0.11856693261144, 0.14239835041468, 0.65071662920455, 0.68148899534493, 0.18053958458111, -0.15507600534975, -0.08607087472073, 0.01615880872592, -0.00726163475093, -0.01803889724192, 0.00995298252351, 0.01239698836665, -0.00190540768985, -0.00393232527980, 0.00005840042869, 0.00071982706421, 0.00002520793314, -0.00007607124406, -0.00000245271634, 0.00000379125319])
        dec_hi = np.array([-0.00000379125319, -0.00000245271634, 0.00007607124406, 0.00002520793314, -0.00071982706421, 0.00005840042869, 0.00393232527980, -0.00190540768985, -0.01239698836665, 0.00995298252351, 0.01803889724192, -0.00726163475093, -0.01615880872592, -0.08607087472073, 0.15507600534975, 0.18053958458111, -0.68148899534493, 0.65071662920455, -0.14239835041468, -0.11856693261144, -0.01727117821052, 0.10475461484223, -0.01790395221434, -0.03329138349236, 0.00481921280318, 0.01048236693303, -0.00085677007019, -0.00274167597568, 0.00013864230268, 0.00047599638026, 0.00001350638340, -0.00006293702598, -0.00000278012669, 0.00000429734333])
        rec_lo = np.array([0.00000379125319, -0.00000245271634, -0.00007607124406, 0.00002520793314, 0.00071982706421, 0.00005840042869, -0.00393232527980, -0.00190540768985, 0.01239698836665, 0.00995298252351, -0.01803889724192, -0.00726163475093, 0.01615880872592, -0.08607087472073, -0.15507600534975, 0.18053958458111, 0.68148899534493, 0.65071662920455, 0.14239835041468, -0.11856693261144, 0.01727117821052, 0.10475461484223, 0.01790395221434, -0.03329138349236, -0.00481921280318, 0.01048236693303, 0.00085677007019, -0.00274167597568, -0.00013864230268, 0.00047599638026, -0.00001350638340, -0.00006293702598, 0.00000278012669, 0.00000429734333])
        rec_hi = np.array([0.00000429734333, -0.00000278012669, -0.00006293702598, 0.00001350638340, 0.00047599638026, 0.00013864230268, -0.00274167597568, -0.00085677007019, 0.01048236693303, 0.00481921280318, -0.03329138349236, -0.01790395221434, 0.10475461484223, -0.01727117821052, -0.11856693261144, -0.14239835041468, 0.65071662920455, -0.68148899534493, 0.18053958458111, 0.15507600534975, -0.08607087472073, -0.01615880872592, -0.00726163475093, 0.01803889724192, 0.00995298252351, -0.01239698836665, -0.00190540768985, 0.00393232527980, 0.00005840042869, -0.00071982706421, 0.00002520793314, 0.00007607124406, -0.00000245271634, -0.00000379125319])
    elif family=="sym18":
        dec_lo = np.array([0.00000261261256, 0.00000135491576, -0.00004524675787, -0.00001402099258, 0.00039616840638, 0.00007021273459, -0.00231387181451, -0.00041152110924, 0.00950216439096, 0.00164298639728, -0.03032509108937, -0.00507708516076, 0.08421992997039, 0.03399566710395, -0.15993814866932, -0.05202915898395, 0.47396905989394, 0.75362914010179, 0.40148386057062, -0.03248057329014, -0.07379920729061, 0.02852959703904, 0.00627794455431, -0.03171268473181, -0.00326074420007, 0.01501235634425, 0.00108778478960, -0.00523978968303, -0.00018877623941, 0.00142808632708, 0.00004741614518, -0.00026583011024, -0.00000985881603, 0.00002955743762, 0.00000078472981, -0.00000151315307])
        dec_hi = np.array([0.00000151315307, 0.00000078472981, -0.00002955743762, -0.00000985881603, 0.00026583011024, 0.00004741614518, -0.00142808632708, -0.00018877623941, 0.00523978968303, 0.00108778478960, -0.01501235634425, -0.00326074420007, 0.03171268473181, 0.00627794455431, -0.02852959703904, -0.07379920729061, 0.03248057329014, 0.40148386057062, -0.75362914010179, 0.47396905989394, 0.05202915898395, -0.15993814866932, -0.03399566710395, 0.08421992997039, 0.00507708516076, -0.03032509108937, -0.00164298639728, 0.00950216439096, 0.00041152110924, -0.00231387181451, -0.00007021273459, 0.00039616840638, 0.00001402099258, -0.00004524675787, -0.00000135491576, 0.00000261261256])
        rec_lo = np.array([-0.00000151315307, 0.00000078472981, 0.00002955743762, -0.00000985881603, -0.00026583011024, 0.00004741614518, 0.00142808632708, -0.00018877623941, -0.00523978968303, 0.00108778478960, 0.01501235634425, -0.00326074420007, -0.03171268473181, 0.00627794455431, 0.02852959703904, -0.07379920729061, -0.03248057329014, 0.40148386057062, 0.75362914010179, 0.47396905989394, -0.05202915898395, -0.15993814866932, 0.03399566710395, 0.08421992997039, -0.00507708516076, -0.03032509108937, 0.00164298639728, 0.00950216439096, -0.00041152110924, -0.00231387181451, 0.00007021273459, 0.00039616840638, -0.00001402099258, -0.00004524675787, 0.00000135491576, 0.00000261261256])
        rec_hi = np.array([0.00000261261256, -0.00000135491576, -0.00004524675787, 0.00001402099258, 0.00039616840638, -0.00007021273459, -0.00231387181451, 0.00041152110924, 0.00950216439096, -0.00164298639728, -0.03032509108937, 0.00507708516076, 0.08421992997039, -0.03399566710395, -0.15993814866932, 0.05202915898395, 0.47396905989394, -0.75362914010179, 0.40148386057062, 0.03248057329014, -0.07379920729061, -0.02852959703904, 0.00627794455431, 0.03171268473181, -0.00326074420007, -0.01501235634425, 0.00108778478960, 0.00523978968303, -0.00018877623941, -0.00142808632708, 0.00004741614518, 0.00026583011024, -0.00000985881603, -0.00002955743762, 0.00000078472981, 0.00000151315307])
    elif family=="sym19":
        dec_lo = np.array([0.00000054877328, -0.00000064636513, -0.00001188051827, 0.00000887331217, 0.00011553923334, -0.00004612039600, -0.00063576451500, 0.00015915804768, 0.00212142502818, -0.00116070325721, -0.00512220500258, 0.00796843832061, 0.01579743929567, -0.02265199337825, -0.04663598353494, 0.00701557385717, 0.00895459117304, -0.06752505804029, 0.10902582508128, 0.57814494533861, 0.71955552571639, 0.25826616923728, -0.17659686625203, -0.11624173010740, 0.09363084341590, 0.08407267627925, -0.01690823486135, -0.02770989693131, 0.00431935187489, 0.00826223695553, -0.00061792232780, -0.00170496026116, 0.00012930767651, 0.00027621877686, -0.00001682138703, -0.00002815113866, 0.00000206231706, 0.00000175093680])
        dec_hi = np.array([-0.00000175093680, 0.00000206231706, 0.00002815113866, -0.00001682138703, -0.00027621877686, 0.00012930767651, 0.00170496026116, -0.00061792232780, -0.00826223695553, 0.00431935187489, 0.02770989693131, -0.01690823486135, -0.08407267627925, 0.09363084341590, 0.11624173010740, -0.17659686625203, -0.25826616923728, 0.71955552571639, -0.57814494533861, 0.10902582508128, 0.06752505804029, 0.00895459117304, -0.00701557385717, -0.04663598353494, 0.02265199337825, 0.01579743929567, -0.00796843832061, -0.00512220500258, 0.00116070325721, 0.00212142502818, -0.00015915804768, -0.00063576451500, 0.00004612039600, 0.00011553923334, -0.00000887331217, -0.00001188051827, 0.00000064636513, 0.00000054877328])
        rec_lo = np.array([0.00000175093680, 0.00000206231706, -0.00002815113866, -0.00001682138703, 0.00027621877686, 0.00012930767651, -0.00170496026116, -0.00061792232780, 0.00826223695553, 0.00431935187489, -0.02770989693131, -0.01690823486135, 0.08407267627925, 0.09363084341590, -0.11624173010740, -0.17659686625203, 0.25826616923728, 0.71955552571639, 0.57814494533861, 0.10902582508128, -0.06752505804029, 0.00895459117304, 0.00701557385717, -0.04663598353494, -0.02265199337825, 0.01579743929567, 0.00796843832061, -0.00512220500258, -0.00116070325721, 0.00212142502818, 0.00015915804768, -0.00063576451500, -0.00004612039600, 0.00011553923334, 0.00000887331217, -0.00001188051827, -0.00000064636513, 0.00000054877328])
        rec_hi = np.array([0.00000054877328, 0.00000064636513, -0.00001188051827, -0.00000887331217, 0.00011553923334, 0.00004612039600, -0.00063576451500, -0.00015915804768, 0.00212142502818, 0.00116070325721, -0.00512220500258, -0.00796843832061, 0.01579743929567, 0.02265199337825, -0.04663598353494, -0.00701557385717, 0.00895459117304, 0.06752505804029, 0.10902582508128, -0.57814494533861, 0.71955552571639, -0.25826616923728, -0.17659686625203, 0.11624173010740, 0.09363084341590, -0.08407267627925, -0.01690823486135, 0.02770989693131, 0.00431935187489, -0.00826223695553, -0.00061792232780, 0.00170496026116, 0.00012930767651, -0.00027621877686, -0.00001682138703, 0.00002815113866, 0.00000206231706, -0.00000175093680])
    elif family=="sym20":
        dec_lo = np.array([0.00000036955375, -0.00000019015676, -0.00000791936141, 0.00000302566606, 0.00007992967836, -0.00001928412301, -0.00049473109157, 0.00007215991188, 0.00208899470819, -0.00030526283180, -0.00660658579909, 0.00142308735946, 0.01700404902339, -0.00331385738362, -0.03162943714496, 0.00812322835601, 0.02557934950941, -0.07899434492840, -0.02981936888033, 0.40583144434845, 0.75116272842273, 0.47199147510149, -0.05108834292107, -0.16057829841525, 0.03625095165393, 0.08891966802820, -0.00684370196507, -0.03537333675660, 0.00193859706724, 0.01215704094879, -0.00061112638580, -0.00347164780284, 0.00012544091723, 0.00074761085978, -0.00002661555034, -0.00011739133516, 0.00000452542221, 0.00001228725278, -0.00000032567026, -0.00000063291290])
        dec_hi = np.array([0.00000063291290, -0.00000032567026, -0.00001228725278, 0.00000452542221, 0.00011739133516, -0.00002661555034, -0.00074761085978, 0.00012544091723, 0.00347164780284, -0.00061112638580, -0.01215704094879, 0.00193859706724, 0.03537333675660, -0.00684370196507, -0.08891966802820, 0.03625095165393, 0.16057829841525, -0.05108834292107, -0.47199147510149, 0.75116272842273, -0.40583144434845, -0.02981936888033, 0.07899434492840, 0.02557934950941, -0.00812322835601, -0.03162943714496, 0.00331385738362, 0.01700404902339, -0.00142308735946, -0.00660658579909, 0.00030526283180, 0.00208899470819, -0.00007215991188, -0.00049473109157, 0.00001928412301, 0.00007992967836, -0.00000302566606, -0.00000791936141, 0.00000019015676, 0.00000036955375])
        rec_lo = np.array([-0.00000063291290, -0.00000032567026, 0.00001228725278, 0.00000452542221, -0.00011739133516, -0.00002661555034, 0.00074761085978, 0.00012544091723, -0.00347164780284, -0.00061112638580, 0.01215704094879, 0.00193859706724, -0.03537333675660, -0.00684370196507, 0.08891966802820, 0.03625095165393, -0.16057829841525, -0.05108834292107, 0.47199147510149, 0.75116272842273, 0.40583144434845, -0.02981936888033, -0.07899434492840, 0.02557934950941, 0.00812322835601, -0.03162943714496, -0.00331385738362, 0.01700404902339, 0.00142308735946, -0.00660658579909, -0.00030526283180, 0.00208899470819, 0.00007215991188, -0.00049473109157, -0.00001928412301, 0.00007992967836, 0.00000302566606, -0.00000791936141, -0.00000019015676, 0.00000036955375])
        rec_hi = np.array([0.00000036955375, 0.00000019015676, -0.00000791936141, -0.00000302566606, 0.00007992967836, 0.00001928412301, -0.00049473109157, -0.00007215991188, 0.00208899470819, 0.00030526283180, -0.00660658579909, -0.00142308735946, 0.01700404902339, 0.00331385738362, -0.03162943714496, -0.00812322835601, 0.02557934950941, 0.07899434492840, -0.02981936888033, -0.40583144434845, 0.75116272842273, -0.47199147510149, -0.05108834292107, 0.16057829841525, 0.03625095165393, -0.08891966802820, -0.00684370196507, 0.03537333675660, 0.00193859706724, -0.01215704094879, -0.00061112638580, 0.00347164780284, 0.00012544091723, -0.00074761085978, -0.00002661555034, 0.00011739133516, 0.00000452542221, -0.00001228725278, -0.00000032567026, 0.00000063291290])
    else:
        bob = 10

    return dec_lo, dec_hi, rec_lo, rec_hi


# -----------------------------------------------------------------------------


def _test_equivalency():
    '''
    Are coefficients the same for wavepy and pywavelets library calls?

    2020-06-23 bjs added more wavelet coefficients. Did a bit of optimization.
      the local code runs about 3 times faster than wavepy-1.0 module, and pywt still
      runs 10 times faster than the local code.  Sigh.

    2020-02-17 bjs tested for Python 3.7.4 and values for coeffs0, coeffs1 and coeffs2 
    showed 10e-5 to 10e-7 diffs between them, as did the reconstructed data. So the 
    libraries are all still in sync for these tests and python 3, pywavelets 1.1.1 vs 0.3.0
    
    '''
    import pywt
    import wavepy as wv     # need to install wavepy-0.1 module

    data = data64

    wavelet = 'coif3'   
    _mode = 'per'
    base_level = int(np.log(len(data))/np.log(2))   

    coeffs0                 = pywt.wavedec(data, wavelet, _mode, level=base_level)
    [coeffs1,length1,flag1] = wv.dwt.dwt(data, base_level, wavelet, _mode)
    coeffs2,length2,alert2   = dwt(wavelet, data, base_level)

    diffA = np.concatenate(coeffs0) - coeffs1
    diffB = coeffs1 - coeffs2

    length1 = [int(item) for item in length1]

    mean0 = np.mean(np.concatenate(coeffs0))
    mean1 = np.mean(coeffs1)
    mean2 = np.mean(coeffs2)

    # first 4 coeffs for pywt are different from other libs, but we never
    # zero these out since they are the broadest dyads, so it is a moot point.
    print("Mean coeffs (pywt, wavepy, local) - ",mean0, mean1, mean2)
    print("Mean data value                   - "+str(np.mean(data)))
    print("Compare coeffs, pywt vs wavepy  - Max Diff = " + str(np.max(diffA[5:]))+"   mean Diff A = "+str(np.mean(diffA[5:])))
    print("Compare coeffs, wavepy vs local - Max Diff = " + str(np.max(diffB))    +"   mean Diff B = "+str(np.mean(diffB)))

    recon0 = pywt.waverec(coeffs0, wavelet, _mode)
    recon1 = wv.dwt.idwt(coeffs1, wavelet, length1, flag1)
    recon2 = idwt(wavelet, coeffs2, length2, alert2)

    diffC = recon0 - recon1
    diffD = recon1 - recon2

    print("Compare recon, pywt vs wavepy  - Max Diff = " + str(np.max(diffC))+"   mean Diff C = "+str(np.mean(diffC)))
    print("Compare recon, wavepy vs local - Max Diff = " + str(np.max(diffD))+"   mean Diff D = "+str(np.mean(diffD)))
    


data64 = [   3.75646062e-01,  -7.52931591e-01,  -9.85668931e-01,
            -5.71847181e-01,  -4.12297490e-01,   5.94913227e-01,
            -7.60136820e-01,  -8.21366067e-01,   4.73895513e-02,
             7.18766056e-01,   5.64785518e-01,  -5.76128617e-01,
            -6.65498320e-01,   4.35884338e-01,  -8.77145793e-01,
            -2.30297001e-01,  -7.76746230e-01,  -3.13412228e-01,
             6.87443670e-01,   1.14553782e-01,   5.66057710e-01,
            -8.77337090e-01,  -6.39754937e-01,   7.18250389e-01,
            -9.24781352e-01,   9.88529513e-01,   5.45465600e-02,
            -8.35248333e-01,  -4.40469828e-01,   1.46422657e-01,
            -1.29651675e-01,  -2.50870615e-01,  -3.52361713e-01,
            -3.62400134e-01,   3.34141270e-01,  -8.83583193e-01,
            -9.70013743e-02,  -7.45088469e-01,  -2.96155143e-01,
             1.61548515e-01,   4.04399869e-01,   4.75285614e-01,
            -3.78595158e-01,   5.19088161e-01,  -1.16044345e-01,
            -7.51618934e-01,  -5.45780616e-01,   9.18209827e-01,
            -6.53026178e-01,   2.39033374e-01,  -6.95793553e-01,
             3.08469171e-01,   7.76517411e-01,   7.80547551e-01,
            -8.68320975e-01,  -8.94686333e-02,   1.14572022e-01,
             6.44247703e-01,   4.11723423e-01,  -8.62504626e-01,
            -6.17812166e-01,   5.91205829e-01,  -9.95255684e-01,
            -7.37408598e-01 ]


# Stub methods to simplify seeing total time in cProfile listing



def _test_timing():

    import pywt
    import wavepy as wv     # need to install wavepy-0.1 module

    def _do_pywt(n):
        for i in range(n):
            coeffs0 = pywt.wavedec(data64, 'coif3', 'per', level=int(np.log(len(data64)) / np.log(2)))
            recon0 = pywt.waverec(coeffs0, 'coif3', 'per')

    def _do_wavepy(n):
        for i in range(n):
            [coeffs1, length1, flag1] = wv.dwt.dwt(data64, int(np.log(len(data64)) / np.log(2)), 'coif3', 'per')
            length1 = [int(item) for item in length1]
            recon1 = wv.dwt.idwt(coeffs1, 'coif3', length1, flag1)

    def _do_local(n):
        for i in range(n):
            [coeffs2, width2, alert2] = dwt('coif3', data64, int(np.log(len(data64)) / np.log(2)))
            recon2 = idwt('coif3', coeffs2, width2, alert2)

    n = 100
    _do_pywt(n)
    _do_wavepy(n)
    _do_local(n)

if __name__ == '__main__':

    import cProfile
    import pstats as ps

    _test_equivalency()

    if os.path.exists("profile.data"):
        os.remove("profile.data")

    cProfile.run('_test_timing()', 'profile.data')
    p = ps.Stats('profile.data')
    p.strip_dirs().sort_stats('cumulative').print_stats()


        


